/******************************************************************************
 * @file:		gps_driver.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	Driver for the U-blox NEO-7P GPS module.
 ******************************************************************************/


#ifndef GPS_DRIVER_H_
#define GPS_DRIVER_H_

#include "em_gpio.h"

#include "shared.h"

/*******************************************************************************
 * 								Typedefs
 *******************************************************************************/

typedef enum {
	nr_utcTime,
	nr_geoPosition,
	nr_velNed,
	nr_navStatus,
	nr_pvtSol
} GPS_NavRes;


typedef enum {
	op_ppp, 				/* Precise point positioning. Requires continuous tracking and SBAS */
	op_continuousNoSBAS,	/* Continuous tracking with SBAS de-activated */
	op_continuousSBAS,		/* Continuous tracking with SBAS activated */
	op_psm_onOff,			/* Power save mode: On/Off operation for periods from 10s and higher. SBAS de-activated */
	op_psm_cyclic			/* Power save mode: Cyclic tracking for position fix between 1s and 10s. SBAS de-activated */
} GPS_OperationalMode;


typedef struct {
	GPS_OperationalMode mode;					/* Operational mode */
	uint32_t 			updatePeriod_ms;		/* Position fix update period. Valid for all modes */
	uint32_t 			searchPeriod_ms;		/* If position acquisition timeout, receiver will retry after specified time (PSM only)*/
	uint16_t			onPeriod_s;				/* Specifies how long the receiver stays in tracking state (PSM only) */
	bool				wakeToUpdateRTC;		/* If set, the receiver wakes every 5 min to calibrate RTC */
	bool				wakeToUpdateEPH;		/* If set, the receiver wakes approx. every 30 min to update Ephemeris */
	bool				waitForTimeFix;			/* Stays in acquisition mode until the GPS time is valid */
	bool				enterInactiveOnNoFix;	/* If set, the receiver enters inactive state as specified by "searchPeriod" (PSM only) */
	bool				extIntAwake;			/* If set, the receiver is kept awake as long as the EXTINT pin is high */
	bool				extIntBackup;			/* If set, the receiver is forced into inactive mode as long as the EXTINT pin is low */
} GPS_OpConfig;


typedef struct {

} GPS_TPConfig;

typedef struct {
	int32_t		nano;
	uint8_t 	sec;
	uint8_t 	min;
	uint8_t 	hour;
	uint8_t 	day;
	uint8_t 	month;
	uint16_t 	year;
	bool		valid;
} GPS_Time;


typedef struct {
	int32_t		velNorth_mms;
	int32_t		velEast_mms;
	int32_t		velDown_mms;
	int32_t		speed3d_mms;
	int32_t		gSpeed2d_mms;
	int32_t		heading_deg;
} GPS_Velocity;


typedef struct {
	GeoPosition 	position;
	GPS_Velocity	velocity;
	GPS_Time 		time;
} GPS_PVTSolution;


/*******************************************************************************
 * 							Function declarations
 *******************************************************************************/

bool GPS_Init();
bool GPS_PowerOn();
void GPS_PowerOff();
void GPS_Reset();
bool GPS_UsingTxReady();

bool GPS_LoadPermanentConfig();
bool GPS_ClearPermanentConfig();
bool GPS_SaveNewPermanentConfig();

bool GPS_DefaultMsgConfig();
bool GPS_DefaultSPIPortConfig(bool useTxReady);
bool GPS_DefaultNAV5Config();
bool GPS_SetOperationalMode(GPS_OpConfig * config);
bool GPS_ConfigTimepulse(GPS_TPConfig * config);

bool GPS_GetPVTSolution(GPS_PVTSolution * sol, bool poll);
bool GPS_GetTimeUTC(GPS_Time * utcTime);


/*******************************************************************************
 * 						Default configuration messages
 *******************************************************************************/

#define GPS_DEFAULT_OP_CONFIG														\
{																					\
	op_continuousNoSBAS,	/* Continuous mode with SBAS de-activated */ 			\
	10000,					/* Update fix every 10s */								\
	2000,					/* Go inactive for 2s when searching */					\
	2,						/* On time for 2s */									\
	true,					/* Wake to update RTC */								\
	true,					/* Wake to update Ephemeris */ 							\
	true,					/* Wait for time fix before going into tracking mode */	\
	true,					/* Enter inactive state on no fix */					\
	true,					/* Enable extint awake when high */						\
	false					/* Disable extint inactive when low */					\
}


#endif /* GPS_DRIVER_H_ */
