/******************************************************************************
 * @file:		aqua_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	31. jan. 2016
 ******************************************************************************
 *	Driver for the AquaTroll multi-parameter instrument, connectable through the
 *  RS485_2 interface. The driver is however not complete, and does only
 *  implement the basic power and data handling.
 ******************************************************************************/

#include "shared.h"
#include "timer.h"
#include "err.h"

#include "aqua_driver.h"
#include "aqua_config.h"

#include "fifo_rx.h"
#include "fifo_tx.h"
#include "stream.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static bool TestConnection();
static void SetTxDirection();
static void SetRxDirection();

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool AQUA_Init(){

	LEUART_Init_TypeDef config = AQUA_UART_INIT;

	LEUART_Init(AQUA_UART, &config);

	/* Enable pins */
	GPIO_PinModeSet(AQUA_UART_PORT, AQUA_RX_PIN, gpioModeInputPullFilter, 1);
	GPIO_PinModeSet(AQUA_UART_PORT, AQUA_TX_PIN, gpioModePushPull, 1);
	GPIO_PinModeSet(AQUA_GPIO_PORT, AQUA_DE_PIN, gpioModePushPull, 0);
	GPIO_PinModeSet(AQUA_GPIO_PORT, AQUA_RE_PIN, gpioModePushPull, 1);
	GPIO_PinModeSet(AQUA_PWR_PORT, AQUA_PWR_PIN, gpioModePushPull, 0);

#if !USE_OLD_CONTROLLER
	GPIO_PinModeSet(AQUA_GPIO_PORT, AQUA_BIAS_PIN, gpioModePushPull, 1);
#endif

	/* Enable route and interrupts */

	AQUA_UART->ROUTE |= AQUA_ROUTE | AQUA_LOC ;
	AQUA_UART->IEN |= LEUART_IEN_RXDATAV;

	NVIC_EnableIRQ(AQUA_IRQn);

	return TestConnection();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool AQUA_PowerOn(){

	GPIO_PinOutSet(AQUA_PWR_PORT, AQUA_PWR_PIN);	/* Enable 12V power */

	/* Enable RS485 bias network */
#if !USE_OLD_CONTROLLER
	GPIO_PinOutSet(AQUA_GPIO_PORT, AQUA_BIAS_PIN);
#endif

	return TestConnection();

}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void AQUA_PowerOff(){

	GPIO_PinOutClear(AQUA_PWR_PORT, AQUA_PWR_PIN); 	/* Disable 12V power */

	GPIO_PinOutClear(AQUA_GPIO_PORT, AQUA_DE_PIN); 	/* Set RS485 converter in sleep */
	GPIO_PinOutSet(AQUA_GPIO_PORT, AQUA_RE_PIN);

	/* Disable RS485 bias network */
	#if !USE_OLD_CONTROLLER
	GPIO_PinOutClear(AQUA_GPIO_PORT, AQUA_BIAS_PIN);
	#endif
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TestConnection(){

	FIFO_RX_Reset(stream_aquaTroll);
	ERR_RAISE(err_connectionErr, dev_aquaTroll);
	return false;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void SetTxDirection(){

	GPIO_PinOutSet(AQUA_GPIO_PORT, AQUA_DE_PIN);
	GPIO_PinOutSet(AQUA_GPIO_PORT, AQUA_RE_PIN);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void SetRxDirection(){

	GPIO_PinOutSet(AQUA_GPIO_PORT, AQUA_DE_PIN);
	GPIO_PinOutClear(AQUA_GPIO_PORT, AQUA_RE_PIN);
}


/*******************************************************************************
 * 							Interrupt handlers
 *******************************************************************************/


void LEUART0_IRQHandler(){

	uint8_t data = 0;
	uint32_t flags = AQUA_UART->IF;
	AQUA_UART->IFC = flags;

	/* Receive interrupt */
	if (flags & LEUART_IF_RXDATAV){

		/* Clear the flag(s) */
		AQUA_UART->IFC = AQUA_UART->IF;
		uint8_t rxData = AQUA_UART->RXDATA;

		FIFO_RX_Put(stream_aquaTroll, rxData); // put data to FIFO
	}

	/* Transmit interrupt */
	if ( flags & LEUART_IF_TXBL ){

		if ( FIFO_TX_Get(stream_aquaTroll, &data)){

			AQUA_UART->TXDATA = data; // send byte
		}
		else { // no more data to transfer, FIFO empty

			/* Enables receiver to respond */
			SetRxDirection();

			/* Disable TXBL */
			AQUA_UART->IEN &= ~LEUART_IEN_TXBL;
			NVIC_ClearPendingIRQ(AQUA_IRQn);
		}
	}

}
