/******************************************************************************
 * @file:		gauge_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	30. jan. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include "shared.h"
#include "err.h"

#include "gauge_driver.h"
#include "gauge_config.h"

/* Received data when doing a full read of
 * all registers (excluding control register) */
static uint8_t registerData[44] = {0};


/*******************************************************************************
 * 						Private function declarations
 *******************************************************************************/


static bool I2CTransfer(I2C_TransferSeq_TypeDef * seq);


/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GAUGE_Init(){

	bool res = true;
	I2C_Init_TypeDef init = I2C_INIT_DEFAULT;

	I2C_Init(GAUGE_I2C, &init);

	/* Enable pins */
	GPIO_PinModeSet(GAUGE_I2C_PORT, GAUGE_SDA_PIN, gpioModeWiredAndFilter, 1);
	GPIO_PinModeSet(GAUGE_I2C_PORT, GAUGE_SCL_PIN, gpioModeWiredAndFilter, 1);
	//GPIO_PinModeSet(GAUGE_BI_PORT, GAUGE_BI_PIN, gpioModeInput, 0);

	/* Enable route */

	GAUGE_I2C->ROUTE |= GAUGE_ROUTE | GAUGE_LOC;

	/* Test connection */
 	if (!GAUGE_ReadAllRegisters()){

 		res = false;
 		GAUGE_DeInit();
 		ERR_RAISE(err_connectionErr, dev_gauge);
 	}

 	 return res;
}


/*******************************************************************************
 * @brief Disable the I2C and GPIO pins.
 *******************************************************************************/
void GAUGE_DeInit(){

	I2C_Reset(GAUGE_I2C);

	GPIO_PinModeSet(GAUGE_I2C_PORT, GAUGE_SDA_PIN, gpioModeDisabled, 0);
	GPIO_PinModeSet(GAUGE_I2C_PORT, GAUGE_SCL_PIN, gpioModeDisabled, 0);
	GPIO_PinModeSet(GAUGE_BI_PORT, GAUGE_BI_PIN, gpioModeDisabled, 0);
}


/*******************************************************************************
 * @brief Read all registers of the Gas gauge. The received data can be
 * accessed via the "Get" functions such as "GAUGE__BatteryVoltage_mV.
 *******************************************************************************/
bool GAUGE_ReadAllRegisters(){

	I2C_TransferSeq_TypeDef seq;

	uint8_t * data = GAUGE_AT_RATE;
	uint8_t * returnData = registerData;

	/* Get the data from the registers */

	seq.addr = GAUGE_ADDRESS;
	seq.buf[0].data = data; /* data to write */
	seq.buf[0].len = 2;		/* length */

	if (USE_OLD_GAS_GAUGE) seq.buf[1].len = 44; /* Data to read */
	else seq.buf[1].len = 34;

	seq.buf[1].data = returnData;	/* Buffer for read data */
	seq.flags = I2C_FLAG_WRITE_READ;

	bool transferOK = I2CTransfer(&seq);

#if !USE_OLD_GAS_GAUGE

	/* Set pointer to next position in the array */
	returnData = registerData + 34;

	/* Start reading at register "internal temp" */
	data = GAUGE_INTERNAL_TEMP;

	/* Get the last 8 bytes */

	seq.buf[0].data = data;
	seq.buf[1].len = 8;
	seq.buf[1].data = returnData;

	transferOK = I2CTransfer(&seq);
#endif

	if(!transferOK) ERR_RAISE(err_connectionErr, dev_gauge);

	return transferOK;
}


/*******************************************************************************
 * @brief Returns the two byte value of the gauge control register.
 *******************************************************************************/
uint16_t GAUGE_GetControlStatus(){

	I2C_TransferSeq_TypeDef seq;

	uint8_t * data = GAUGE_CONTROL;

	uint8_t returnData[2] = {0};

	/* Point to the "control" address in the register */
	seq.addr = GAUGE_ADDRESS;
	seq.buf[0].data = data;
	seq.buf[0].len = 2;
	seq.flags = I2C_FLAG_WRITE;

	I2CTransfer(&seq);

	/* Then point at the "control status" register and read the data */
	data = GAUGE_CONTROL_STATUS;

	seq.buf[0].data = data;
	seq.buf[1].len = 2;
	seq.buf[1].data = returnData;
	seq.flags = I2C_FLAG_WRITE_READ;

	I2CTransfer(&seq);
	return (uint16_t)(returnData[0] << 8) | returnData[1];
}


/*******************************************************************************
 * @brief Not implemented!
 *******************************************************************************/
bool GAUGE_EnableSnooze(bool enable){

	return true;
}


/*******************************************************************************
 * @brief Not implemented!
 *******************************************************************************/
bool GAUGE_EnableHibernate(bool enable){

	return true;
}


/*******************************************************************************
 * @brief Not implemented!
 *******************************************************************************/
bool GAUGE_DefaultConfig(){
	return false;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
uint16_t GAUGE_BatteryVoltage_mV(){

	return (uint16_t)(registerData[6] << 8) | registerData[7];
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
uint16_t GAUGE_RemainingCapacity_mAh(){

	return ((uint16_t)(registerData[14] << 8) | registerData[15]);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
uint16_t GAUGE_FullChargeCapacity_mAh(){

	return ((uint16_t)(registerData[16] << 8) | registerData[17]);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
int16_t GAUGE_AverageCurrent_mA(){

	return ((int16_t)(registerData[18] << 8) | registerData[19]);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
int16_t GAUGE_InstCurrent_mA(){

#if USE_OLD_GAS_GAUGE
	return 0;
#else
	return ((int16_t)(registerData[32] << 8) | registerData[33]);
#endif
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
int8_t GAUGE_InternalTemp_C(){

#if USE_OLD_GAS_GAUGE
	return 0;
#else
	float temp = (((uint16_t)(registerData[34] << 8) | registerData[35]) / 10) - 273.15;
	return temp >= 0 ? (int8_t)(temp + 0.5) : (int8_t)(temp - 0.5);
#endif
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
int16_t GAUGE_StandbyCurrent_mA(){

#if USE_OLD_GAS_GAUGE
	return ((int16_t)(registerData[24] << 8) | registerData[25]);
#else
	return ((int16_t)(registerData[22] << 8) | registerData[23]);
#endif
}


/*******************************************************************************
 * @brief Get battery state of charge (SOC) in percent (0-100).
 *******************************************************************************/
uint8_t GAUGE_StateOfCharge(){

#if USE_OLD_GAS_GAUGE
	return registerData[43] / 10;
#else
	return registerData[31] / 10;
#endif
}


/*******************************************************************************
 * 								Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool I2CTransfer(I2C_TransferSeq_TypeDef * seq){

	I2C_TransferReturn_TypeDef returnVal;

	returnVal = I2C_TransferInit(GAUGE_I2C, seq);

	/* Transfer in progress */
	while(returnVal == i2cTransferInProgress){
		returnVal = I2C_Transfer(GAUGE_I2C);
	}

	switch(returnVal){

	case i2cTransferDone:
		return true;

	default:
		return false;
	}
}


//static bool CalibrateCCOffset(){

	// Hold I2C lines low for more than 5 sek
	// Avg current lower than sleep current
	// takes approx. 16 sec
//}
