/******************************************************************************
 * @file:		gps_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	Driver for the U-blox NEO-7P GPS module.
 ******************************************************************************/

#include "timer.h"
#include "err.h"

#include "gps_driver.h"
#include "gps_config.h"
#include "spi_driver.h"
#include "ubx_parser.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static bool 	TestConnection();

static bool 	ConfigPowerManagement(GPS_OpConfig * config);

static bool 	ConfigureRXM(bool psmMode);
static bool 	ConfigureSBAS(bool enable);
static bool 	ConfigurePPP();
static void 	ConfigTXReady(bool useTxReady);

static bool 	GetNAVResult(GPS_NavRes nRes, uint8_t * buf, bool poll);

static void 	WakeUpGPS();
static void 	TransmitByte(uint8_t data);
static uint8_t 	ReceiveByte();


static bool 	usingTxReady;

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GPS_Init(){

	/* Register transmit and receive functions in UBX */
	UBX_Init(TransmitByte, ReceiveByte, true);

	/* Configure GPIO ports */

	GPIO_PinModeSet(GPS_GPIO_PORT, GPS_RESET_PIN, gpioModePushPull, 1);
	GPIO_PinModeSet(GPS_GPIO_PORT, GPS_PWR_PIN, gpioModePushPull, 1);

	return TestConnection();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GPS_UsingTxReady(){

	return usingTxReady;
}


/*******************************************************************************
 * @brief Power on GPS with implemented power switch.
 *******************************************************************************/
bool GPS_PowerOn(){

	GPIO_PinOutSet(GPS_GPIO_PORT, GPS_PWR_PIN);
	Delay_ms(1);
	return TestConnection();
}


/*******************************************************************************
 * @brief Power off GSP module with power switch.
 *******************************************************************************/
void GPS_PowerOff(){

	GPIO_PinOutClear(GPS_GPIO_PORT, GPS_PWR_PIN);
}


/*******************************************************************************
 * @brief Reset GPS through the hardware reset pin.
 *******************************************************************************/
void GPS_Reset(){

	GPIO_PinOutClear(GPS_GPIO_PORT, GPS_RESET_PIN);
	Delay_ms(10);
	GPIO_PinOutSet(GPS_GPIO_PORT, GPS_RESET_PIN);
	Delay_ms(100);
}


/*******************************************************************************
 * @brief Default configuration of the GPS message settings, UBX-MSG.
 * Activates the UBX_NAV_PVT solution to be issued if the TX ready feature
 * is enabled. All NMEA messages are de-activated.
 *******************************************************************************/
bool GPS_DefaultMsgConfig(){

	uint8_t buf[CFG_MSG_LEN];
	UBX_Message msg;
	bool resOk = true;

	msg.msgClass = UBX_CFG;
	msg.msgID = CFG_MSG;
	msg.payloadLen = CFG_MSG_C_LEN;
	msg.recvPayloadLen = 0;
	msg.poll = false;
	msg.payload = buf;

	if (usingTxReady){

		/* Set message rate for NAV PVT message */
		buf[0] = UBX_NAV;
		buf[1] = NAV_PVT;
		buf[2] = 1; /* One per navigation solution */

		resOk = UBX_TransmitMessage(&msg, NULL);
	}

	/* De-activate standard NMEA messages */

	int i = 0;

	while (resOk && GPS_STANDARD_NMEA_MSG[i] != 0xFF){

		buf[0] = GPS_STANDARD_NMEA_MSG[i++];
		buf[1] = GPS_STANDARD_NMEA_MSG[i++];
		buf[2] = 0;

		if (resOk) resOk = UBX_TransmitMessage(&msg, NULL);
	}

	return resOk;
}


/*******************************************************************************
 * @brief Sets the SPI port configuration, CFG-PRT, with the default values
 * set in the GPS config file. Set "useTxReady" to configure the chosen
 * TX ready pin to notify when data is available in the GPS buffer.
 *******************************************************************************/
bool GPS_DefaultSPIPortConfig(bool useTxReady){

	bool 		resOk = true;
	uint8_t 	buf[CFG_PRT_LEN];
	uint8_t 	pay = 4;
	UBX_Message msg = UBX_MESSAGE_POLL_PRT;


	/* Poll SPI port configuration */

	msg.payloadLen = 1;
	msg.payload = &pay; /* SPI port */

	resOk = UBX_TransmitMessage(&msg, buf);

	if (resOk){

		/* Set new SPI port configuration */
		buf[2] = (GPS_TXREADY_THRES << 7) |
				(GPS_TXREADY_PIN << 2) |
				GPS_TXREADY_POL << 1 | useTxReady;

		buf[3] = GPS_TXREADY_THRES >> 1;
		buf[12] = GPS_IN_PROTOCOLS;
		buf[14] = GPS_OUT_PROTOCOLS;
		buf[16] = GPS_EXT_TX_TIM ? 2 : 0;

		msg.payloadLen = CFG_PRT_LEN;
		msg.recvPayloadLen = 0;
		msg.payload = buf;
		msg.poll = false;

		/* Update port configuration */
		resOk = UBX_TransmitMessage(&msg, buf);

		/* Configure the TX ready pin */
		ConfigTXReady(useTxReady);
	}

	return resOk;
}


/*******************************************************************************
 * @brief Configures the default navigation engine settings, CFG-NAV5.
 *******************************************************************************/
bool GPS_DefaultNAV5Config(){

	bool resOk = true;
	UBX_Message msg = UBX_MESSAGE_POLL_NAV5;
	uint8_t buf[CFG_NAV5_LEN];

	/* Poll for current configuration */
	resOk = UBX_TransmitMessage(&msg, buf);

	if (resOk){

		/* Edit current configuration */

		buf[0] = 0x01; /* Apply dynamic model settings */
		buf[2] = 0x05; /* Dynamic model = Sea */


		/* Set new configuration */

		msg.payloadLen = CFG_NAV5_LEN;
		msg.recvPayloadLen = 0;
		msg.payload = buf;
		msg.poll = false;

		resOk = UBX_TransmitMessage(&msg, NULL);
	}

	return resOk;
}


/*******************************************************************************
 * @brief Configures the GPS operational mode with the provided GPS_OpConfig by
 * editing the CFG-PM2, CFG-RATE, CFG-SBAS, CFG-NAVX5 and CFG-RXM settings.
 * Possible power save modes to consider:
 * Set update period to 0, and wake GPS when data needed.
 * Set update period to != 0, and be awaken by GPS when data available(by enabling TX ready).
 * Use power switch to turn on GPS when data needed.
 *******************************************************************************/
bool GPS_SetOperationalMode(GPS_OpConfig * config){

	bool resOk = true;
	bool psmMode;
	bool useSbas;


	WakeUpGPS();

	resOk = ConfigPowerManagement(config);

	switch(config->mode){

	case op_continuousNoSBAS:
		useSbas = false;
		psmMode = false;
		break;

	case op_continuousSBAS:
		useSbas = true;
		psmMode = false;
		break;

	case op_ppp:
		if (resOk) resOk = ConfigurePPP();
		useSbas = true;
		psmMode = false;
		break;

	case op_psm_cyclic:
	case op_psm_onOff:
		useSbas = false;
		psmMode = true;
		break;
	}

	if (resOk) resOk = ConfigureSBAS(useSbas);
	if (resOk) resOk = ConfigureRXM(psmMode);

	return resOk;
}


/*******************************************************************************
 * @brief Configure the time pulse settings of the GPS with the provided
 * GPS_TPConfig. Not fully implemented, but will work as configured by
 * default.
 *******************************************************************************/
bool GPS_ConfigTimepulse(GPS_TPConfig * config){

	bool res = true;

	/* Enable time pulse pin interrupt */
	GPIO_PinModeSet(GPS_GPIO_PORT, GPS_TIMEPULSE_PIN, gpioModeInput, 0);
	GPIO_IntConfig(GPS_GPIO_PORT, GPS_TIMEPULSE_PIN, true, false, true);

	/* Set time pulse parameters */
	/* Disable SBAS for accuracy */
	/* CFG Rate */
	/* Antenna control settings */
	return res;
}


/*******************************************************************************
 * @brief Poll or receive the GPS PVT (Position,Velocity,Time) solution from
 * the GPS module. If not polling, the specific message is expected to be
 * available in the GPS buffer.
 *******************************************************************************/
bool GPS_GetPVTSolution(GPS_PVTSolution * sol, bool poll){

	bool resOk = true;
	uint8_t buf[NAV_PVT_LEN];

	resOk = GetNAVResult(nr_pvtSol, buf, poll);

	if (resOk){

		/* Time */
		sol->time.year 		= buf[5] << 8 | buf[4];
		sol->time.month 	= buf[6];
		sol->time.day 		= buf[7];
		sol->time.hour 		= buf[8];
		sol->time.min 		= buf[9];
		sol->time.sec 		= buf[10];
		sol->time.valid 	= buf[11] & 0x04;
		sol->time.nano 		= buf[19] << 24 | buf[18] << 16 | buf[17] << 8 | buf[16];

		/* Position */
		sol->position.fixType 					= buf[20];
		sol->position.valid 					= buf[21] & 0x01;
		sol->position.longitude 				= buf[27] << 24 | buf[26] << 16 | buf[25] << 8 | buf[24];
		sol->position.latitude 					= buf[31] << 24 | buf[30] << 16 | buf[29] << 8 | buf[28];
		sol->position.heightAboveEllipsoid_mm 	= buf[35] << 24 | buf[34] << 16 | buf[33] << 8 | buf[32];
		sol->position.heightAboveMeanSea_mm 	= buf[39] << 24 | buf[38] << 16 | buf[37] << 8 | buf[36];

		/* Velocity */
		sol->velocity.velNorth_mms 	= buf[51] << 24 | buf[50] << 16 | buf[49] << 8 | buf[48];
		sol->velocity.velEast_mms 	= buf[55] << 24 | buf[54] << 16 | buf[53] << 8 | buf[52];
		sol->velocity.velDown_mms 	= buf[59] << 24 | buf[58] << 16 | buf[57] << 8 | buf[56];
		sol->velocity.gSpeed2d_mms 	= buf[63] << 24 | buf[62] << 16 | buf[61] << 8 | buf[60];
		sol->velocity.heading_deg 	= buf[67] << 24 | buf[66] << 16 | buf[65] << 8 | buf[64];
	}
	return resOk;
}


/*******************************************************************************
 * @brief Polls the GPS module for a GPS time solution. Valid flag defines
 * whether the time is valid or not.
 *******************************************************************************/
bool GPS_GetTimeUTC(GPS_Time * utcTime){

	bool resOk = true;
	uint8_t buf[NAV_TIMEUTC_LEN];

	resOk = GetNAVResult(nr_utcTime, buf, true);

	if (resOk){
		utcTime->nano 	= buf[11] << 24 | buf[10] << 16 | buf[9] << 8 | buf[8];
		utcTime->year 	= buf[13] << 8 | buf[12];
		utcTime->month 	= buf[14];
		utcTime->day 	= buf[15];
		utcTime->hour 	= buf[16];
		utcTime->min 	= buf[17];
		utcTime->sec 	= buf[18];
		utcTime->valid 	= buf[19] & 0x04;
	}

	return resOk;
}


/*******************************************************************************
 * @brief Load permanent stored configuration from non-volatile memory.
 *******************************************************************************/
bool GPS_LoadPermanentConfig(){

	bool res;
	UBX_Message msg = UBX_MESSAGE_LOAD_PERM_CONFIG;

	res = UBX_TransmitMessage(&msg, NULL);

	return res;
}


/*******************************************************************************
 * @brief Clear permanent stored configuration in non-volatile memory, and
 * replace it with default configuration.
 *******************************************************************************/
bool GPS_ClearPermanentConfig(){

	UBX_Message msg = UBX_MESSAGE_LOAD_DEFAULT_CONFIG;

	return UBX_TransmitMessage(&msg, NULL);

}


/*******************************************************************************
 * @brief Save current configuration as new permanent configuration
 * to non-volatile memory.
 *******************************************************************************/
bool GPS_SaveNewPermanentConfig(){

	int count = 0;
	bool res;
	UBX_Message msg = UBX_MESSAGE_SAVE_CURRENT_CONFIG;


	while (!(res = UBX_TransmitMessage(&msg, NULL) && count++ < 10));

	return res;
}


/*******************************************************************************
 * 								Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Test connection to GPS module by polling for data. Allows retries
 * if buffer is not empty.
 *******************************************************************************/
static bool TestConnection(){

	UBX_Message msg = UBX_MESSAGE_POLL_ANT;
	uint8_t buf[CFG_ANT_LEN];
	bool res = true;
	int count = 0;

	/* Let the GPS buffer be emptied */
	while (!(res = UBX_TransmitMessage(&msg, buf)) && count++ < 20);

	if (!res){
		ERR_RAISE(err_connectionErr, dev_gps);
		res = false;
	}

	return res;
}


/*******************************************************************************
 * @brief Poll or receive the defined navigation data from the GPS module.
 * If not polling, the specific message is expected to be available in
 * the GPS buffer. A buffer must provided for the received data.
 *******************************************************************************/
static bool GetNAVResult(
		GPS_NavRes nRes,
		uint8_t * buf,
		bool poll){

	UBX_Message msg;
	bool resOk = true;

	switch(nRes){

	case nr_geoPosition:
		msg = (UBX_Message)UBX_MESSAGE_POLL_POSLLH;
		break;

	case nr_navStatus:
		msg = (UBX_Message)UBX_MESSAGE_POLL_NAV_STAT;
		break;

	case nr_utcTime:
		msg = (UBX_Message)UBX_MESSAGE_POLL_TIMEUTC;
		break;

	case nr_velNed:
		msg = (UBX_Message)UBX_MESSAGE_POLL_VELNED;
		break;

	case nr_pvtSol:
		msg = (UBX_Message)UBX_MESSAGE_POLL_PVT;
	}

	WakeUpGPS();

	if (poll) resOk = UBX_TransmitMessage(&msg, buf);

	else resOk = UBX_ReceiveMessage(&msg, buf);

	return resOk;
}


/*******************************************************************************
 * @brief Configures the power management settings of the GPS module with the
 * provided GPS_OpConfig, by accessing the CFG-PM2 and CFG-RATE settings.
 *******************************************************************************/
static bool ConfigPowerManagement(GPS_OpConfig * config){

	bool 		resOk = true;
	uint8_t 	buf[CFG_PM2_LEN];
	UBX_Message msg = UBX_MESSAGE_POLL_PM2;


	/* Poll power management configuration */

	resOk = UBX_TransmitMessage(&msg, buf);

	if (resOk){

		/* Edit received configuration */

		buf[4] = config->extIntAwake << 5 |
				config->extIntBackup << 6;

		buf[5] = config->waitForTimeFix << 2 |
				config->wakeToUpdateRTC << 3 |
				config->wakeToUpdateEPH << 4;

		if (config->mode == op_psm_cyclic){
			buf[6] 	= (!config->enterInactiveOnNoFix) | 0x02;
		}
		else buf[6] = !config->enterInactiveOnNoFix;

		buf[8] 	= config->updatePeriod_ms;
		buf[9] 	= config->updatePeriod_ms >> 8;
		buf[10] = config->updatePeriod_ms >> 16;
		buf[11] = config->updatePeriod_ms >> 24;

		buf[12] = config->searchPeriod_ms;
		buf[13] = config->searchPeriod_ms >> 8;
		buf[14] = config->searchPeriod_ms >> 16;
		buf[15] = config->searchPeriod_ms >> 24;

		buf[20] = config->onPeriod_s;
		buf[21] = config->onPeriod_s >> 8;

		buf[22] = 2; /* Minimum acquisition timeout in seconds */
		buf[23] = 0;

		/* Edit message for configuration */
		msg.recvPayloadLen = 0;
		msg.payloadLen = CFG_PM2_LEN;
		msg.payload = buf;
		msg.poll = false;

		/* Update power management configuration */
		resOk = UBX_TransmitMessage(&msg, NULL);
	}

	/* Poll nav/measurement rate settings */

	msg = (UBX_Message)UBX_MESSAGE_POLL_RATE;

	if (resOk) resOk = UBX_TransmitMessage(&msg, buf);

	if (resOk){

		/* Update nav/measurement rate settings */

		buf[0] = config->updatePeriod_ms;
		buf[1] = config->updatePeriod_ms >> 8;

		/* Edit message for configuration */
		msg.payloadLen = CFG_RATE_LEN;
		msg.recvPayloadLen = 0;
		msg.payload = buf;
		msg.poll = false;

		resOk = UBX_TransmitMessage(&msg, NULL);
	}

	return resOk;
}


/*******************************************************************************
 * @brief Configures the PPP (Precise Point Positioning) feature, which enables
 * positioning witch accuracy lower or equal to 1m.
 *******************************************************************************/
static bool ConfigurePPP(){

	bool 		resOk = true;
	uint8_t 	buf[CFG_NAVX5_LEN];
	UBX_Message msg = UBX_MESSAGE_POLL_NAVX5;


	/* Poll navigation engine expert configuration */

	resOk = UBX_TransmitMessage(&msg, buf);

	/* Edit configuration */

	/* Enable PPP */
	buf[26] = 1;


	/* Update new configuration */

	msg.payloadLen = CFG_NAVX5_LEN;
	msg.recvPayloadLen = 0;
	msg.payload = buf;
	msg.poll = false;

	if (resOk) resOk = UBX_TransmitMessage(&msg, NULL);

	return resOk;
}


/*******************************************************************************
 * @brief Configure the SBAS receiver subsystem such as WAAS, EGNOS, MSAS.
 *******************************************************************************/
static bool ConfigureSBAS(bool enable){

	bool 		resOk = true;
	uint8_t 	buf[CFG_SBAS_LEN];
	UBX_Message msg = UBX_MESSAGE_POLL_SBAS;


	/* Poll SBAS configuration */

	resOk = UBX_TransmitMessage(&msg, buf);

	if (resOk){

		buf[0] = enable;
		buf[1] = GPS_SBAS_USAGE;
		buf[3] = GPS_SBAS_SCANMODE >> 32;
		buf[4] = GPS_SBAS_SCANMODE;
		buf[5] = GPS_SBAS_SCANMODE >> 8;
		buf[6] = GPS_SBAS_SCANMODE >> 16;
		buf[7] = GPS_SBAS_SCANMODE >> 24;

		/* Update SBAS configuration */

		msg.payloadLen = CFG_SBAS_LEN;
		msg.recvPayloadLen = 0;
		msg.payload = buf;
		msg.poll = false;

		resOk = UBX_TransmitMessage(&msg, NULL);
	}

	return resOk;
}


/*******************************************************************************
 * @brief Configures the GPS power mode.
 *******************************************************************************/
static bool ConfigureRXM(bool psmMode){

	bool resOk;
	UBX_Message msg;
	uint8_t buf[CFG_RXM_LEN];

	buf[0] = 8; /* Always set to 8 */
	buf[1] = psmMode; /* Set PSM mode */

	msg.msgClass = UBX_CFG;
	msg.msgID = CFG_RXM;
	msg.payloadLen = CFG_RXM_LEN;
	msg.recvPayloadLen = 0;
	msg.poll = false;
	msg.payload = buf;

	resOk = UBX_TransmitMessage(&msg, NULL);

	return resOk;
}


/*******************************************************************************
 * @brief Configures the TX ready feature for the selected pin.
 *******************************************************************************/
static void ConfigTXReady(bool useTxReady){

	usingTxReady = useTxReady;

	if (useTxReady){

		GPIO_PinModeSet(GPS_GPIO_PORT, GPS_EXTINT_PIN, gpioModeInput, 0);
		GPIO_IntConfig(GPS_GPIO_PORT, GPS_EXTINT_PIN, true, true, true);
	}
	else GPIO_PinModeSet(GPS_GPIO_PORT, GPS_EXTINT_PIN, gpioModePushPull, 0);
}


/*******************************************************************************
 * @brief Send wake up signal to GPS in case it is in a sleep/low power mode.
 *******************************************************************************/
static void WakeUpGPS(){

	if (!usingTxReady){
		GPIO_PinOutSet(GPS_GPIO_PORT, GPS_EXTINT_PIN);
		Delay_ms(1);
		GPIO_PinOutClear(GPS_GPIO_PORT, GPS_EXTINT_PIN);
	}
	else TransmitByte(0xFF);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void TransmitByte(uint8_t data){

	SPI_TransferByte(data);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static uint8_t ReceiveByte(){

	return SPI_ReceiveByte();
}
