/******************************************************************************
 * @file:		gsm_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	31. jan. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include "em_int.h"

#include "shared.h"
#include "timer.h"
#include "err.h"

#include "gsm_driver.h"
#include "gsm_config.h"

#include "fifo_rx.h"
#include "fifo_tx.h"
#include "stream.h"

/*******************************************************************************
 * 							Private variable declarations
 *******************************************************************************/

static bool clearToSend = true;

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

#if (GSM_USE_RTS)
static void SetRTS(bool set);
#endif
static bool EnableTransmit();
static bool TestConnection();

static char * pinCode;

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GSM_Init(){

	USART_InitAsync_TypeDef config = GSM_UART_INIT;

	USART_InitAsync(GSM_UART, &config);

	/* Enable pins */
	GPIO_PinModeSet(GSM_UART_PORT, GSM_RX_PIN, gpioModeInputPullFilter, 1);
	GPIO_PinModeSet(GSM_UART_PORT, GSM_TX_PIN, gpioModePushPull, 1);

	GPIO_PinModeSet(GSM_PWR_PORT, GSM_PWR_PIN, gpioModePushPull, 0);
	GPIO_PinModeSet(GSM_CTS_PORT, GSM_CTS_PIN, gpioModeInput, 0);
	GPIO_PinModeSet(GSM_DTR_PORT, GSM_DTR_PIN, gpioModePushPull, 0);
	GPIO_PinModeSet(GSM_RI_PORT, GSM_RI_PIN, gpioModeInput, 1);

#if (GSM_USE_RESET)
	GPIO_PinModeSet(GSM_RESET_PORT, GSM_RESET_PIN, gpioModeWiredAnd, 1);
#endif
#if (GSM_USE_PWRKEY)
	GPIO_PinModeSet(GSM_PWRKEY_PORT, GSM_PWRKEY_PIN, gpioModeWiredAnd, 1);
#endif

#if (GSM_USE_RTS)
	GPIO_PinModeSet(GSM_RTS_PORT, GSM_RTS_PIN, gpioModeWiredAnd, 0);

	FIFO_RX_SetRtsCallBack(stream_gsm, SetRTS);
#endif

	/* Register stream flush function to call */
	STREAM_SetFlushCallback(stream_gsm, EnableTransmit);


	/* Enable route and interrupts */

	GPIO_IntConfig(GSM_CTS_PORT, GSM_CTS_PIN, true, true, true);
	GPIO_IntConfig(GSM_RI_PORT, GSM_RI_PIN, false, true, true);

	GSM_UART->ROUTE |= GSM_ROUTE | GSM_LOC ;
	GSM_UART->IEN |= UART_IEN_RXDATAV;

	NVIC_EnableIRQ(GSM_RX_IRQn);
	NVIC_EnableIRQ(GSM_TX_IRQn);

	GSM_Reset();

	return TestConnection();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GSM_PowerOn(){

#if GSM_USE_SLEEP_MODE
	GSM_LowPowerMode(false);
#elif GSM_USE_PWRKEY
	GPIO_PinOutClear(GSM_PWRKEY_PORT, GSM_PWRKEY_PIN);
	Sleep_ms(2000, true);
	GPIO_PinOutSet(GSM_PWRKEY_PORT, GSM_PWRKEY_PIN);
#else
	GPIO_PinOutClear(GSM_PWR_PORT, GSM_PWR_PIN);
#endif

	Sleep_ms(500, true);
	return TestConnection();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GSM_PowerOff(){

	bool res = true;
#if GSM_USE_SLEEP_MODE
	GSM_LowPowerMode(true);
#elif GSM_USE_PWRKEY
	GPIO_PinOutClear(GSM_PWRKEY_PORT, GSM_PWRKEY_PIN);
	Sleep_ms(2000, true);
	GPIO_PinOutSet(GSM_PWRKEY_PORT, GSM_PWRKEY_PIN);
#else
	GPIO_PinOutSet(GSM_PWR_PORT, GSM_PWR_PIN);
#endif

	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void GSM_LowPowerMode(bool enter){

	if (enter) GPIO_PinOutSet(GSM_DTR_PORT, GSM_DTR_PIN);
	else GPIO_PinOutClear(GSM_DTR_PORT, GSM_DTR_PIN);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void GSM_Reset(){

	uint8_t cancelSMS[] = {0x1B,0};
	/* Reset modem to initial state */
	STREAM_Printf(stream_gsm, "%s", cancelSMS);
	STREAM_Printf(stream_gsm, "AT+CIPSHUT\r");
	STREAM_WaitAltResponse(stream_gsm, 1000, 2, "OK", "SHUT OK");

#if GSM_USE_RESET
	GPIO_PinOutClear(GSM_RESET_PORT, GSM_RESET_PIN);
	Delay_ms(10);
	GPIO_PinOutSet(GSM_RESET_PORT, GSM_RESET_PIN);
	Sleep_ms(1800, true);
#elif GSM_USE_PWRKEY //Timing not correct
	GPIO_PinOutClear(GSM_PWRKEY_PORT, GSM_PWRKEY_PIN);
	Sleep_ms(2000, true);
	GPIO_PinOutSet(GSM_PWRKEY_PORT, GSM_PWRKEY_PIN);
#endif
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void GSM_SetCTS(bool clear){

	clearToSend = clear;
	if (clear) EnableTransmit();
	else GSM_UART->IEN &= ~UART_IEN_TXBL;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void GSM_SetPin(char * pin){

	if(pin) pinCode = pin;
}


/*******************************************************************************
 * @brief Checks GSM network registration status.
 * @return GSM_RegStatus_rs result.
 *******************************************************************************/
GSM_RegStatus_rs GSM_Registered(){

	int fRes;
	GSM_RegStatus_rs res = rs_notRegistered;

	STREAM_Printf(stream_gsm, "AT+CREG?\r");

	fRes = STREAM_WaitAltResponse(stream_gsm, 1000, 2, "+CREG:0,1", "+CREG:0,5");
	STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");

	if (fRes > 0){
		res = rs_gsmRegistered;
		fRes = STREAM_PrintfWaitResp(stream_gsm, 500, "+CGATT:1", "AT+CGATT?\r");
		STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");
	}

	if (fRes > 0 && res == rs_gsmRegistered){

		res = rs_gsmGprsRegistered;
	}

	else if (!fRes) res = rs_notRegistered;
	else res = rs_error;

	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GSM_DefaultConfig(){

	int res = 1;

	// Reset to factory settings
	//if (res) res = STREAM_PrintfWaitResp(stream_gsm, 1000, "AT&F");

	/* Disable local echo */
	if (res > 0) res = STREAM_Printf(stream_gsm, "ATE0&W\r");
	if (res > 0) res = STREAM_WaitAltResponse(stream_gsm, 1000, 2, "AT", "OK");

	/* Disable startup message */
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CIURC=0;&W\r");

	/* Set DTR function mode */
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT&D1\r");

	/* Disable mobile equipment error */
	//if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CMEE=0\r");

	// Control GPIO->turn off LED

	/********* Connection settings **********/

	/* Set fixed baud rate */
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+IPR=%d\r", GSM_BAUDRATE);

	/* Enable sleep mode functionality with DTR signal */
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CSCLK=1\r");

	/* Set hardware flow control (RTS and CTS) */
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+IFC=%d,2\r", GSM_USE_RTS ? 2:0);


	//USART_BaudrateAsyncSet(GSM_UART, 0, GSM_BAUDRATE, usartOVS16);

	if (!res) ERR_RAISE(err_dataNotValid, dev_gsm);
	return res > 0 ? true : false;
}


/*******************************************************************************
 * @brief Default SIM card configuration.
 *******************************************************************************/
bool GSM_ConfigureSmsSim(){

	/* Select phone book memory storage */
	int res = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK", "AT+CPBS=\"SM\"\r");
	//if (res) res = STREAM_PrintfWaitResp(stream_gsm, 1000, "SHUT OK", "AT+CIPSHUT");

	/********* SMS settings ************/

	/* Set SIM pin */
	if (pinCode && res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CPIN=%s\r", pinCode);

	/* Set SMS text mode */
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CMGF=1\r");
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CLIP=1\r");

	/* Select 7 bit character set */
	if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CSCS=\"GSM\"\r");

	/* Set SMS service center address */
	//if (res > 0) res = STREAM_PrintfWaitResp(stream_gsm, 500, "AT+CSCA=%s\r", smscNumber);


	if (res > 0) res = GSM_InitSMSMemory();

	/* SMS delivery */
	//if (res) res = STREAM_PrintfWaitResp(stream_gsm, 500, "OK", "AT+CNMI"

	if (!res) ERR_RAISE(err_simError, dev_gsm);
 	return res > 0 ? true : false;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool GSM_InitSMSMemory(){

	/* Delete all SMS */
	int fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK", "AT+CMGDA=\"DEL ALL\"\r");

	/* Disable messages about new SMS from the GSM module */
	if(fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK", "AT+CNMI=2,0\r");

	/* Send AT command to init memory for SMS in the SIM card
	 * fResponse:
	 * +CPMS: <usedr>,<totalr>,<usedw>,<totalw>,<useds>,<totals> */
	if (fRes > 0) fRes = STREAM_PrintfWaitResp(
			stream_gsm, 1000, "+CPMS:", "AT+CPMS=\"SM\",\"SM\",\"SM\"\r");

	STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_gsm);

	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool EnableTransmit(){

	if (clearToSend){
		/* Enable TX interrupt */
		GSM_UART->IEN |= UART_IEN_TXBL;
	}
	return clearToSend;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TestConnection(){

	int res = 0, count = 0;

	while (res < 1 && count < 3){

		STREAM_Printf(stream_gsm, "AT\r");

		res = STREAM_WaitAltResponse(stream_gsm, 100, 2, "OK", "AT");
		count++;
	}

	if (res < 1) ERR_RAISE(err_connectionErr, dev_gsm);

	return res > 0 ? true : false;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
#if (GSM_USE_RTS)
static void SetRTS(bool set){

	if (set) GPIO_PinOutClear(GSM_RTS_PORT, GSM_RTS_PIN);
	else GPIO_PinOutSet(GSM_RTS_PORT, GSM_RTS_PIN);
}
#endif


/*******************************************************************************
 * 							Interrupt handlers
 *******************************************************************************/


void UART0_TX_IRQHandler(){

	uint8_t data;
	uint32_t flags = GSM_UART->IF;
	GSM_UART->IFC = flags;

	if ( flags & UART_IF_TXBL ){

		if (FIFO_TX_Get(stream_gsm, &data) ) {

			/* Send byte from FIFO */
			GSM_UART->TXDATA = data;
		}
		/* FIFO empty */
		else {
			/* Disable TXBL */
			GSM_UART->IEN &= ~UART_IEN_TXBL;
			NVIC_ClearPendingIRQ(GSM_TX_IRQn);
		}
	}
}


void UART0_RX_IRQHandler(){

	/* Clear interrupt */
	GSM_UART->IFC = GSM_UART->IF;

	uint8_t rxData = GSM_UART->RXDATA;

	FIFO_RX_Put(stream_gsm, rxData);

#if (GSM_USE_RTS)
	if (FIFO_RX_Full(stream_gsm)){ /* Assert RTS */
		GPIO_PinOutSet(GSM_RTS_PORT, GSM_RTS_PIN);
	}
#endif
}

