/******************************************************************************
 * @file:		imu_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	IMU driver for the InvenSense MPU-9250. Driver is not complete, only
 *	implemented for test and power control.
 ******************************************************************************/

#include <math.h>

#include "shared.h"
#include "timer.h"
#include "err.h"

#include "imu_driver.h"
#include "imu_config.h"
#include "spi_driver.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/


static bool 	TestConnection();
static void 	GetHLData(uint16_t * recvData, uint8_t * reg);
static uint8_t 	ReadRegister(uint8_t reg);
static bool 	WriteRegister(uint8_t reg, uint8_t data);


/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool IMU_Init(){

	GPIO_PinModeSet(IMU_GPIO_PORT, IMU_FSYNC_PIN, gpioModePushPull, 0);
	GPIO_PinModeSet(IMU_GPIO_PORT, IMU_INT_PIN, gpioModeInputPull, 1);

	return TestConnection();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool IMU_EnterSleepMode(){

	uint8_t readVal;
	readVal = ReadRegister(IMU_PWR_MGMT_1);
	WriteRegister(IMU_PWR_MGMT_1, readVal | 0x40);

	return true;
}


/*******************************************************************************
 * @brief Not implemented!
 *******************************************************************************/
bool IMU_ExitSleepMode(){


	return true;
}


/*******************************************************************************
 * @brief Not implemented!
 *******************************************************************************/
void IMU_SelfTest(){

	SPI_Select(spi_imu);
	SPI_DeSelect(spi_imu);
}


/*******************************************************************************
 * @brief Get raw accel data.
 *******************************************************************************/
void IMU_GetAccelData(uint16_t * recvData){

	uint8_t * reg = IMU_ACCEL_PACKET;
	GetHLData(recvData, reg);
}


/*******************************************************************************
 * @brief Get temperature reading.
 *******************************************************************************/
int8_t IMU_GetTemp(){

	uint16_t recvData = 0;
	uint8_t * reg = IMU_TEMP_PACKET;

	GetHLData(&recvData, reg);

	return (int8_t)/*round*/((recvData - 21) / 333.87) + 21;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void IMU_GetGyroData(uint16_t * recvData){

	uint8_t * reg = IMU_GYRO_PACKET;
	GetHLData(recvData, reg);
}


/*******************************************************************************
 * 								Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void GetHLData(uint16_t * recvData, uint8_t * reg){

	bool endOfPacket = false;
	int i = 0;
	//uint8_t data1,data2 = 0;

	SPI_Select(spi_imu);

	while (!endOfPacket){

		SPI_TransferByte(reg[i]);

		if ((i % 2) == 0)
			recvData[i] = (uint16_t)(SPI_ReceiveByte() << 8);

		else
			recvData[i-1] |= SPI_ReceiveByte();

		i++;

		if (reg[i] == IMU_PACKET_END)
			endOfPacket = true;
	}
	SPI_DeSelect(spi_imu);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TestConnection(){

	bool testOK = true;

	SPI_Select(spi_imu);

	SPI_TransferByte(IMU_WHOAMI);

	if (SPI_ReceiveByte() != 0x71){

		testOK = false;
		ERR_RAISE(err_connectionErr, dev_imu);
	}
	SPI_DeSelect(spi_imu);
	return testOK;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static uint8_t ReadRegister(uint8_t reg){

	uint8_t readVal;

	SPI_Select(spi_imu);

	SPI_TransferByte(reg | IMU_REGISTER_READ);
	readVal = SPI_ReceiveByte();

	SPI_DeSelect(spi_imu);

	return readVal;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool WriteRegister(uint8_t reg, uint8_t data){

	bool resOk = true;
	SPI_Select(spi_imu);

	SPI_TransferByte(reg);
	SPI_TransferByte(data);

	SPI_DeSelect(spi_imu);

	return resOk;
}
