/******************************************************************************
 * @file:		radio_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	Driver for the Radiocrafts RC11xx- TinyMesh module.
 ******************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "em_int.h"

#include "shared.h"
#include "timer.h"
#include "err.h"

#include "radio_driver.h"
#include "radio_config.h"

#include "fifo_rx.h"
#include "fifo_tx.h"
#include "stream.h"

/*******************************************************************************
 * 							Private variable declarations
 *******************************************************************************/

static volatile bool inConfigMode = false;
static volatile bool promptReceived = false;
static bool inSleep = false;
static bool memReset = false;
static bool clearToSend = true;

/* Byte returned when reading configuration
 * parameters in configuration mode */
static uint8_t returnedConfigByte;

typedef enum {
	enterConfigMode,
	receivePrompt,
	enterMemConfigMode,
	sendCommand,
	exitMemConfigMode,
	exitConfigMode
} ConfigStages;

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/


static bool SendConfig(uint8_t * data, bool memConfig, bool readConfig);
static void EnterConfig();
static bool TestConnection();
static bool EnableTransmit();
static bool AutoBaud();

#if (RADIO_USE_RTS)
static void SetRTS(bool set);
#endif

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Initiates the UART interface, GPIOs and interrupts. Verifies
 * the connection with the Tinymesh module.
 *******************************************************************************/
bool RADIO_Init(){

	bool res = true;
	USART_InitAsync_TypeDef config = RADIO_UART_INIT;

	USART_InitAsync(RADIO_UART, &config);

//	/* Enable pins */
	GPIO_PinModeSet(RADIO_UART_PORT, RADIO_RX_PIN, gpioModeInputPullFilter, 1);
	GPIO_PinModeSet(RADIO_UART_PORT, RADIO_TX_PIN, gpioModePushPull, 1);
	GPIO_PinModeSet(RADIO_RESET_PORT, RADIO_RESET_PIN, gpioModePushPull, 1);
	GPIO_PinModeSet(RADIO_CONFIG_PORT, RADIO_CONFIG_PIN, gpioModePushPull, 1);

	#if RADIO_USE_RTS
		GPIO_PinModeSet(RADIO_RTS_PORT, RADIO_RTS_PIN, gpioModePushPull, 0);

		/* Register function to call when RTS can be de-asserted */
		FIFO_RX_SetRtsCallBack(stream_radio, SetRTS);
	#endif

	#if RADIO_USE_CTS
		GPIO_PinModeSet(RADIO_CTS_PORT, RADIO_CTS_PIN, gpioModeInput, 0);
		GPIO_IntConfig(RADIO_CTS_PORT, RADIO_CTS_PIN, true, true, true);

	#endif

	/* Enable route and interrupts */
	RADIO_UART->ROUTE |= RADIO_ROUTE | RADIO_LOC;
	RADIO_UART->IEN |= USART_IEN_RXDATAV;

	NVIC_EnableIRQ(RADIO_RX_IRQn);
	NVIC_EnableIRQ(RADIO_TX_IRQn);

	/* Register stream flush function to call */
	STREAM_SetFlushCallback(stream_radio, EnableTransmit);


	if (!TestConnection()){

		res = RADIO_Reset();
		if (!res) res = AutoBaud();
		if (res) ERR_RESET();
	}
	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void RADIO_SetBaudRate(int baud){

	USART_BaudrateAsyncSet(RADIO_UART, 0,
			baud, usartOVS16);
}


/*******************************************************************************
 * @brief Set the preferred operational mode for the radio: router, gateway or
 * end device.
 *******************************************************************************/
bool RADIO_SetMode(RADIO_OpMode_om mode){

	bool res;

	/* Poll for current device mode */
	uint8_t buf[] = {RADIO_MEM_DEV_TYPE, 0, RADIO_CONFIG_END};
	res = SendConfig(buf, false, true);

	switch(mode){

	case om_gateway:
		if (buf[1] != 1) res = SendConfig(RADIO_SET_GATEWAY_MODE, false, false);
		break;

	case om_router:
		if (buf[1] != 2) res = SendConfig(RADIO_SET_ROUTER_MODE, false, false);
		break;

	case om_end:
		if (buf[1] != 3) res = SendConfig(RADIO_SET_END_MODE, false, false);
		break;
	}

	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void RADIO_SetCTS(bool clear){

	clearToSend = clear;
	if (clear) EnableTransmit();
	/* Stop transmit */
	else {
		RADIO_UART->IEN &= ~UART_IEN_TXBL;
		NVIC_ClearPendingIRQ(RADIO_TX_IRQn);
	}
}


/*******************************************************************************
 * @brief Reset the radio module.
 * @return Result of testConnection
 *******************************************************************************/
bool RADIO_Reset(){

	/* Ensure that config pin is de-asserted */
	GPIO_PinOutSet(RADIO_CONFIG_PORT, RADIO_CONFIG_PIN);

	/* Send exit config command to ensure that
	 * device is not in config mode */
	RADIO_UART->TXDATA = RADIO_EXIT_CONFIG;
	Sleep_ms(10, true);
	/* Reset device */
	GPIO_PinOutClear(RADIO_RESET_PORT, RADIO_RESET_PIN);
	Sleep_ms(1, true);
	GPIO_PinOutSet(RADIO_RESET_PORT, RADIO_RESET_PIN);
	Sleep_ms(10, true);

	return TestConnection();
}


/*******************************************************************************
 * @brief Enters low power sleep mode.
 *******************************************************************************/
bool RADIO_EnterSleepMode(){

	inSleep = true;
	return SendConfig(RADIO_SLEEP, false, false);
}


/*******************************************************************************
 * @brief Wakes up device from low power sleep mode.
 *******************************************************************************/
bool RADIO_ExitSleepMode(){

	GPIO_PinOutSet(RADIO_CONFIG_PORT, RADIO_CONFIG_PIN);
	inSleep = false;

	/* Wait for the device to enter idle state */
	Sleep_ms(200, true);

	/* Check that connection is OK */
	return TestConnection();
}


/*******************************************************************************
 * @brief Set or get the unique ID of the radio module. Used to identify the
 * module in the network.
 *******************************************************************************/
bool RADIO_UniqueID(bool set, uint32_t * uid){

	bool res;
	uint8_t data[12];
	int i = 0;

	/* Copy UID message */
	while (RADIO_UID_MESSAGE[i] != RADIO_CONFIG_END){

		data[i] = RADIO_UID_MESSAGE[i];
		i++;
	}
	data[i] = RADIO_CONFIG_END;

	if (set){

		data[1] = *uid;
		data[3] = *uid >> 8;
		data[5] = *uid >> 16;
		data[7] = *uid >> 24;

		/* Set provided UID */
		res = RADIO_ConfigMem(data);
	}

	else {

		/* Read UID from radio module */
		res = SendConfig(data, false, true);

		if (res){
			*uid = data[7] << 24 | data[5] << 16 | data[3] << 8 | data[1];
		}
	}

	return res;
}


/*******************************************************************************
 * @brief Set the system ID for the radio network. All devices supposed to be
 * on the same network must have the same system ID.
 *******************************************************************************/
bool RADIO_SetSID(uint32_t * sid){

	uint8_t data[10];
	int i = 0;

	while (RADIO_SID_MESSAGE[i] != RADIO_CONFIG_END){
		data[i] = RADIO_SID_MESSAGE[i];
		i++;
	}

	data[i++] = RADIO_CONFIG_END;

	data[1] = *sid;
	data[3] = *sid >> 8;
	data[5] = *sid >> 16;
	data[7] = *sid >> 24;

	/* Set provided system ID */
	return RADIO_ConfigMem(data);
}


/*******************************************************************************
 * @brief Returns the current input signal strength. Can be used as a indication
 * of fading margin, or as a carrier sense signal to avoid collisions.
 *******************************************************************************/
float RADIO_GetRSSI_dBm(){

	if (SendConfig((uint8_t[]){RADIO_RSSI, RADIO_CONFIG_END}, false, false)){

		return (float)(-returnedConfigByte / 2);
	}
	else return 0.0;
}


/*******************************************************************************
 * @brief Returns the measured temperature in deg C.
 *******************************************************************************/
int8_t RADIO_GetTemp_C(){

	if (SendConfig((uint8_t[]){RADIO_TEMP, RADIO_CONFIG_END}, false, false)){

		return (int8_t)(returnedConfigByte - 128);
	}
	else return 0;
}


/*******************************************************************************
 * @brief Returns the measured voltage in the radio module in Volts[V].
 *******************************************************************************/
float RADIO_GetVoltage_V(){

	if (SendConfig((uint8_t[]){RADIO_VOLTAGE, RADIO_CONFIG_END}, false, false)){

		return (float)(returnedConfigByte * 0.03);
	}
	else return 0.0;
}


/*******************************************************************************
 * @brief Configures the non-volatile memory with the default configuration
 * RADIO_DEFAULT_CONFIG.
 *******************************************************************************/
bool RADIO_DefaultMemConfig(){

	bool res;

	res = RADIO_ConfigMem(RADIO_DEFAULT_CONFIG);

	if (res){

		RADIO_SetBaudRate(RADIO_BAUDRATE);
		res = RADIO_Reset();
	}

	return res;
}


/*******************************************************************************
 * @brief Configures the non-volatile memory using the provided config data.
 * If the current config is the same, the memory will not be overwritten.
 * The provided array must terminate with a RADIO_CONFIG_END char!
 *******************************************************************************/
bool RADIO_ConfigMem(uint8_t * config){

	bool res;
	bool doConfig;
	int i = 0, retryCount = 0;
	uint8_t configCopy[RADIO_MAX_CONFIG_LEN];

	while (config[i] != RADIO_CONFIG_END){
		configCopy[i] = config[i];
		i++;
	}
	configCopy[i] = RADIO_CONFIG_END;

	do {

		i = 0;
		retryCount++;
		doConfig = false;

		/* Read current radio configuration */
		res = SendConfig(configCopy, false, true);

		if (res){

			/* Compare current configuration read from radio,
			 * with provided configuration */
			while (config[i] != RADIO_CONFIG_END){

				if (configCopy[i] != config[i]){
					doConfig = true;
					break;
				}
				i++;
			}
			if (doConfig){
				/* Update current radio configuration */
				res = SendConfig(config, true, false);
			}
		}

	} while(doConfig && retryCount < 11);

	return res;
}


/*******************************************************************************
 * @brief Initiates a reset of the non-volatile memory back to factory default.
 *******************************************************************************/
bool RADIO_MemReset(){

	memReset = true;
	bool res;

	res = SendConfig(RADIO_RESET_MEM, false, false);

	memReset = false;

	if (res){
		RADIO_SetBaudRate(RADIO_BAUDRATE_FACTORY);
		res = TestConnection();
	}
	return res;
}


/*******************************************************************************
 * @brief Sets data rate[kbit/s] in non-volatile memory.
 *******************************************************************************/
bool RADIO_SetDataRate_kbps(RADIO_DataRate_dar rate){

	return RADIO_ConfigMem(((uint8_t[]){RADIO_MEM_RF_DATA_RATE,
		rate, RADIO_CONFIG_END}));
}


/*******************************************************************************
 * @brief Sets output power[dBm] in non-volatile memory.
 *******************************************************************************/
bool RADIO_SetOutputPower_dBm(RADIO_OutputPower_op power){

	return RADIO_ConfigMem((uint8_t[]){RADIO_MEM_RF_PWR,
		power, RADIO_CONFIG_END});

}


/*******************************************************************************
 * @brief Sets channel in non-volatile memory. 1-18 valid channels.
 *******************************************************************************/
bool RADIO_SetChannel(uint8_t ch){

	return RADIO_ConfigMem((uint8_t[]){RADIO_MEM_RF_CH,
		ch, RADIO_CONFIG_END});
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Enters configuration mode. The radio prompt is sent if radio enters
 * configuration mode successfully.
 *******************************************************************************/
static void EnterConfig(){

	GPIO_PinOutClear(RADIO_CONFIG_PORT, RADIO_CONFIG_PIN);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TestConnection(){

	return SendConfig(NULL, false, false);
}


/*******************************************************************************
 * @brief Autobaud feature not fully implemented. Only resets baud rate to
 * factory setting.
 *******************************************************************************/
static bool AutoBaud(){

	bool res;
	int baud = RADIO_BAUDRATE_FACTORY;

	RADIO_SetBaudRate(baud);
	res = TestConnection();
	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
#if (RADIO_USE_RTS)
static void SetRTS(bool set){

	if (set) GPIO_PinOutClear(RADIO_RTS_PORT, RADIO_RTS_PIN);
	else GPIO_PinOutSet(RADIO_RTS_PORT, RADIO_RTS_PIN);
}
#endif


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool EnableTransmit(){

	if (clearToSend){
		/* Enable TX interrupt */
		RADIO_UART->IEN |= UART_IEN_TXBL;
	}
	return clearToSend;
}


/*******************************************************************************
 * @brief Read or write to/from the radio configuration memory.
 *******************************************************************************/
static bool SendConfig(uint8_t * data, bool memConfig, bool readConfig) {

	int i = 0;
	bool res = true;

	ConfigStages stage = enterConfigMode;
	ConfigStages prevStage = enterConfigMode;
	inConfigMode = true;

	while (inConfigMode){

		switch(stage){

		/***********************************************/
		case enterConfigMode:

			EnterConfig();
			stage = receivePrompt;
			break;

		/***********************************************/
		case receivePrompt:

			StartTimer_ms(RADIO_WAIT_FOR_PROMPT);

			/* Wait till prompt is received, or timer has elapsed */
			while(!promptReceived){

				if (TimeElapsed()){

					inConfigMode = false;
					res = false;
					ERR_RAISE(err_connectionErr, dev_radio);
					goto exit;
				}
			}

			promptReceived = false;

			switch(prevStage){

			case receivePrompt:
				break;

			case enterConfigMode:
				if (memConfig) stage = enterMemConfigMode;
				else if (!data) stage = exitConfigMode;
				else stage = sendCommand;
				break;

			case enterMemConfigMode:
				stage = sendCommand;
				break;

			case sendCommand:
				stage = sendCommand;
				break;

			case exitMemConfigMode:
				stage = exitConfigMode;
				break;

			case exitConfigMode:
				inConfigMode = false;
				break;
			}

			prevStage = receivePrompt;
			break;

		/***********************************************/
		case enterMemConfigMode:

			RADIO_UART->TXDATA = RADIO_MEM_CONFIG;

			stage = receivePrompt;
			prevStage = enterMemConfigMode;
			break;

		/***********************************************/
		case sendCommand:

			if (data[i] == RADIO_CONFIG_END){

				if (memConfig) stage = exitMemConfigMode;
				else stage = exitConfigMode;
			}

			/* Read current configuration from radio */
			else if (readConfig && i % 2 == 1){
				/* Delay to ensure commands in
				 * memory config mode are processed correctly */
				Delay_ms(5);
				data[i] = returnedConfigByte;
			}

			else {

				if (readConfig){

					Delay_ms(10);
					RADIO_UART->TXDATA = RADIO_READ_MEM_CONFIG;
					stage = receivePrompt;
				}

				Delay_ms(5);

				RADIO_UART->TXDATA = data[i];
			}

			prevStage = sendCommand;
			i++;

			break;

		/***********************************************/
		case exitMemConfigMode:

			RADIO_UART->TXDATA = RADIO_MEM_EXIT;

			stage = receivePrompt;
			prevStage = exitMemConfigMode;
			break;

		/***********************************************/
		case exitConfigMode:

			if (!inSleep){
				/* De-assert config pin before sending exit command */
				GPIO_PinOutSet(RADIO_CONFIG_PORT, RADIO_CONFIG_PIN);
				RADIO_UART->TXDATA = RADIO_EXIT_CONFIG;
			}

			if (memReset) Sleep_ms(100, true);

			inConfigMode = false;

			prevStage = exitConfigMode;
			break;
		}
	}

exit:
	/* Delay to let the module enter idle after config */
	Sleep_ms(5, true);

	return res;
}


/*******************************************************************************
 * 							Interrupt handlers
 *******************************************************************************/


void USART2_TX_IRQHandler(){

	uint8_t data;
	uint32_t flags = RADIO_UART->IF;
	RADIO_UART->IFC = flags;

	if (flags & USART_IF_TXBL){

		if (FIFO_TX_Get(stream_radio, &data)){

			/* Send byte from FIFO */
			RADIO_UART->TXDATA = data;

		}/* FIFO empty */
		else {
			/* Disable TXBL */
			RADIO_UART->IEN &= ~UART_IEN_TXBL;
			NVIC_ClearPendingIRQ(RADIO_TX_IRQn);
		}
	}
}


void USART2_RX_IRQHandler(){

	/* Clear interrupt */
	RADIO_UART->IFC = RADIO_UART->IF;

	uint8_t rxData = RADIO_UART->RXDATA;

	if (inConfigMode){

		if (rxData == RADIO_PROMPT){
			promptReceived = true;
		}

		else returnedConfigByte = rxData;
	}

	else {

		FIFO_RX_Put(stream_radio, rxData);

	/* Set RTS pin if buffer full */
	#if (RADIO_USE_RTS)
		if (FIFO_RX_Full(stream_radio)){
			GPIO_PinOutSet(RADIO_RTS_PORT, RADIO_RTS_PIN);
		}
	#endif
	}
}
