/******************************************************************************
 * @file:		sd_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	31. jan. 2016
 ******************************************************************************
 *	A SD driver making a interface layer between the SPI driver and the diskio layer.
 *	The code is copied from Silicon Labs' microsd driver,
 *  with some minor adjustments and modifications.
 ******************************************************************************/

/**************************************************************************//**
 * @file
 * @brief Micro SD card driver for the EFM32GG_DK3750 development kit.
 *        This file provides the interface layer between the DK and the
 *        fat filesystem provided.
 * @version 3.20.5
 ******************************************************************************
 * @section License
 * <b>(C) Copyright 2014 Silicon Labs, http://www.silabs.com</b>
 *******************************************************************************
 *
 * This file is licensed under the Silabs License Agreement. See the file
 * "Silabs_License_Agreement.txt" for details. Before using this software for
 * any purpose, you must agree to the terms of that agreement.
 *
 ******************************************************************************/

#include "sd_driver.h"
#include "sd_config.h"

#include "spi_driver.h"
#include "spi_config.h"

/** @cond DO_NOT_INCLUDE_WITH_DOXYGEN */
static uint32_t timeOut, xfersPrMsec;


/**************************************************************************//**
 * @brief Wait for micro SD card ready.
 * @return 0xff: micro SD card ready, other value: micro SD card not ready.
 *****************************************************************************/
static uint8_t WaitReady(void)
{
  uint8_t res;
  uint32_t retryCount;

  /* Wait for ready in timeout of 500ms */
  retryCount = 500 * xfersPrMsec;
  do
    res = SD_XferSpi(0xff);
  while ((res != 0xFF) && --retryCount);

  return res;
}
/** @endcond */


void SD_Init(){

	GPIO_PinModeSet(SD_POWER_PORT, SD_POWER_PIN, gpioModePushPull, 0);
}

/**************************************************************************//**
 * @brief
 *  Do one SPI transfer.
 *
 * @param data
 *  Byte to transmit.
 *
 * @return
 *  Byte received.
 *****************************************************************************/
uint8_t SD_XferSpi(uint8_t data)
{
  if ( timeOut )
    timeOut--;

  return SPI_SendReceiveByte(data);
}
/**************************************************************************//**
 * @brief Deselect the micro SD card and release the SPI bus.
 *****************************************************************************/
void SD_DeSelect()
{
	SPI_DeSelect(spi_sdCard);
	SD_XferSpi(0xFF);
}

/**************************************************************************//**
 * @brief Select the micro SD card and wait for the card to become ready.
 * @return 1:Successful, 0:Timeout.
 *****************************************************************************/
int SD_Select(void){

	SPI_Select(spi_sdCard);
	if (WaitReady() != 0xFF){

		SPI_DeSelect(spi_sdCard);
		return 0;
	}

  return 1;
}

/**************************************************************************//**
 * @brief Turn on micro SD card power.
 *        DK doesn't support socket power control, only enable the SPI clock.
 *****************************************************************************/
void SD_PowerOn(void)
{
	GPIO_PinOutSet(SD_POWER_PORT, SD_POWER_PIN);
}

/**************************************************************************//**
 * @brief Turn off micro SD card power.
 *        DK doesn't support socket power control, only disable the SPI clock.
 *****************************************************************************/
void SD_PowerOff(void)
{
  /* Wait for micro SD card ready */
  SD_Select();
  SD_DeSelect();

  GPIO_PinOutClear(SD_POWER_PORT, SD_POWER_PIN);
}

/**************************************************************************//**
 * @brief Receive a data block from micro SD card.
 * @param[out] buff
 *  Data buffer to store received data.
 * @param btr
 *  Byte count (must be multiple of 4).
 * @return
 *  1:OK, 0:Failed.
 *****************************************************************************/
int SD_BlockRx(uint8_t *buff, uint32_t btr)
{
  uint8_t token;
  uint16_t val;
  uint32_t retryCount, framectrl, ctrl;

  /* Wait for data packet in timeout of 100ms */
  retryCount = 100 * xfersPrMsec;
  do
  {
    token = SD_XferSpi(0xff);
  } while ((token == 0xFF) && --retryCount);

  if (token != 0xFE)
  {
    /* Invalid data token */
    return 0;
  }

  /* Save current configuration. */
  framectrl = SPI_USART->FRAME;
  ctrl      = SPI_USART->CTRL;

  /* Set frame payload_len to 16 bit. This will increase the effective data rate. */

  SPI_USART->FRAME = (SPI_USART->FRAME & (~_USART_FRAME_DATABITS_MASK))
                         | USART_FRAME_DATABITS_SIXTEEN;
  SPI_USART->CTRL |= USART_CTRL_BYTESWAP;


  /* Clear send and receive buffers. */
  SPI_USART->CMD = USART_CMD_CLEARRX | USART_CMD_CLEARTX;

  if ( timeOut >= btr + 2 )
  {
    timeOut -= btr + 2;
  }
  else
  {
    timeOut = 0;
  }

  /* Pipelining - The USART has two buffers of 16 bit in both
   * directions. Make sure that at least one is in the pipe at all
   * times to maximize throughput. */
  SPI_USART->TXDOUBLE = 0xffff;
  do
  {
    SPI_USART->TXDOUBLE = 0xffff;

    while (!(SPI_USART->STATUS & USART_STATUS_RXDATAV));

    val = SPI_USART->RXDOUBLE;
    *buff++ = val;
    *buff++ = val >> 8;

    btr -= 2;
  } while (btr);

  /* Next two bytes is the CRC which we discard. */
  while (!(SPI_USART->STATUS & USART_STATUS_RXDATAV));
  SPI_USART->RXDOUBLE;

  /* Restore old settings. */
  SPI_USART->FRAME = framectrl;
  SPI_USART->CTRL  = ctrl;

  return 1;     /* Return with success */
}

/**************************************************************************//**
 * @brief Send a data block to micro SD card.
 * @param[in] buff 512 bytes data block to be transmitted.
 * @param token Data token.
 * @return 1:OK, 0:Failed.
 *****************************************************************************/
#if _READONLY == 0
int SD_BlockTx(const uint8_t *buff, uint8_t token)
{
  uint8_t resp;
  uint16_t val;
  uint32_t bc = 512;
  uint32_t framectrl, ctrl;

  if (WaitReady() != 0xFF)
  {
    return 0;
  }

  SD_XferSpi(token);         /* Xmit a token */

  if (token == 0xFD)
  {
    /* StopTran token */
    return 1;
  }

  /* Save current configuration. */
  framectrl = SPI_USART->FRAME;
  ctrl      = SPI_USART->CTRL;

  /* Set frame payload_len to 16 bit. This will increase the effective data rate. */
  SPI_USART->FRAME = (SPI_USART->FRAME & (~_USART_FRAME_DATABITS_MASK))
                         | USART_FRAME_DATABITS_SIXTEEN;
  SPI_USART->CTRL |= USART_CTRL_BYTESWAP;

  /* Clear send and receive buffers. */
  SPI_USART->CMD = USART_CMD_CLEARRX | USART_CMD_CLEARTX;

  if ( timeOut >= bc + 2 )
  {
    timeOut -= bc + 2;
  }
  else
  {
    timeOut = 0;
  }

  do
  {
    /* Transmit a 512 byte data block to the SD-Card. */

    val  = *buff++;
    val |= *buff++ << 8;
    bc  -= 2;

    while (!(SPI_USART->STATUS & USART_STATUS_TXBL));

    SPI_USART->TXDOUBLE = val;
  } while (bc);

  while (!(SPI_USART->STATUS & USART_STATUS_TXBL));

  /* Transmit two dummy CRC bytes. */
  SPI_USART->TXDOUBLE = 0xFFFF;

  while (!(SPI_USART->STATUS & USART_STATUS_TXC));

  /* Clear send and receive buffers. */
  SPI_USART->CMD = USART_CMD_CLEARRX | USART_CMD_CLEARTX;

  /* Restore old settings. */
  SPI_USART->FRAME = framectrl;
  SPI_USART->CTRL  = ctrl;

  resp = SD_XferSpi(0xff); /* Receive a data response */

  if ((resp & 0x1F) != 0x05)    /* If not accepted, return with error */
  {
    return 0;
  }

  return 1;
}
#endif  /* _READONLY */

/**************************************************************************//**
 * @brief
 *  Send a command packet to micro SD card.
 * @param[in] cmd
 *  Command byte.
 * @param[in] arg
 *  Argument.
 * @return
 *  Response value.
 *****************************************************************************/
uint8_t SD_SendCmd(uint8_t cmd, DWORD arg)
{
  uint8_t  n, res;
  uint32_t retryCount;

  if (cmd & 0x80)
  { /* ACMD<n> is the command sequense of CMD55-CMD<n> */
    cmd &= 0x7F;
    res  = SD_SendCmd(CMD55, 0);
    if (res > 1)
    {
      return res;
    }
  }

  /* Select the card and wait for ready */
  SD_DeSelect();
  if (!SD_Select())
  {
    return 0xFF;
  }

  /* Send command packet */
  SD_XferSpi(0x40 | cmd);            /* Start + Command index */
  SD_XferSpi((uint8_t)(arg >> 24));  /* Argument[31..24] */
  SD_XferSpi((uint8_t)(arg >> 16));  /* Argument[23..16] */
  SD_XferSpi((uint8_t)(arg >> 8));   /* Argument[15..8] */
  SD_XferSpi((uint8_t) arg);         /* Argument[7..0] */
  n = 0x01;                               /* Dummy CRC + Stop */
  if (cmd == CMD0)
  {
    n = 0x95;                             /* Valid CRC for CMD0(0) */
  }
  if (cmd == CMD8)
  {
    n = 0x87;                             /* Valid CRC for CMD8(0x1AA) */
  }
  SD_XferSpi(n);

  /* Receive command response */
  if (cmd == CMD12)
  {
    SD_XferSpi(0xff);                /* Skip a stuff byte when stop reading */
  }
  retryCount = 10;                        /* Wait for a valid response in timeout of 10 attempts */
  do
  {
    res = SD_XferSpi(0xff);
  } while ((res & 0x80) && --retryCount);

  return res;             /* Return with the response value */
}

/**************************************************************************//**
 * @brief Set SPI clock to a low frequency suitable for initial
 *        card initialization.
 *****************************************************************************/
void SD_SpiClkSlow(void)
{
  SPI_SetCurrentClk(SD_SPI_LO_CLOCK);
  xfersPrMsec = SD_SPI_LO_CLOCK / 8000;
}

/**************************************************************************//**
 * @brief Set SPI clock to maximum frequency.
 *****************************************************************************/
void SD_SpiClkFast(void)
{
  SPI_SetCurrentClk(SD_SPI_HI_CLOCK);
  xfersPrMsec = SD_SPI_HI_CLOCK / 8000;
}

/**************************************************************************//**
 * @brief
 *  Set a timeout value. The timeout value will be decremented towards zero
 *  when SPI traffic to/from the micro SD card takes place. Use @ref
 *  MICROSD_TimeOutElapsed() to check if timeout has elapsed.
 * @param[in] msec
 *  Millisecond timeout value (very approximate).
 *****************************************************************************/
void SD_TimeOutSet(uint32_t msec)
{
  timeOut = xfersPrMsec * msec;
}

/**************************************************************************//**
 * @brief
 *  Check if timeout value set with @ref MICROSD_TimeOutSet() has elapsed.
 * @return
 *  True if timeout has elapsed.
 *****************************************************************************/
bool SD_TimeOutElapsed(void)
{
  return timeOut == 0;
}
