/******************************************************************************
 * @file:		tbr_driver.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	31. jan. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#include <stdint.h>

#include "shared.h"
#include "timer.h"
#include "err.h"

#include "tbr_driver.h"
#include "tbr_config.h"

#include "fifo_rx.h"
#include "fifo_tx.h"
#include "stream.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static void 	SetTxDirection();
static void 	SetRxDirection();
static bool 	EnableTransmit();
static bool 	EnterCommandMode();

static char 	CalculateLuhn(time_t * time);

/*******************************************************************************
 * 							Private variable declarations
 *******************************************************************************/

static int8_t 	lastTempSensorRead = 0;
static uint16_t nTagDetections = 0;
static uint32_t lastFlashCode = 0;

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool TBR_Init(){

	LEUART_Init_TypeDef config = TBR_UART_INIT;
	LEUART_Init(TBR_UART, &config);

	/* Enable pins */
	GPIO_PinModeSet(TBR_UART_PORT, TBR_RX_PIN, gpioModeInputPullFilter, 1);
	GPIO_PinModeSet(TBR_UART_PORT, TBR_TX_PIN, gpioModePushPull, 1);
	GPIO_PinModeSet(TBR_GPIO_PORT, TBR_DE_PIN, gpioModePushPull, 0);
	GPIO_PinModeSet(TBR_GPIO_PORT, TBR_RE_PIN, gpioModePushPull, 1);
	GPIO_PinModeSet(TBR_PWR_PORT, TBR_PWR_PIN, gpioModePushPull, 0);

#if !USE_OLD_CONTROLLER
	GPIO_PinModeSet(TBR_GPIO_PORT, TBR_BIAS_PIN, gpioModePushPull, 1);
#endif

	/* Enable route and interrupts */

	TBR_UART->ROUTE |= TBR_ROUTE | TBR_LOC ;
	TBR_UART->IEN |= LEUART_IEN_RXDATAV;

	NVIC_EnableIRQ(TBR_IRQn);

	STREAM_SetFlushCallback(stream_tbr700, EnableTransmit);

	/* Make sure that the TBR is not in command mode */
	TBR_ExitCommandMode();

	return TBR_TestConnection();
}


/*******************************************************************************
 * @brief Returns true if connection is OK.
 *******************************************************************************/
bool TBR_TestConnection(){


	int fRes = STREAM_PrintfWaitResp(
			stream_tbr700, 500, "SN=", TBR_GET_SERIAL_NORMAL_MODE);

	/* Clear buffer for received serial number */
	fRes = STREAM_WaitAltResponse(stream_tbr700, 500, 2, "<><", "><>"); /*  */

	if (fRes < 1) ERR_RAISE(err_connectionErr, dev_tbr700);

	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Enables 12V power to the TBR700 and returns true if connection is OK.
 *******************************************************************************/
bool TBR_PowerOn(){

	GPIO_PinOutSet(TBR_PWR_PORT, TBR_PWR_PIN); /* Enable 12V power */

	/* Enable RS485 bias network */
#if !USE_OLD_CONTROLLER
	GPIO_PinOutSet(TBR_GPIO_PORT, TBR_BIAS_PIN);
#endif

#if (STARTUP_MSG_TBR700)
	SetRxDirection();
	Sleep_ms(4000, true); /* Wait for any "welcome message" from the TBR */
	FIFO_RX_Reset(stream_tbr700);
#endif

	return TBR_TestConnection();
}


/*******************************************************************************
 * @brief Disables 12V power to the TBR700. If "disableRs485" is set, the
 * RS485 transceiver is put in sleep mode.
 *******************************************************************************/
void TBR_DisablePower(bool disableRs485){

	GPIO_PinOutClear(TBR_PWR_PORT, TBR_PWR_PIN); 	/* Disable 12V power */

	if (disableRs485){
		GPIO_PinOutClear(TBR_GPIO_PORT, TBR_DE_PIN); /* Set RS485 converter in sleep */
		GPIO_PinOutSet(TBR_GPIO_PORT, TBR_RE_PIN);

		/* Disable RS485 bias network */
	#if !USE_OLD_CONTROLLER
		GPIO_PinOutClear(TBR_GPIO_PORT, TBR_BIAS_PIN);
	#endif
	}
}


/*******************************************************************************
 * @return true if 12V power is enabled, otherwise false.
 *******************************************************************************/
bool TBR_GetPowerState(){

	return GPIO_PinOutGet(TBR_PWR_PORT, TBR_PWR_PIN);
}


/*******************************************************************************
 * @brief Restarts the TBR700.
 *******************************************************************************/
bool TBR_Restart(){

	int fRes = STREAM_PrintfWaitResp(
			stream_tbr700, 500, TBR_RESTART, TBR_RESTART);

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_tbr700);

	return fRes >= 0 ? true : false;
}


/*******************************************************************************
 * @brief Delete all tag detections and sensor logs from flash memory.
 *******************************************************************************/
bool TBR_ClearMemory(){

	int fRes = STREAM_PrintfWaitResp(
			stream_tbr700, 500, TBR_DELETEALL, TBR_DELETEALL);

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_tbr700);

	return fRes >= 0 ? true : false;
}


/*******************************************************************************
 * @brief Return TBR to normal mode. Returns true if connection is OK.
 *******************************************************************************/
bool TBR_ExitCommandMode(){

	int res = STREAM_PrintfWaitResp(
			stream_tbr700, 500, TBR_EXIT_COMMAND, "q%s", TBR_EXIT_COMMAND);

	return res > 0 ? true : false;
}


/*******************************************************************************
 * @brief Returns the last temperature sensor measurement from the TBR700, if
 * the TBR is configured to take measurements.
 *******************************************************************************/
int8_t TBR_GetLastTempRead(){

	return lastTempSensorRead;
}


/*******************************************************************************
 * @brief Returns the number of tag detections received from the TBR700 since
 * last.
 *******************************************************************************/
uint16_t TBR_GetTagDetections(){

	uint16_t detections = nTagDetections;
	nTagDetections = 0;
	return detections;
}


/*******************************************************************************
 * @brief Returns the last read flash code from the incoming data.
 *******************************************************************************/
uint32_t * TBR_GetLastFlashCode(){

	return &lastFlashCode;
}


/*******************************************************************************
 * @brief Request a log print from the TBR700.
 *******************************************************************************/
bool TBR_RequestLog(int log, uint32_t * flashCode){


	bool res = EnterCommandMode();

	if (res){

		if (*flashCode == 0) *flashCode = 1;

		if (log) res = STREAM_Printf(
					stream_tbr700, "%s%07lu", TBR_GET_TAG_LOG, *flashCode);

		else res = STREAM_Printf(
					stream_tbr700, "%s%07lu", TBR_GET_SENSOR_LOG, *flashCode);
	}
	return res;
}


/*******************************************************************************
 * @brief Get or set TBR temperature logging interval.
 *******************************************************************************/
bool TBR_TempInt_interval(
		bool set,
		TBR_TempInt_int * interval){

	bool res;
	int fRes = -1;

	res = EnterCommandMode();

	if (res){

		if (set) fRes = STREAM_PrintfWaitResp(
				stream_tbr700, 500, TBR_SET_TEMP_LOG_INT, "%s%02d", TBR_SET_TEMP_LOG_INT, *interval);

		else fRes = STREAM_PrintfWaitResp(
				stream_tbr700, 500, TBR_SET_TEMP_LOG_INT, TBR_GET_TEMP_LOG_INT);


		if (fRes > 0) *interval = STREAM_GetValue(stream_tbr700, 100);
	}

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_tbr700);

	TBR_ExitCommandMode();
	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Sync current system time with the TBR700.
 *******************************************************************************/
bool TBR_SyncTime(time_t * newTime){

	char buf[11];
	int fRes;

	if (newTime){ /* Update current UNIX time stamp */

		int len = sprintf(buf, "%ld", *newTime); /* Format time stamp */

		buf[len - 1] = CalculateLuhn(newTime); /* Replace last digit with luhn's number */

		fRes = STREAM_PrintfWaitResp(stream_tbr700, 1000,
				USE_OLD_FW_TBR700 ? TBR_CLOCKUPDATE_OK_2 : TBR_CLOCKUPDATE_OK, "%s%s", TBR_CLOCKSYNC, buf);
	}

	/* Send clock sync sign */
	else fRes = STREAM_PrintfWaitResp(stream_tbr700, 1000,
			USE_OLD_FW_TBR700 ? TBR_CLOCKSYNC_OK_2: TBR_CLOCKSYNC_OK, TBR_CLOCKSYNC);


	if (!fRes) ERR_RAISE(err_dataNotValid, dev_tbr700);

	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool TBR_IdentifyNewTagData(uint8_t * data){

	bool found = false;

	if (*data == '$' || *data == '@'){
		nTagDetections++;
		found = true;
	}
	return found;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void TBR_IdentifyTemperatureData(uint8_t * data){

	static bool newReading;
	static int tempFound;
	static uint16_t temperatureData;

	if (*data == '$' || *data == '@'){

		newReading = true;
		tempFound = 0;
	}

	/* Check for some of the characters in a
	 * temperature sensor reading */
	else if (tempFound == 0 && *data == 'o') tempFound++;
	else if (tempFound == 1 && *data == 'r') tempFound++;
	else if (tempFound == 2 && *data == ',') tempFound++;

	/* Temperature reading identified */
	else if (newReading && tempFound > 2){

		/* Finished reading temperature */
		if (*data == ','){
			lastTempSensorRead =
					(int8_t)/*round*/((temperatureData - 50.0) / 10.0);

			newReading = false;
			temperatureData = 0;
			nTagDetections--; /* Decrement since the data was a sensor reading */
		}
		/* Convert ASCII to integer value */
		else if (*data >= '0' && *data <= '9'){
			temperatureData = 10 * temperatureData + *data - '0';
		}
	}
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void TBR_IdentifyFlashCode(uint8_t * data){

	static int commaCount = 0;

	if (*data == '$' || *data == '@'){

		commaCount = 0;
		lastFlashCode = 0;
	}

	/* Read flash code number from TBR data */
	else if (commaCount >= 7 && *data >= '0' && *data <= '9'){

		lastFlashCode = 10 * lastFlashCode + *data - '0';
	}

	else if (*data == ',') commaCount++;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool EnableTransmit(){

	SetTxDirection();
	TBR_UART->IEN |= UART_IEN_TXBL;

	return true;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void SetTxDirection(){

	GPIO_PinOutSet(TBR_GPIO_PORT,TBR_DE_PIN);
	GPIO_PinOutSet(TBR_GPIO_PORT,TBR_RE_PIN);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void SetRxDirection(){

	GPIO_PinOutClear(TBR_GPIO_PORT,TBR_DE_PIN);
	GPIO_PinOutClear(TBR_GPIO_PORT,TBR_RE_PIN);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool EnterCommandMode(){

	int fRes = STREAM_PrintfWaitResp(
			stream_tbr700, 1000, TBR_ENTER_COMMAND, TBR_ENTER_COMMAND);

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_tbr700);

	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Calculate Luhn verification number to use for time stamp updates.
 *******************************************************************************/
static char CalculateLuhn(time_t * time){

	int digit, counter;
	uint8_t timeStampArray[9];
	uint32_t luhn_sum = 0;

	time_t timestamp = *time / 10; /* Cut last digit */

	/* Make an array of digits of the time stamp */
	for (counter = 9; counter >= 1; counter--){
		digit = timestamp % 10;
		timestamp = timestamp / 10;
		timeStampArray[counter - 1] = digit;
	}

	/* Calculate luhn sum */
	for (int i = 0; i < 9; i++){

		if (i % 2 == 0){
		timeStampArray[i] += timeStampArray[i];
		}

		if (timeStampArray[i] >= 10){
			luhn_sum += timeStampArray[i] % 10 + 1;
	}

		else luhn_sum += timeStampArray[i];
	}

	return (luhn_sum * 9) % 10 + '0';
}


/*******************************************************************************
 * 							Interrupt handlers
 *******************************************************************************/


void LEUART1_IRQHandler(){

	uint8_t data = 0;
	uint32_t flags = TBR_UART->IF;

	/* Clear the flags */
	TBR_UART->IFC = flags;

	/* Receive interrupt */
	if (flags & LEUART_IF_RXDATAV){

		uint8_t rxData = TBR_UART->RXDATA;

		FIFO_RX_Put(stream_tbr700, rxData); // put data in FIFO
	}
	/* Transmit interrupt */
	if (flags & LEUART_IF_TXBL){

		if (FIFO_TX_Get(stream_tbr700, &data)){

			TBR_UART->TXDATA = data;

			/* Delay when sending commands to make sure the
			 * TBR has time to do all critical operations */
			Delay_ms(1);
		}
		else { // no more data to transfer, FIFO empty

			/* Enables receiver to respond */
			SetRxDirection();

			/* Disable TXBL */
			TBR_UART->IEN &= ~LEUART_IEN_TXBL;
			NVIC_ClearPendingIRQ(TBR_IRQn);
		}
	}
}
