/******************************************************************************
 * @file:		bmp_com.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	25. feb. 2016
 ******************************************************************************
 *	BMP (Buoy Message Protocol) communication module. Implements all of the
 *	interpretation and handling of received -and sent BMP messages.
 *	Handles all access to the log files and the system info structures,
 *	as well as the BMP parser.
 ******************************************************************************/

#ifndef BMP_COM_H_
#define BMP_COM_H_

#include "shared.h"

#include "bmp_parser.h"
#include "log_manager.h"

/******************************************************************************
 *								Settings
 ******************************************************************************/

/* Time before system expect that the polled controllers have answered */
#define BMP_DATA_RECEIVE_TIMEOUT 	5

/* System info "sysFlags" masks */

#define BMP_NAK_MASK				0x01
#define BMP_ACK_MASK				0x02
#define BMP_TF_MASK					0x04
#define BMP_RF_MASK					0x08
#define BMP_SESSION_END_MASK 		0x10
#define BMP_ACK_TIMEOUT_MASK		0x20
#define BMP_TIP_MASK				0x40
#define BMP_AWAIT_ACK_MASK			0x80


#define BMP_RADIO_POLL_SYS_INFO	\
{								\
	0,							\
	BMP_BROADCAST_ADDR, 		\
	acm_poll,					\
	bc_systemStatus,			\
	0,							\
    NULL						\
}

#define BMP_RADIO_END			\
{								\
	0,							\
	BMP_BROADCAST_ADDR, 		\
	acm_response,				\
	bc_sessionFinished,			\
	0,							\
    NULL						\
}

#define BMP_RADIO_POLL_LAST		\
{								\
	0,							\
	BMP_BROADCAST_ADDR,			\
	acm_poll,  					\
	bc_fromLast,				\
	dev_noDevice,				\
    NULL						\
}

/******************************************************************************
 *								Typedefs
 ******************************************************************************/

typedef enum {
	msg_pollClientStatus,
	msg_sessionFinished,
	msg_pollTbrDetections,
	msg_pollGaugeData,
	msg_pollGpsData,
	msg_pollAquaData,
	msg_pollWsData,
	msg_pollImuData,
	msg_nak
} BMP_MessageType_msg;


typedef void (*OnReceiveFile)(char *);
typedef void (*OnReceiveFinished)();

typedef struct {
	/* The message to transfer next */
	BMP_Message 		* msgToTransfer;

	/* The last received BMP message */
	BMP_Message 		* receivedMsg;

	/* The last sent BMP write message */
	BMP_Message			* writeMsg;

	/* The stream arguments to be used, if file transfer */
	LOG_StreamArgs 		* streamArgs;

	/* The transfer flags for each registered buoy node. Indexed by "sysIndex" */
	uint8_t 			* flags;

	/* System index of the system to send data to, or receive from. */
	uint8_t	 			sysIndex;

	/* Defines whether to include the BMP header in the transmission */
	bool				includeHeaderInMsg;

	/* Callback function when data reception has finished */
	OnReceiveFinished	receiveFinished;

	/* Callback function when file data is received */
	OnReceiveFile		fileReceived;
} BMP_TransferArgs;

/******************************************************************************
 *							Function declarations
 ******************************************************************************/

void 			BMP_Init();

BMP_Message * 	BMP_GetMessage(BMP_MessageType_msg type);

bool 			BMP_ReceiveTransmission(BMP_TransferArgs * args);
bool 			BMP_TransmitMessage(BMP_TransferArgs * args);
void 			BMP_RegisterNewMessage(BMP_TransferArgs * args);

#endif /* BMPMUNICATION_H_ */
