/******************************************************************************
 * @file:		bmp_parser.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	6. feb. 2016
 ********************************************************************************************
 *	Parser for the proprietary buoy message protocol.
 *	Radio message protocol description:												(  If file transfer	 )
 *	byte 0-2	byte 3-6	  byte 7-10	    byte 11	    byte 12	  byte 13	 byte 14-25   byte 26	byte 27(14)	byte 28(15)..n	 byte n+1
 *  SOH chars  sender addr  receiver addr  access mode  command  device log  File name	 EOFN char	 STX char     Payload	     EOT char
 ********************************************************************************************/

#ifndef BMP_PARSER_H_
#define BMP_PARSER_H_

#include "shared.h"

/******************************************************************************
 *						Protocol chars, commands etc.
 ******************************************************************************/

#define BMP_TX_FIFO_TIMEOUT_MS	1500
#define BMP_RX_FIFO_TIMEOUT_MS	500

#define BMP_BROADCAST_ADDR		0xFFFFFFFF
#define BMP_HEADER_LEN			15
#define BMP_HEADER_W_FILE_LEN	(BMP_HEADER_LEN + 12 + 1)

/* Control chars */
#define BMP_SOH_CHARS		(uint8_t[]){0xF4, 0xD0, 0x2B}
#define BMP_N_SOH_CHARS		2

#define BMP_STX				0x04 	/* Start of text */
#define BMP_EOFN			0x1C	/* End of file name (if file transmission) */
#define BMP_EOT				0x02 	/* End of transmission */

/* Access modes */
#define BMP_UPDATE			0x10 /* Contains data, but does not require ACK. */
#define BMP_POLL			0x20
#define BMP_WRITE			0x40 /* Write messages must end with a "transFinished" message, which then must be acknowledged. */
#define BMP_RESPONSE		0x80

/* Commands */
#define BMP_STATUS			0x01 | BMP_POLL | BMP_UPDATE
#define BMP_ID				0x02 | BMP_POLL | BMP_RESPONSE 	/* Not implemented */
#define BMP_CONFIG			0x03 | BMP_POLL | BMP_WRITE 	/* Not implemented */

#define BMP_SESSION_END		0x04 | BMP_RESPONSE
#define BMP_TR_FINISHED		0x05 | BMP_RESPONSE
#define BMP_ACK				0x06 | BMP_RESPONSE
#define BMP_NAK				0x07 | BMP_RESPONSE

#define BMP_LAST_ENTRY		0x08 | BMP_POLL | BMP_WRITE
#define BMP_ALL_LOG			0x09 | BMP_POLL | BMP_WRITE 	/* Not implemented */
#define BMP_LAST_LOG		0x0A | BMP_POLL | BMP_WRITE
#define BMP_SPECIFIED_LOG	0x0B | BMP_POLL | BMP_WRITE

/* Masks */
#define BMP_FILE_MASK		0x08 						/* The file mask is used to identify the BMP commands that require file access */
#define BMP_DATA_MASK		(BMP_WRITE | BMP_UPDATE) 	/* Identifies messages that contain data in addition to the header */

/*******************************************************************************
 * 								Typedefs
 *******************************************************************************/

/*******************************************************************************
 * @brief RMP message command types.
 *******************************************************************************/
typedef enum {
	bc_systemStatus = 		BMP_STATUS,			/* System info */
	bc_entireLog = 			BMP_ALL_LOG,		/* All log files */
	bc_specifiedLog =  		BMP_SPECIFIED_LOG,	/* Specified log file */
	bc_lastLogFile = 		BMP_LAST_LOG,		/* Last created log file */
	bc_fromLast = 			BMP_LAST_ENTRY,		/* Last data entry to the last log file */
	bc_config = 			BMP_CONFIG,			/* Configuration of device/unit etc. */
	bc_ack =				BMP_ACK,			/* Acknowledge to a received message */
	bc_nak =				BMP_NAK,			/* Not acknowledge to a received message */
	bc_identify = 			BMP_ID,				/* Identify the device */
	bc_transFinished = 		BMP_TR_FINISHED,	/* Notify transfer finished */
	bc_sessionFinished = 	BMP_SESSION_END		/* Notify that current session is finished */
} BMP_Command_bc;


/*******************************************************************************
 * @brief RMP message access mode types.
 *******************************************************************************/
typedef enum {
	acm_poll =		BMP_POLL, 	 	/* Poll another unit for updates, data etc. */
	acm_write = 	BMP_WRITE,	 	/* Write data, such as config data, to another unit */
	acm_response = 	BMP_RESPONSE, 	/* A response to a received message */
	acm_update = 	BMP_UPDATE		/* An update message, such as "receive finished */
} BMP_AccessMode_acm;


/*******************************************************************************
 * @brief Structure containing the information needed to transfer an RMP message.
 *******************************************************************************/
typedef struct {
	/* Sender unique ID */
	uint32_t  			senderAddr;

	/* The unique ID of the receiver to respond */
	uint32_t			receiverAddr;

	/* The access mode for this message */
	BMP_AccessMode_acm 	mode;

	/* The command defines which action to take */
	BMP_Command_bc 		cmd;

	/* Device log to access */
	Device_dev			deviceLog;

	/* Name of transferred file / file to transfer */
	char				* fileName;
} BMP_Message;


/******************************************************************************
 *							Function declarations
 ******************************************************************************/

bool 	 BMP_EndTransfer(Stream stream, uint8_t waitAfterEnd_ms);
bool 	 BMP_MakeMessageHeader(BMP_Message * msg, Stream stream);
bool 	 BMP_IdentifyMessage(BMP_Message * msg, Stream stream);
bool 	 BMP_ReceiveEOT(Stream stream);

#endif /* BMP_PARSER_H_ */
