/******************************************************************************
 * @file:		device_config.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	10. mars 2016
 ******************************************************************************
 *	Provides initial configuration of the "DEV_Settings" structures and the
 *	specific device flags. Flags can be set or removed according to the
 *	wanted operation of the buoy controller, and the use of the connected
 *  instruments and peripherals.
 ******************************************************************************/

#ifndef DEVICE_CONFIG_H_
#define DEVICE_CONFIG_H_

#include "shared.h"

#include "gauge_config.h"
#include "radio_config.h"
#include "gsm_config.h"
#include "aqua_config.h"
#include "spi_config.h"
#include "tbr_config.h"


/*******************************************************************************
 * @brief Device structure containing the different properties for each installed device.
 * All of the included data structures, flags etc. are not applicable to all
 * devices.
 *******************************************************************************/
typedef struct {

	/* Defines the time between wake up */
	uint32_t interval_s;

	/* Defines the active time in seconds */
	uint16_t onPeriod_s;

	/********************************************************************************************
	 * Device flags.
	 * Bit explanation:
	 * 	bit7		bit6		bit5		bit4		bit3		bit2		bit1		bit0
	 * 	shutdown  awaitData	 useAwaitData  active  noDeepSleep  collectData   alwaysOn	  enabled
	 *
	 * Explanations:
	 * enabled:     Asserted when device is initially enabled, and available (working).
	 * alwaysOn:	Assert to configure device to be always be powered on (cannot be
	 * 				shut off / set to sleep).
	 * collectData:	Asserted when device provides data / is used for data acquisition.
	 * noDeepSleep: Asserted when device does not allow deep sleep operation when active(on).
	 * active:		Asserted when the device is active, meaning it is powered on.
	 * useAwaitData:Assert to enable awaitData for device.
	 * awaitData:	Asserted when device is expected to provide data before any shutdown of the
	 * 				device.
	 * shutdown:	Asserted when shutdown of device has been attempted, but failed because of
	 * 				"awaitData" flag set.
	 ********************************************************************************************/
	uint8_t  flags;

	/* The associated clock */
	const CMU_Clock_TypeDef clock;

	bool (*Init)();

	bool (*WakeUp)();

	bool (*Configure)();

	bool (*CollectData)();

	bool (*Reset)();
} DEV_Settings;


/*** Masks ***/
#define DEV_ENABLED_MASK		0x01
#define DEV_ALWAYS_ON_MASK		0x02
#define DEV_COLLECT_DATA_MASK 	0x04
#define DEV_NO_DEEP_SLEEP_MASK	0x08
#define DEV_ACTIVE_MASK			0x10
#define DEV_USE_AWAIT_DATA_MASK	0x20
#define DEV_AWAIT_DATA_MASK		0x40
#define DEV_SHDWN_MASK			0x80

/******************************************************************************
*							Initial device flags
******************************************************************************/


#define DEV_GPS_COLLECT_DATA 	GPS_USE_TX_READY ? DEV_ENABLED_MASK : DEV_COLLECT_DATA_MASK


#define DEV_RADIO_FLAGS_INIT	DEV_ENABLED_MASK 						 | DEV_NO_DEEP_SLEEP_MASK
#define DEV_GSM_FLAGS_INIT		DEV_ENABLED_MASK 						 | DEV_NO_DEEP_SLEEP_MASK
#define DEV_SD_FLAGS_INIT		DEV_ENABLED_MASK

#define DEV_TBR_FLAGS_INIT		DEV_ENABLED_MASK | DEV_COLLECT_DATA_MASK | DEV_NO_DEEP_SLEEP_MASK | DEV_ALWAYS_ON_MASK
#define DEV_GPS_FLAGS_INIT		DEV_ENABLED_MASK | DEV_GPS_COLLECT_DATA												   | DEV_USE_AWAIT_DATA_MASK
#define DEV_GAUGE_FLAGS_INIT	DEV_ENABLED_MASK | DEV_COLLECT_DATA_MASK

#define DEV_AQUA_FLAGS_INIT		/*DEV_ENABLED_MASK |*/DEV_COLLECT_DATA_MASK
#define DEV_IMU_FLAGS_INIT		/*DEV_ENABLED_MASK |*/DEV_COLLECT_DATA_MASK
#define DEV_WS_FLAGS_INIT		/*DEV_ENABLED_MASK |*/DEV_COLLECT_DATA_MASK


/******************************************************************************
*					Misc. power-related ports and pins
******************************************************************************/

#define CH_FAULT_PORT	gpioPortA	/* Charger output, active low (LTC4121-4.2) */
#define CH_CHRG_PORT	gpioPortE	/* Charger output, active low (LTC4121-4.2) */
#define BI_TOUT_PORT	gpioPortE
#define V12_PWR_PORT	gpioPortE	/* 12V enable, active high */
#define LED_PORT		gpioPortD

#define CH_FAULT_PIN	15
#define CH_CHRG_PIN 	14
#define BI_TOUT_PIN 	15
#define V12_PWR_PIN		1
#define LED_GR_PIN     	2
#define LED_YL_PIN     	3
#define LED_RED_PIN    	4


/*******************************************************************************
 * 						Initially defined device settings
 *******************************************************************************/


#define INIT_DEVICE_RADIO									\
{															\
	100, 					/* Interval time in seconds */	\
	80, 					/* On time in seconds */		\
	DEV_RADIO_FLAGS_INIT, 	/* Init flags */				\
	RADIO_CMUCLOCK 			/* Associated CMU clock */		\
}

#define	INIT_DEVICE_TBR										\
{															\
	0, 						/* Interval time in seconds */	\
	0, 						/* On time in seconds */		\
	DEV_TBR_FLAGS_INIT, 	/* Init flags */				\
	TBR_CMUCLOCK 			/* Associated CMU clock */		\
}

#define INIT_DEVICE_AQUA									\
{															\
	100, 					/* Interval time in seconds */	\
	0, 						/* On time in seconds */		\
	DEV_AQUA_FLAGS_INIT, 	/* Init flags */				\
	AQUA_CMUCLOCK 			/* Associated CMU clock */		\
}

#define INIT_DEVICE_GSM										\
{															\
	0, 						/* Interval time in seconds */	\
	0,	 					/* On time in seconds */		\
	DEV_GSM_FLAGS_INIT, 	/* Init flags */				\
	GSM_CMUCLOCK, 			/* Associated CMU clock */		\
}

#define INIT_DEVICE_WS										\
{															\
	0,	 					/* Interval time in seconds */	\
	0, 						/* On time in seconds */		\
	DEV_WS_FLAGS_INIT, 		/* Init flags */				\
	/*WS_CMUCLOCK*/ 		/* Associated CMU clock */		\
}

#define INIT_DEVICE_GPS										\
{															\
	300,					/* Interval time in seconds */	\
	60, 					/* On time in seconds */		\
	DEV_GPS_FLAGS_INIT, 	/* Init flags */				\
	SPI_CMUCLOCK, 			/* Associated CMU clock */		\
}

#define INIT_DEVICE_SD										\
{															\
	0, 						/* Interval time in seconds */	\
	0, 						/* On time in seconds */		\
	DEV_SD_FLAGS_INIT, 		/* Init flags */				\
	SPI_CMUCLOCK, 			/* Associated CMU clock */		\
}

#define INIT_DEVICE_IMU										\
{															\
	100, 					/* Interval time in seconds */	\
	0, 						/* On time in seconds */		\
	DEV_IMU_FLAGS_INIT, 	/* Init flags */				\
	SPI_CMUCLOCK, 			/* Associated CMU clock */		\
}

#define INIT_DEVICE_GAUGE									\
{															\
	100,					/* Interval time in seconds */	\
	0, 						/* On time in seconds */		\
	DEV_GAUGE_FLAGS_INIT, 	/* Init flags */				\
	GAUGE_CMUCLOCK,			/* Associated CMU clock */		\
}


#endif /* DEVICE_CONFIG_H_ */
