/******************************************************************************
 * @file:		file.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	The file module act as a interface to the FatFs module. Basic functionality
 *	as changing directories, creating, opening and closing of files is
 *	implemented. In addition, the module enables several files to be open
 *	simultaneously, by storing the file information in static FIL structures,
 *	and the file names in static char arrays.
 *	This allows for fast access to files when reading and writing from / to files.
 *	The module also enables stream functionality to -and from the files.
 *
 *	The current implementation depends on all file names being unique, even if
 *	they are placed in different directories. The file names must also be kept
 *	less than 13 characters, as specified in the FatFs module.
 *
 *	An error handler is included in the module, which transform the FRESULT
 *	enumeration to the ERR_Result_err defined in err.h.
 ******************************************************************************/

#ifndef FILE_H_
#define FILE_H_

#include "ff.h"
#include "diskio.h"
#include "integer.h"

#include "shared.h"

#define MAX_OPEN_FILES	N_MAX_SYSTEMS + 1

#define FILE_STREAM_BUF_LEN			100
#define FILE_STREAM_RX_TIMEOUT_MS  	500

typedef uint (* FILE_InStreamFunc)(uint8_t * data, uint sense);
typedef uint (* FILE_OutStreamFunc)(const uint8_t * data, uint sense);

/*******************************************************************************
 * 							Function declarations
 *******************************************************************************/

bool 		FILE_InitFileSystem();
int16_t 	FILE_GetFree_MB();
uint8_t 	FILE_GetNOpenFiles();

FIL * 		FILE_OpenNew(const char * fName);
FIL * 		FILE_OpenExisting(const char * fName);
FIL * 		FILE_OpenAppend(const char * fName);
bool 		FILE_Close(const char * fileName);
bool 		FILE_CloseAll();
bool 		FILE_Delete(const char * fName);

bool 		FILE_OpenDir(DIR * dirObj, const char * dirName);
bool		FILE_ChangeDir(const char * dirName);
bool 		FILE_CreateDir(const char * dirName);
bool 		FILE_CloseDir(DIR * dirObj);

char * 		FILE_FindString(FIL * fObj, const char * string, char * buf, int bufLen, uint32_t * filePointer);
bool 		FILE_MoveToPosition(FIL * fObj, uint32_t index);

bool 		FILE_TransferToStream(FIL * fObj, FILE_OutStreamFunc streamFunc, uint nBytesToRead);
bool 		FILE_WriteFromStream(FIL * fObj, FILE_InStreamFunc streamFunc);

bool 		FILE_Write(FIL *fp, const char * buffer, uint nBytesToWrite);
bool 		FILE_Read(FIL *fp, char * buffer, uint nBytesToRead);
bool 		FILE_ReadString(FIL *fObj, char * buffer, int bufferLen);


#endif /* FILE_H_ */
