/******************************************************************************
 * @file:		shared.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	4. feb. 2016
 ******************************************************************************
 *	Defines most of the system settings, and several enumerations and macros
 *	used in most modules.
 ******************************************************************************/

#ifndef SHARED_H_
#define SHARED_H_

#include <stdint.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "em_cmu.h"

/*******************************************************************************
 * 								Settings
 *******************************************************************************/

extern bool firstTimeOnly; 					/* For first time configuration */


#define DEBUG_OPERATION				0 		/* 0 = No debug, normal operation.
								   	   	   	 * 1 = Debug with GPS active/in range (time sync from GPS).
								   	   	   	 * 2 = Debug without GPS active/in range.
								   	   	   	 *
								   	   	   	 * Note that debug operation enables SWO output, and different debug strings will be output on the SWO line.
								   	   	     * These will add delays to the operation of the buoy controller, which may delay or corrupt operations that
								   	   	     * are time dependent. */

#define USE_OLD_CONTROLLER			1		/* Set to 1 for the old buoy controller prototypes. Some pins have been changed in the new revision */

#define USE_OLD_GAS_GAUGE			1		/* Set to 1 if the older BQ27510-G2 model is used (new = G3)*/

#define USE_OLD_FW_TBR700			1		/* The TBR700 answers with different commands depending on firmware. Check the expected responses,
										     * otherwise may the TBR700 be considered faulty. */

#define STARTUP_MSG_TBR700			0 		/* Set to 1 if TBR700 sends startup messages (only when not using battery (applying power)?) */

#define USE_PROTOCOL_GSM			0 		/* 0 = Use defined GSM commands to access data via GSM/GPRS. No header is added to the data transferred.
											 * 1 = Use the proprietary BMP protocol for all data transfer via GSM/GPRS */

#define UPDATE_SYSTEM_TIME_RATE 	2 		/* 1: Update every time a valid solution is received from the GPS
								   	   	     * 2: Update every second time, and so on */

#define UPDATE_TBR_CLOCK_RATE		2		/* The rate at which the full time stamp is sent to the TBR700. Otherwise the (+) clock sync is sent.
 	 	 	 	 	 	 	 	 	 	 	 * The rate is also dependent on the fact that TBR700 time is only updated on last digit == 0. */

#define SYNC_SCHEDULER_RATE			5 		/* Synchronization rate of scheduler to UTC and midnight time.
											 * Multiply with "UPDATE_SYSTEM_TIME_RATE" to get actual rate. */

#define SAVE_GPS_POS_RATE			2  		/* GPS position save rate, meaning the rate at which a valid GPS position is saved to log file */

#define SAVE_TBR_DET_RATE			10		/* The rate to save the number of tag detections and last TBR flash code to file. */


#if !USE_PROTOCOL_GSM
#define STANDARD_UID_GSM			255 	/* Use a fixed UID for the GSM module for identification in the bmp_com module,
 	 	 	 	 	 	 	 	 	 	 	 * when not using the BMP protocol for data transfer via the module */
#endif


#define N_DEVICES					9		/* Total number of devices */
#define N_STREAM_DEVICES			5		/* Number of devices that use FIFO queues(UART devices) */
#define N_LOG_DEVICES				6		/* Number of devices that the system collects and logs data for */
#define N_MAX_SYSTEMS				10		/* Max systems to operate in one network (value selected randomly)*/

#define FILENAME_MAX_SIZE			13		/* SFN (FAT short file name) */

#define TIMEOUT_GPS_FIX				60 		/* Timeout to acquire GPS position fix in seconds */


#define BAT_VOLTAGE_LOW_THRES		3000	/* Low voltage threshold in mV */
#define BAT_SOC_LOW_THRES			10		/* Low state of charge threshold in % */

#define SYS_LASTCONTACT_THRES		3600 * 24	/* Last contact threshold in seconds */
#define SYS_BOARD_TEMP_THRES		50 			/* On board temperature threshold in degC */



/* Clock settings */
#define HFRCO_BAND					cmuHFRCOBand_28MHz
#define HFPER_DIV					cmuClkDiv_1
#define HFCORE_DIV					cmuClkDiv_1

/* RTC settings */
#define RTC_CLK_DIV					cmuClkDiv_1
#define RTC_TICKS_PR_SEC			32768 / RTC_CLK_DIV 	/* clkFreq / cmuClkDiv */
#define RTC_LEN						0x00FFFFFF

/* LE timer settings */
#define LETIMER_CLK_DIV				cmuClkDiv_32768
#define LETIMER_TICKS_PR_SEC		32768 / LETIMER_CLK_DIV
#define LETIMER_LEN					0xFFFF


#define TIMER_MULTI					28 		/* Used for the timers to get approx. 1ms resolution
		 	 	 	 	 	 	 	 	 	 * Approx. the same as HFRCO band(MHz) / HFPER_DIV */

/*******************************************************************************
 * 								Flag masks
 *******************************************************************************/


/*******************************************************************************
 * 								Typedefs
 *******************************************************************************/

typedef unsigned int uint;

/*******************************************************************************
 * @brief Enum defining all of the accessible devices (if installed).
 * Is used for accessing logs as well as the devices. The devices that are used
 * to collect data (log devices) must be at the top (such as GPS, gauge).
 *******************************************************************************/
typedef enum {
	dev_tbr700 = 0,
	dev_aquaTroll,
	dev_ws,
	dev_gps,
	dev_imu,
	dev_gauge,
	dev_sdCard,
	dev_radio,
	dev_gsm,
	dev_noDevice /* Keep at the end! */
} Device_dev;


/*******************************************************************************
 * @brief Enum defining all of stream devices, which are the devices that are
 * connected through the UART interfaces and require buffers for outgoing -and
 * incoming data.
 *******************************************************************************/
typedef enum {
	stream_radio = 0,
	stream_tbr700,
	stream_aquaTroll,
	stream_ws,
	stream_gsm,
	stream_debug,
	stream_none /* Keep at the end! */
} Stream;


typedef struct {
	time_t 	atTime;						/* The time at which the position was acquired */
	int32_t longitude;					/* Longitude degrees, 1e-7 scale */
	int32_t latitude;					/* Latitude degrees, 1e-7 scale */
	int32_t heightAboveMeanSea_mm;		/* Height in mm */
	int32_t heightAboveEllipsoid_mm;	/* Height above the ellipsoid in mm */
	uint8_t	fixType;					/* Fix type: 2: 2D-fix, 3: 3D-fix + more */
	bool	valid;						/* Defines whether the position is within the accuracy masks */
} GeoPosition;


/******************************************************************************
 *								Macros
 ******************************************************************************/

#define STREAM_TO_DEV(stream) 						\
	stream == stream_radio ? dev_radio : 			\
	stream == stream_tbr700 ? dev_tbr700 : 			\
	stream == stream_aquaTroll ? dev_aquaTroll :	\
	stream == stream_ws ? dev_ws :					\
	stream == stream_gsm ? dev_gsm : dev_noDevice	\

#define DEV_TO_STREAM(dev) 						\
	dev == dev_tbr700 ? stream_tbr700 :			\
	dev == dev_aquaTroll ? stream_aquaTroll :	\
	dev == dev_ws ? stream_ws :					\
	dev == dev_radio ? stream_radio :			\
	dev == dev_gsm ? stream_gsm : stream_none

#define BYTE_TO_BINARY_PATTERN "%d%d%d%d%d%d%d%d"

#define BYTE_TO_BINARY(byte)  \
  (byte & 0x80 ? 1 : 0), \
  (byte & 0x40 ? 1 : 0), \
  (byte & 0x20 ? 1 : 0), \
  (byte & 0x10 ? 1 : 0), \
  (byte & 0x08 ? 1 : 0), \
  (byte & 0x04 ? 1 : 0), \
  (byte & 0x02 ? 1 : 0), \
  (byte & 0x01 ? 1 : 0)


#define DEVICE_TO_STRING(device) 				\
	device == dev_tbr700 ? "TBR700" :			\
	device == dev_aquaTroll ? "AquaTroll" :		\
	device == dev_ws ? "Weather station" :		\
	device == dev_gps ? "GPS" :					\
	device == dev_imu ? "IMU" :					\
	device == dev_gauge ? "Gas gauge" :			\
	device == dev_sdCard ? "SD card" :			\
	device == dev_radio ? "Radio" :				\
	device == dev_gsm ? "GSM" : "Unknown"   	\

#endif /* SHARED_CONFIG_H_ */
