/******************************************************************************
 * @file:		system.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	24. feb. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#ifndef SYSTEM_H_
#define SYSTEM_H_

#include "shared.h"


#define SYS_INFO_LEN 			36
#define SYS_INFO_FORMAT_LEN		150

#define SYS_DATA_ACQ_RETRIES	6
#define SYS_DATA_ACQ_TIMEOUT_S	10

/* System info "flags" masks */

#define SYS_CHARGING_MASK		0x01
#define SYS_CHRGERR_MASK		0x02
#define SYS_FAULTY_DEV_MASK		0x04
#define SYS_NOCONTACT_MASK		0x08
#define SYS_HIGHTEMP_MASK		0x10
#define SYS_LOWVOLTAGE_MASK		0x20
#define SYS_LOWSOC_MASK			0x40


/* System flags masks */

#define SYS_NEW_TBR_DET_MASK		0x01
#define SYS_DATA_ACQ_FINISHED_MASK	0x02
#define SYS_START_DATA_ACQ_MASK		0x04
#define SYS_END_ACQ_SESSION_MASK	0x08

/*******************************************************************************
 * 								Typedefs
 *******************************************************************************/


typedef enum {
	task_ok = 0,
	task_startDataAcq,
	task_pollTbrData,
	task_sysParamAboveTH,
	task_pollGpsData,
	task_pollGaugeData,
	task_pollAquaData,
	task_pollWsData,
	task_pollImuData,
	task_endDataAcqSession
} SYS_Task;


/*******************************************************************************
 * @brief Struct containing the most essential information about the controller,
 * such as gateway/slave status, battery status, etc.
 *******************************************************************************/
typedef struct {
	/* The unique id of the system (Radio UID) */
	uint32_t id;

	/* Defines whether the controller is a network gateway, or router */
	bool gateway;

	/* A UNIX time stamp which is used by the gateway system to keep control of
	 * the last time it had contact with the specified router system */
	time_t lastContact;

	/* Latest registered GPS position */
	GeoPosition lastPosition;


	/*******************************************************************************
	 * @brief Variable containing the controller status flags.
	 * Bit explanation:
	 *    bit7	    bit6	 bit5    bit4	  bit3	    bit2	    bit1	   bit0
	 *   		   lowSOC  lowVolt highTemp noContact faultyDev(s) chargerErr  charging
	 *******************************************************************************/
	uint8_t flags;

	/* Provides the last measured battery voltage */
	uint16_t battVoltage_mV;

	/* Provides the average current.. */
	int16_t avgCurrent_mA;

	/* Provides the current state of charge in percent */
	uint8_t stateOfCharge;

	/* Provides the temperature from the built-in temperature sensor in the MCU. */
	int8_t onBoardTemp_degC;

	/* Provides the sea water temperature last measured by the TBR700. */
	int8_t seaTemp_degC;

	/* Provides a bit field of the enabled devices, where bit 0 corresponds to
	 * the first device in the device list, and so on.. */
	uint16_t enabledDevices;

	/* A bit field of the devices, with which no contact was a established */
	uint16_t faultyDevices;

	/* The total number of tag detections from the TBR700 acoustic receiver */
	uint32_t tbrTagDetections;

} SystemInfo;


/******************************************************************************
 *							Function declarations
 ******************************************************************************/

void 			SYS_InitSystemClocks();
void 			SYS_SetupSWOForPrint();

SYS_Task		SYS_GetTask();
void 			SYS_StartDataAcquisition();
void 			SYS_DataAcqFinished();
void 			SYS_AnalyzeSystemInfo();

bool 			SYS_GetFaultyDevice(Device_dev dev);
bool 			SYS_GetEnabledDevice(Device_dev dev);

uint32_t * 		SYS_GetID();
uint32_t *		SYS_GetGateway();
uint8_t 		SYS_IdentifySystem(uint32_t addr);
uint32_t *		SYS_IdentifyIndex(uint8_t index);
uint8_t			SYS_GetnRegSystems();
void	 		SYS_PrintSystemInfo(Stream stream, uint8_t systemIndex, bool format);
bool			SYS_PrintShortSysInfo(Stream stream);


void 			SYS_SetFlags(uint8_t mask);
void 			SYS_ClearFlags(uint8_t mask);

void			SYS_SetID(uint32_t * uid);
void 			SYS_SetEnabledDevice(Device_dev dev);
void 			SYS_SetFaultyDevice(Device_dev dev);


void 			SYS_UpdateGeoPos(GeoPosition * pos);
void			SYS_UpdateLastContact(uint8_t systemIndex);
void 			SYS_UpdateOnBoardTemp(int8_t temp);
void			SYS_UpdateSeaTemp(int8_t temp);
void 			SYS_UpdatePowerInfo(uint16_t battVoltage_mV, int16_t avgCurrent_mA, uint8_t stateOfCharge);
bool 			SYS_UpdateSystemInfo(Stream stream, uint8_t systemIndex);

void 			SYS_AddTbrDetections(uint32_t nRec);
uint32_t *  	SYS_TbrDetections(uint32_t * detections);
uint32_t * 		SYS_TbrFlashCode(uint32_t * flashCode);

#endif /* SYSTEM_H_ */
