/******************************************************************************
 * @file:		ubx_parser.h
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/


#ifndef UBX_PARSER_H_
#define UBX_PARSER_H_

#include "spi_driver.h"
#include "timer.h"
#include "shared.h"

/*******************************************************************************
 * 								Typedefs
 *******************************************************************************/

typedef struct {
	uint8_t   	msgClass;
	uint8_t   	msgID;
	uint16_t  	recvPayloadLen;	/* Length of the payload to be received */
	uint16_t  	payloadLen;		/* Length of the payload */
	bool	  	poll;			/* Set if the message is a poll message */
	uint8_t * 	payload;		/* Pointer to the message to be sent */
} UBX_Message;

/*******************************************************************************
 * 							Function declarations
 *******************************************************************************/

void UBX_Init(void (*TransmitBytePtr)(uint8_t), uint8_t (*ReceiveBytePtr)(), bool spiMode);
bool UBX_TransmitMessage(UBX_Message * msg, uint8_t * recvPayload);
bool UBX_ReceiveMessage(UBX_Message * msg, uint8_t * recvPayload);

/*******************************************************************************
 * 							UBX protocol message classes -and IDs
 *******************************************************************************/

#define UBX_RECEIVE_TIMEOUT_MS	5

// UBX Protocol Sync chars

#define UBX_SYNC_CHAR_1			0xB5
#define UBX_SYNC_CHAR_2			0x62


// UBX Message classes

#define UBX_NAV					0x01
#define UBX_RXM					0x02
#define UBX_INF					0x04
#define UBX_ACK					0x05
#define UBX_CFG					0x06
#define UBX_MON					0x0A
#define UBX_AID					0x0B
#define UBX_TIM					0x0D
#define UBX_LOG					0x21

// UBX Message IDs

#define ACK_ACK					0x01
#define ACK_NAK					0x00

#define AID_ALPSRV				0x32
#define AID_ALP					0x50
#define AID_AOP					0x33
#define AID_DATA				0x10
#define AID_EPH					0x31
#define AID_HUI					0x02
#define AID_INI					0x01
#define AID_REQ					0x00

#define CFG_ANT					0x13
#define CFG_CFG					0x09
#define CFG_DAT					0x06
#define CFG_GNSS				0x3E
#define CFG_INF					0x02
#define CFG_ITFM				0x39
#define CFG_LOGFILTER			0x47
#define CFG_MSG					0x01
#define CFG_NAV5				0x24
#define CFG_NAVX5				0x23
#define CFG_NMEA				0x17
#define CFG_NVS					0x22
#define CFG_PM2					0x3B
#define CFG_PRT					0x00
#define CFG_RATE				0x08
#define CFG_RINV				0x34
#define CFG_RST					0x04
#define CFG_RXM					0x11
#define CFG_SBAS				0x16
#define CFG_TP5					0x31
#define CFG_USB					0x1B

#define INF_DEBUG				0x04
#define INF_ERROR				0x00
#define INF_NOTICE				0x02
#define INF_TEST				0x03
#define INF_WARNING				0x01

#define LOG_CREATE				0x07
#define LOG_ERASE				0x03
#define LOG_FINDTIME			0x0E
#define LOG_INFO				0x08
#define LOG_RETRIEVEPOS			0x0B
#define LOG_RETRIEVESTRING		0x0D
#define LOG_RETRIEVE			0x09
#define LOG_STRING				0x04

#define MON_HW2					0x0B
#define MON_HW					0x09
#define MON_IO					0x02
#define MON_MSGPP				0x06
#define MON_RXBUF				0x07
#define MON_RXR					0x21
#define MON_TXBUF				0x08
#define MON_VER					0x04

#define NAV_AOPSTATUS			0x60
#define NAV_CLOCK				0x22
#define NAV_DGPS				0x31
#define NAV_DOP					0x04
#define NAV_POSECEF				0x01
#define NAV_POSLLH				0x02
#define NAV_PVT					0x07
#define NAV_SBAS				0x32
#define NAV_SOL					0x06
#define NAV_STATUS				0x03
#define NAV_SVINFO				0x30
#define NAV_TIMEGPS				0x20
#define NAV_TIMEUTC				0x21
#define NAV_VELECEF				0x11
#define NAV_VELNED				0x12

#define RXM_ALM					0x30
#define RXM_EPH					0x31
#define RXM_PMREQ				0x41
#define RXM_RAW					0x10
#define RXM_SFRB				0x11
#define RXM_SVSI				0x20

#define TIM_TM2					0x03
#define TIM_TP					0x01
#define TIM_VRFY				0x06


/*******************************************************************************
 * 							Standard UBX message lengths
 *******************************************************************************/

#define ACK_LEN					2
#define CFG_CFG_LEN				12
#define CFG_MSG_C_LEN			3

/* COnfig messages */

#define CFG_PRT_LEN				20
#define CFG_USB_LEN				108
#define CFG_MSG_LEN				8
#define CFG_INF_LEN				0 /* Varying */
#define CFG_NAV5_LEN			36
#define CFG_NAVX5_LEN			40
#define CFG_DAT_LEN				44
#define CFG_RATE_LEN			6
#define CFG_SBAS_LEN			8
#define CFG_NMEA_LEN			4
#define CFG_GNSS_LEN			0 /* Varying */
#define CFG_TP5_LEN				32
#define CFG_RXM_LEN				2
#define CFG_PM2_LEN				44
#define CFG_ITFM_LEN			8
#define CFG_RINV_LEN			0 /* Varying */
#define CFG_ANT_LEN				4
#define CFG_LOGFILTER_LEN		12

/* NAV message lengths */

#define NAV_TIMEUTC_LEN			20
#define NAV_POSLLH_LEN			28
#define NAV_POSECEF_LEN			20
#define NAV_STATUS_LEN			16
#define NAV_VELECEF_LEN			20
#define NAV_VELNED_LEN			36
#define NAV_PVT_LEN				84

#define MON_HW_LEN				60

/*******************************************************************************
 * 							UBX configuration messages
 *******************************************************************************/

#define UBX_MESSAGE_POLL_PRT								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_PRT,			/* Message ID */					\
	CFG_PRT_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_USB								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_USB,			/* Message ID */					\
	CFG_USB_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_MSG								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_MSG,			/* Message ID */					\
	CFG_MSG_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_INF								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_INF,			/* Message ID */					\
	CFG_INF_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_NAV5								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_NAV5,			/* Message ID */					\
	CFG_NAV5_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_NAVX5								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_NAVX5,			/* Message ID */					\
	CFG_NAVX5_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_DAT								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_DAT,			/* Message ID */					\
	CFG_DAT_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_RATE								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_RATE,			/* Message ID */					\
	CFG_RATE_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_SBAS								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_SBAS,			/* Message ID */					\
	CFG_SBAS_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_NMEA								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_NMEA,			/* Message ID */					\
	CFG_NMEA_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_GNSS								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_GNSS,			/* Message ID */					\
	CFG_GNSS_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_TP5								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_TP5,			/* Message ID */					\
	CFG_TP5_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_RXM								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_RXM,			/* Message ID */					\
	CFG_RXM_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_PM2								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_PM2,			/* Message ID */					\
	CFG_PM2_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_ITFM								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_ITFM,			/* Message ID */					\
	CFG_ITFM_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_RINV								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_PM2,			/* Message ID */					\
	CFG_PM2_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_ANT								\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_ANT,			/* Message ID */					\
	CFG_ANT_LEN,		/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}

#define UBX_MESSAGE_POLL_LOGFILTER							\
{ 															\
	UBX_CFG,			/* Message class */					\
	CFG_LOGFILTER,		/* Message ID */					\
	CFG_LOGFILTER_LEN,	/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}


#define UBX_MESSAGE_POLL_HW									\
{ 															\
	UBX_MON,			/* Message class */					\
	MON_HW,				/* Message ID */					\
	MON_HW_LEN,			/* Number of bytes to receive */	\
	0,					/* Number of bytes to transmit */	\
	true,				/* Poll message */					\
	(uint8_t[]){0}		/* Payload */						\
}
/*******************************************************************************
 * 							UBX NAV poll messages
 *******************************************************************************/

#define UBX_MESSAGE_POLL_TIMEUTC	\
{ 									\
	UBX_NAV,						\
	NAV_TIMEUTC,					\
	NAV_TIMEUTC_LEN,				\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

#define UBX_MESSAGE_POLL_POSLLH		\
{ 									\
	UBX_NAV,						\
	NAV_POSLLH,						\
	NAV_POSLLH_LEN,					\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

#define UBX_MESSAGE_POLL_NAV_STAT	\
{									\
	UBX_NAV,						\
	NAV_STATUS,						\
	NAV_STATUS_LEN,					\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

#define UBX_MESSAGE_POLL_PVT		\
{ 									\
	UBX_NAV,						\
	NAV_PVT,						\
	NAV_PVT_LEN,					\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

#define UBX_MESSAGE_POLL_VELECEF	\
{ 									\
	UBX_NAV,						\
	NAV_VELECEF,					\
	NAV_VELECEF_LEN,				\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

#define UBX_MESSAGE_POLL_VELNED		\
{ 									\
	UBX_NAV,						\
	NAV_VELNED,						\
	NAV_VELNED_LEN,					\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

/*******************************************************************************
 * 							UBX config messages
 *******************************************************************************/

#define UBX_MESSAGE_LOAD_PERM_CONFIG		\
{ 											\
	UBX_CFG,								\
	CFG_CFG,								\
	0,										\
	CFG_CFG_LEN,							\
	false,									\
	(uint8_t[])								\
	{0x00, 0x00, 0x00, 0x00,				\
	 0x00, 0x00, 0x00, 0x00,				\
	 0x1F, 0x06, 0x00, 0x00}				\
}

#define UBX_MESSAGE_SAVE_CURRENT_CONFIG		\
{ 											\
	UBX_CFG,								\
	CFG_CFG,								\
	0,										\
	CFG_CFG_LEN,							\
	false,									\
	(uint8_t[])								\
	{0x00, 0x00, 0x00, 0x00,				\
	 0x1F, 0x06, 0x00, 0x00,				\
	 0x00, 0x00, 0x00, 0x00}				\
}

#define UBX_MESSAGE_LOAD_DEFAULT_CONFIG		\
{ 											\
	UBX_CFG,								\
	CFG_CFG,								\
	0,										\
	CFG_CFG_LEN,							\
	false,									\
	(uint8_t[])								\
	{0x1F, 0x06, 0x00, 0x00,				\
	 0x00, 0x00, 0x00, 0x00,				\
	 0x00, 0x00, 0x00, 0x00}				\
}

#define UBX_MESSAGE_ACK				\
{ 									\
	UBX_ACK,						\
	ACK_ACK, 						\
	ACK_LEN,						\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

#define UBX_MESSAGE_NAK				\
{ 									\
	UBX_ACK,						\
	ACK_NAK, 						\
	ACK_LEN,						\
	0,								\
	true,							\
	(uint8_t[]){0}					\
}

#endif /* UBX_PARSER_H_ */
