/******************************************************************************
 * @file:		bmp_com.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	25. feb. 2016
 ******************************************************************************
 *	BMP (Buoy Message Protocol) communication module. Implements all of the
 *	interpretation and handling of received -and sent BMP messages.
 *	Handles all access to the log files and the system info structures,
 *	as well as the BMP parser.
 ******************************************************************************/

#include <string.h>

#include "system.h"
#include "timer.h"
#include "err.h"

#include "bmp_com.h"
#include "fifo_tx.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static bool TransmitData(BMP_TransferArgs * args);
static bool ReceiveDataTransfer(BMP_TransferArgs * args);
static bool SetStreamArgs(bool send, BMP_TransferArgs * args);

static void Error_Handler(BMP_TransferArgs * args);

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

static BMP_Message 	bmpMsg = {0};

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Returns a pointer to the selected message type.
 *******************************************************************************/
BMP_Message * BMP_GetMessage(BMP_MessageType_msg type){

	switch(type){

	case msg_pollClientStatus:

		bmpMsg = (BMP_Message)BMP_RADIO_POLL_SYS_INFO;
		break;

	case msg_sessionFinished:

		bmpMsg = (BMP_Message)BMP_RADIO_END;
		break;

	case msg_pollTbrDetections:

		bmpMsg = (BMP_Message)BMP_RADIO_POLL_LAST;
		bmpMsg.deviceLog = dev_tbr700;
		break;

	case msg_pollGaugeData:

		bmpMsg = (BMP_Message)BMP_RADIO_POLL_LAST;
		bmpMsg.deviceLog = dev_gauge;
		break;

	case msg_pollGpsData:

		bmpMsg = (BMP_Message)BMP_RADIO_POLL_LAST;
		bmpMsg.deviceLog = dev_gps;
		break;

	case msg_nak:
		bmpMsg.cmd = bc_nak;
		bmpMsg.mode = acm_response;
		break;

	default:
		break;
	}

	bmpMsg.senderAddr = *SYS_GetID();

	return &bmpMsg;
}


/*******************************************************************************
 * @brief Receive incoming BMP message from stream. The BMP_TransferArgs defines
 * the arguments needed to receive the message.
 *******************************************************************************/
bool BMP_ReceiveTransmission(BMP_TransferArgs * args){

	bool res = true;

	res = BMP_IdentifyMessage(args->receivedMsg, args->streamArgs->stream);

	if (!res) goto exit;

	/* Check message addressing */
	if (args->receivedMsg->receiverAddr != BMP_BROADCAST_ADDR &&
			args->receivedMsg->receiverAddr != *SYS_GetID()){

		goto exit; /* Cancel reception if the message is not addressed to this node */
	}

	else if (args->receivedMsg->cmd == bc_fromLast &&
			args->receivedMsg->mode == acm_poll){

		/* Check if new data is available in local log file. */
		if (!LOG_NewDataAvailable(
			args->receivedMsg->deviceLog, type_local)){

			/* Disregard the message */
			goto exit;
		}
	}

	BMP_RegisterNewMessage(args);

	/* Receive data if message contains data */
	if (args->receivedMsg->mode & BMP_DATA_MASK) res = ReceiveDataTransfer(args);


exit:
	if (!res) Error_Handler(args);
	return res;
}


/*******************************************************************************
 * @brief Transmit the specified BMP message via the specified stream.
 * If data transmission, specified log file will be accessed and transferred.
 *******************************************************************************/
bool BMP_TransmitMessage(BMP_TransferArgs * args){

	bool res = true;


	if(args->msgToTransfer->cmd == bc_fromLast &&
			(args->msgToTransfer->mode & BMP_DATA_MASK)){

		/* Retrieve last read file name, to send in the BMP header */
		LOG_GetLastReadFile(
			args->msgToTransfer->deviceLog,
			args->streamArgs->logToAccess,
			args->msgToTransfer->fileName);
	}

	if (args->includeHeaderInMsg) res = BMP_MakeMessageHeader(
			args->msgToTransfer, args->streamArgs->stream);


	/* Transmit data if data transfer message */
	if (res && (args->msgToTransfer->mode & BMP_DATA_MASK)) res = TransmitData(args);

	/* End transfer immediately if no data transfer */
	else if (res) res = BMP_EndTransfer(args->streamArgs->stream,
			args->streamArgs->waitAfterTransferEnd_ms);

	if (!res) goto exit;

	/* Start receive timeout timer in case the router nodes have no new data to send */
	else if ((args->msgToTransfer->cmd & BMP_FILE_MASK) &&
			args->msgToTransfer->mode == acm_poll){

		StartTimer_s(BMP_DATA_RECEIVE_TIMEOUT, args->receiveFinished);
	}

exit:
	if (!res) Error_Handler(args);
	return res;
}


/*******************************************************************************
 * @brief Registers the received BMP message in the specified BMP_TransferArgs
 * structure. Actions are based on the received message's command and access mode.
 *******************************************************************************/
void BMP_RegisterNewMessage(BMP_TransferArgs * args){


	args->sysIndex = SYS_IdentifySystem(args->receivedMsg->senderAddr);


	if (args->receivedMsg->cmd == bc_identify ||
			args->receivedMsg->cmd == bc_config ||
			args->receivedMsg->cmd == bc_entireLog){

		return; /* Not implemented */
	}

	switch(args->receivedMsg->mode){

	case acm_poll:

		SYS_UpdateLastContact(args->sysIndex);

		/* Notify transfer in progress */
		args->flags[args->sysIndex] |= BMP_TIP_MASK;

		if (args->receivedMsg->cmd & BMP_WRITE){

			args->msgToTransfer->mode = acm_write;
			args->flags[args->sysIndex] |= BMP_AWAIT_ACK_MASK;
		}

		else if (args->receivedMsg->cmd & BMP_UPDATE){

			args->msgToTransfer->mode = acm_update;
		}

		/* Copy received message parameters to the message to transfer */
		args->msgToTransfer->cmd = args->receivedMsg->cmd;
		args->msgToTransfer->deviceLog = args->receivedMsg->deviceLog;

		args->msgToTransfer->receiverAddr =	args->receivedMsg->senderAddr;
		args->msgToTransfer->senderAddr = *SYS_GetID();

		if (args->receivedMsg->cmd & BMP_FILE_MASK){
			strcpy(args->msgToTransfer->fileName, args->receivedMsg->fileName);
		}

		/* Copy into "write" message as well, to have a copy of the last sent
		 * acm_write or acm_update message */
		if (args->writeMsg){

			if (args->receivedMsg->cmd & BMP_FILE_MASK){
				strcpy(args->writeMsg->fileName, args->receivedMsg->fileName);
			}

			/* Keep track of file name pointer */
			void * filePtr = args->writeMsg->fileName;

			*args->writeMsg = *args->msgToTransfer;
			args->writeMsg->fileName = filePtr;
		}

		break;

	case acm_write:

		/* Reset LE timer to make sure that all write messages
		 * are received before timeout LE timer is set when a
		 * "transFinished" message is received */
		ResetTimer();

		break;

	case acm_update:

		SYS_UpdateLastContact(args->sysIndex);


		switch(args->receivedMsg->cmd){

		case bc_systemStatus:

			/* Waiting for all systems to send their system status.
			 * After the timeout, the callback function will be called */
			StartTimer_s(BMP_DATA_RECEIVE_TIMEOUT, args->receiveFinished);
			break;

		default:
			break;
		}
		break; /* acm_update */

	case acm_response:

		SYS_UpdateLastContact(args->sysIndex);

		switch(args->receivedMsg->cmd){

		case bc_ack:

			args->flags[args->sysIndex] |= BMP_ACK_MASK;
			args->flags[args->sysIndex] &= ~BMP_AWAIT_ACK_MASK;

			break;

		case bc_nak:

			/* Set NAK flag */
			args->flags[args->sysIndex] |= BMP_NAK_MASK;


			/* Check if the NAK is for the last sent "data message" */
			if (args->writeMsg->cmd & BMP_DATA_MASK){


				/* Keep track of file name pointer */
				void * filePtr = args->msgToTransfer->fileName;

				/* Copy last sent "write" message into "msgToTransfer" */
				*args->msgToTransfer = *args->writeMsg;
				args->msgToTransfer->fileName = filePtr;


				/* Last transferred message was file data */
				if (args->writeMsg->cmd & BMP_FILE_MASK){

					/* Copy file name from write message */
					strcpy(args->receivedMsg->fileName, args->writeMsg->fileName);

					/* Initiate transfer */
					args->flags[args->sysIndex] |= BMP_TIP_MASK;
				}
			}

			break;

		case bc_transFinished:

			/* Set receive finished flag */
			args->flags[args->sysIndex] |= BMP_RF_MASK;

			/* Define ACK message to be issued to the sender */

			args->msgToTransfer->receiverAddr =
					args->receivedMsg->senderAddr;

			args->msgToTransfer->senderAddr = *SYS_GetID();

			args->msgToTransfer->mode = acm_response;
			args->msgToTransfer->cmd = bc_ack;

			/* Waiting to receive data from all systems. After timeout, call back
		     * to the specified function */
			StartTimer_s(BMP_DATA_RECEIVE_TIMEOUT, args->receiveFinished);

			break;

		case bc_sessionFinished:

			/* Set session end flag, meaning that
			 * connection can be closed */
			args->flags[args->sysIndex] |= BMP_SESSION_END_MASK;
			break;

		default:
			break;
		}

		break; /* acm_response */
	}
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Error handler for the BMP Com module.
 *******************************************************************************/
static void Error_Handler(BMP_TransferArgs * args){

	switch(error.erRes){

	case err_txTimeout: /* TX stream was not empty when trying to write to it */

		if (error.errCount < 3) error.retVal = ret_resolvedRetry;

		else {
			FIFO_TX_Reset(DEV_TO_STREAM((Device_dev)error.origin));

			/* Call transfer complete to stop transfer */
			if (args->streamArgs->transferComplete) args->streamArgs->transferComplete();
			error.retVal = ret_resolved;
		}

		break;

	case err_rxTimeout: /* Data reception ended abruptly, expected data was not received */

		error.retVal = ret_notResolved;
		break;

	case err_badFileName:
	case err_noFile:
	case err_headerErr:

		if (error.errCount >= 2){
			/* Call transfer complete to stop transfer */
			if (args->streamArgs->transferComplete) args->streamArgs->transferComplete();
		}

		break;

	default: /* Could not resolve */
		error.retVal = ret_notResolved;
		break;
	}

	if (error.retVal == ret_resolved) ERR_RESET();
}


/*******************************************************************************
 * @brief Sets the LOG_StreamArgs accessMode based on the incoming -or outgoing
 * BMP message registered in the BMP_TransferArgs structure.
 *******************************************************************************/
static bool SetStreamArgs(bool send, BMP_TransferArgs * args){

	bool res = true;
	BMP_Message * msgPtr;

	if (send){
		msgPtr = args->msgToTransfer;
		args->sysIndex = SYS_IdentifySystem(msgPtr->receiverAddr);
	}

	else {
		msgPtr = args->receivedMsg;

		/* Callback with file name of received data */
		if (args->fileReceived) args->fileReceived(msgPtr->fileName);
	}


	/* Set log file to access */
	args->streamArgs->fileToAccess =
			msgPtr->fileName;


	/* Set device log to access */
	args->streamArgs->deviceLogToAccess =
			msgPtr->deviceLog;


	/* Convert BMP command to LOG accessMode */
	switch (msgPtr->cmd){

	case bc_fromLast:

		if (send) args->streamArgs->accessMode = am_readFromLast;
		else args->streamArgs->accessMode = am_appendToSpecifiedFile;
		break;

	case bc_specifiedLog:

		if (send) args->streamArgs->accessMode = am_readSpecifiedFile;
		else args->streamArgs->accessMode = am_appendToSpecifiedFile;
		break;

	case bc_lastLogFile:

		if (send) args->streamArgs->accessMode = am_readCurrentFile;
		else args->streamArgs->accessMode = am_appendToCurrentFile;
		break;

	default:
		break;
	}

	return res;
}


/*******************************************************************************
 * @brief Handles all data transfer from log file, or from the system info
 * structures.
 *******************************************************************************/
static bool TransmitData(BMP_TransferArgs * args){

	bool res = true;

	/* Send file data if the BMP message specifies file transfer */
	if (args->msgToTransfer->cmd & BMP_FILE_MASK){

		SetStreamArgs(true, args);
		res = LOG_StreamFromStorage(args->streamArgs);
	}

	/* Transfer system info data */
	else if (args->msgToTransfer->cmd == bc_systemStatus){

		bool format = false;
		uint8_t nSystemInfoToTransfer = 1;


		if (*SYS_GetID() == *SYS_GetGateway()){ /* Check if this is the gateway node */

			/* Format the system info to ASCII if not using
			 * the BMP protocol for data transfer */
			if (!USE_PROTOCOL_GSM) format = true;
			nSystemInfoToTransfer = SYS_GetnRegSystems();
		}


		for (int i = 0; i < nSystemInfoToTransfer; i++){

			SYS_PrintSystemInfo(args->streamArgs->stream, i, format);
		}

		/* Initiate transfer */
		res = BMP_EndTransfer(args->streamArgs->stream,
					args->streamArgs->waitAfterTransferEnd_ms);

		/* Notify transfer finished */
		args->streamArgs->transferComplete();
	}

	return res;
}


/*******************************************************************************
 * @brief Handle received data message, whether it is "write" or "update" data.
 *******************************************************************************/
static bool ReceiveDataTransfer(BMP_TransferArgs * args){

	bool res = true;

	if (args->receivedMsg->cmd & BMP_FILE_MASK){

		/* Configure the stream args provided in the BMP_Transferargs structure pointer,
		 * for this particular transmission. */
		SetStreamArgs(false, args);

		/* Receive data from stream and write to specified log file */
		res = LOG_StreamToStorage(args->streamArgs);
	}

	else if (args->receivedMsg->cmd == bc_systemStatus){

		/* Update system info if received */
		res = SYS_UpdateSystemInfo(args->streamArgs->stream, args->sysIndex);
		if (res) res = BMP_ReceiveEOT(args->streamArgs->stream);
	}

	return res;
}
