/******************************************************************************
 * @file:		bmp_parser.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	7. feb. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include "timer.h"
#include "err.h"

#include "bmp_parser.h"

#include "fifo_rx.h"
#include "fifo_tx.h"
#include "stream.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static bool CheckDeviceLog(Device_dev device);
static bool CheckAccessMode(uint8_t acm);
static bool CheckCommand(uint8_t cmd, uint8_t acm);

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Puts the last byte (RMP EOT char) in the
 * specified stream buffer and enables transmit.
 *******************************************************************************/
bool BMP_EndTransfer(Stream stream, uint8_t waitAfterEnd_ms){

	bool res = true, finished = false;

	StartTimer_ms(BMP_TX_FIFO_TIMEOUT_MS);

	while (!finished){

		if (TimeElapsed()){
			res = false;
			finished = true;
			ERR_RAISE(err_txTimeout, STREAM_TO_DEV(stream));
		}

		if (FIFO_TX_Free(stream)){

			FIFO_TX_Put(stream, BMP_EOT);
			finished = true;
		}
	}

	STREAM_Flush(stream);

	/* Delay to ensure that module transfer message before next comes */
	Delay_ms(waitAfterEnd_ms);

	return res;
}


/*******************************************************************************
 * @brief Creates the radio message protocol header using the provided data
 * structure. The data is put in the specified output stream buffer.
 *******************************************************************************/
bool BMP_MakeMessageHeader(
		BMP_Message * msg,
		Stream stream){


	bool res = true;
	StartTimer_ms(BMP_TX_FIFO_TIMEOUT_MS);

	/* Wait for TX buffer to become empty */
	while (!FIFO_TX_Empty(stream) && !TimeElapsed()) STREAM_Flush(stream);

	if (FIFO_TX_Empty(stream)){

		/* "Start of header" bytes */
		for (int i = 0; i < BMP_N_SOH_CHARS; i++){
			FIFO_TX_Put(stream, BMP_SOH_CHARS[i]);
		}

		FIFO_TX_Put(stream, msg->senderAddr);		/* The UID (address) of this unit */
		FIFO_TX_Put(stream, msg->senderAddr >> 8);
		FIFO_TX_Put(stream, msg->senderAddr >> 16);
		FIFO_TX_Put(stream, msg->senderAddr >> 24);

		FIFO_TX_Put(stream, msg->receiverAddr);		/* The UID (address) of the unit to respond */
		FIFO_TX_Put(stream, msg->receiverAddr >> 8);
		FIFO_TX_Put(stream, msg->receiverAddr >> 16);
		FIFO_TX_Put(stream, msg->receiverAddr >> 24);

		FIFO_TX_Put(stream, msg->mode);
		FIFO_TX_Put(stream, msg->cmd);
		FIFO_TX_Put(stream, msg->deviceLog);

		if(msg->cmd & BMP_FILE_MASK){	/* Send file name if file data is transferred */

			int i = 0;

			if (msg->fileName){

				while (msg->fileName[i] != '\0' && i < FILENAME_MAX_SIZE - 1){
					FIFO_TX_Put(stream, msg->fileName[i++]);
				}
			}

			FIFO_TX_Put(stream, BMP_EOFN); /* "End of file name" byte */
		}

		FIFO_TX_Put(stream, BMP_STX);	/* "Start of text" byte */
	}

	else {
		ERR_RAISE(err_txTimeout, STREAM_TO_DEV(stream));
		res = false;
	}

	return res;
}


/*******************************************************************************
 * @brief Identifies the incoming BMP message from the specified input stream
 * buffer.
 * @return false if RX timeout, or if the message did not fulfill the BMP
 * specification.
 *******************************************************************************/
bool BMP_IdentifyMessage(BMP_Message * msg, Stream stream){

	bool res = true;
	uint8_t data;
	int stepNr = 0, sohCount = 0, sohCounter = 0;
	bool finished = false;

	/* Reset sender -and receiver addresses */
	msg->receiverAddr = 0;
	msg->senderAddr = 0;


	while (!finished){

		StartTimer_ms(BMP_RX_FIFO_TIMEOUT_MS);

		/* Allow the fifo buffer to fill up while reading */
		while (!FIFO_RX_Get(stream, &data)){
			if (TimeElapsed()){
				res = false;
				ERR_RAISE(err_rxTimeout, STREAM_TO_DEV(stream));
				goto end;
			}
		}

		switch(stepNr){

		case 0: /* Start of header */

			if (data == BMP_SOH_CHARS[sohCount]){
				sohCount++;

				if (sohCount == BMP_N_SOH_CHARS) stepNr++;
			}

			else if (sohCount > 0 && sohCounter > 50){ /* Looking for SOH char */
				res = false;
				ERR_RAISE(err_sohTimeout, STREAM_TO_DEV(stream));
				finished = true;
			}

			else sohCounter++;

			break;

		case 1:{ /* Sender address */

			static int i = 0;

			msg->senderAddr |= data << 8 * i;
			i++;

			if (i > 3){
				i = 0;
				stepNr++;
			}
		}
			break;

		case 2:{ /* Receiver address */

			static int i = 0;

			msg->receiverAddr |= data << 8 * i;
			i++;

			if (i > 3){
				i = 0;
				stepNr++;
			}
		}
			break;

		case 3: /* Access mode */

			/* Check if received access mode is valid */
			if (CheckAccessMode(data)) msg->mode = (BMP_AccessMode_acm)data;

			else {
				res = false;
				finished = true;
			}
			stepNr++;
			break;

		case 4: /* Command */

			/* Check if received command is valid */
			if (CheckCommand(data, msg->mode)) msg->cmd = (BMP_Command_bc)data;

			else {
				res = false;
				finished = true;
			}
			stepNr++;

			break;

		case 5: /* Log to access */

			/* Check if device log is valid */
			if (CheckDeviceLog((Device_dev)data)) msg->deviceLog = (Device_dev)data;

			else {
				res = false;
				finished = true;
			}

			/* Get file name if request for, -or incoming
			 * file transmission */
			if (msg->cmd & BMP_FILE_MASK) stepNr++;
			else stepNr += 2;

			break;

		case 6:{ /* Get file name */

			static int i = 0;

			/* Check for "end of file name" byte */
			if (data == BMP_EOFN || i >= FILENAME_MAX_SIZE - 1){

				msg->fileName[i] = '\0';
				stepNr++;
				i = 0;
			}

			else msg->fileName[i++] = data;
		}
			break;

		case 7: /* Start of text */

			if (BMP_STX == data){ /* Check that start of text char is received */

				if (msg->mode == acm_write || msg->mode == acm_update) finished = true;

				/* Continue to reading end char if message is not a "write" -or "update" message */
				else stepNr++;
			}
			else {
				res = false;
				finished = true;
			}

			break;

		case 8:

			/* Check end of transmission char */
			if (BMP_EOT == data) finished = true;

			else res = false;

			break;
		}
	}

	if (!res && stepNr > 2) ERR_RAISE(err_headerErr, STREAM_TO_DEV(stream));
end:
	return res;
}


/*******************************************************************************
 * @brief Called to identify the end of transmission (EOT) char of a received
 * BMP message.
 *******************************************************************************/
bool BMP_ReceiveEOT(Stream stream){

	bool finished = false, res = true;
	uint8_t data;

	StartTimer_ms(100);

	while (!finished){

		FIFO_RX_Get(stream, &data);

		if (data == BMP_EOT) finished = true;

		else if (TimeElapsed()){
			res = false;
			ERR_RAISE(err_rxTimeout, STREAM_TO_DEV(stream));
		}
	}
	return res;
}

/*******************************************************************************
 * @brief Compares the provided access mode with the allowed values.
 *******************************************************************************/
static bool CheckDeviceLog(Device_dev deviceLog){

	bool res = false;

	if (deviceLog >= 0 && deviceLog < N_LOG_DEVICES) res = true;

	return res;
}


/*******************************************************************************
 * @brief Compares the provided access mode with the allowed values.
 *******************************************************************************/
static bool CheckAccessMode(uint8_t acm){

	bool res = false;
	if (
			acm == acm_poll 	||
			acm == acm_write 	||
			acm == acm_response ||
			acm == acm_update){

		res = true;
	}
	return res;
}


/*******************************************************************************
 * @brief Compares the provided command with the allowed values. Also compares
 * it to the access mode, to ensure that they are a valid combination.
 *******************************************************************************/
static bool CheckCommand(uint8_t cmd, uint8_t acm){

	bool res = false;

	if (
			cmd == bc_ack 				||
			cmd == bc_config 			||
			cmd == bc_entireLog 		||
			cmd == bc_fromLast			||
			cmd == bc_identify			||
			cmd == bc_lastLogFile		||
			cmd == bc_nak				||
			cmd == bc_sessionFinished	||
			cmd == bc_specifiedLog		||
			cmd == bc_systemStatus		||
			cmd == bc_transFinished){

		if (cmd & acm) res = true;
	}
	return res;
}
