/******************************************************************************
 * @file:		comm.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	13. mai 2016
 ******************************************************************************
 *	The communication module is a high level module which implements the
 *	BMP and the GSM modules to allow data transfer to -and from both the
 *	radio module and the GSM module.
 ******************************************************************************/

#include <sys/time.h>
#include <string.h>

#include "timer.h"
#include "system.h"
#include "system_format.h"
#include "err.h"

#include "comm.h"

#include "bmp_parser.h"
#include "bmp_com.h"

#include "gsm_com.h"
#include "gsm_inet.h"
#include "radio_config.h"


/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static void InitializeArgsRadio();
static void InitializeArgsGsm();

static bool ReceiveRadio();
static bool ReceiveGsm();

static bool TransferSpecifiedRadio(BMP_Message * msg);
static bool TransferSpecifiedGsm(BMP_Message * msg);

static bool TransferRadio();
static bool TransferGsm();

static void OnTransferCompleteRadio();
static void OnTransferCompleteGsm();

static BMP_Message * CreateNewBmpMsg();

static void OnRadioAckTimeout();
static void OnGsmAckTimeout();

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

#define N_ARGS			2
#define N_MAX_MSG		N_ARGS * 3

static const uint8_t 	gsmArgs = 0, radioArgs = 1;

static LOG_StreamArgs 	streamArgs[N_ARGS];
static BMP_TransferArgs remoteSystemArgs[N_ARGS];
static BMP_TransferArgs * rArgs, * gArgs;

static uint8_t 			flags[N_MAX_SYSTEMS] = {0};
static BMP_Message		messages[N_MAX_MSG];
static char 			filesNames[N_MAX_MSG][FILENAME_MAX_SIZE];

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Returns the communication state of the system. The device log that has
 * been accessed is returned for saving log parameters.
 *******************************************************************************/
COM_State COM_GetState(Device_dev * deviceLog){

	COM_State state = state_idle;
	int mask = 1;
	uint8_t nSystems, sysIndex;
	bool maskMatch = false;

	*deviceLog = dev_noDevice; /* Set initial value */

	for (uint8_t arg = 0; arg < N_ARGS; arg++){

		/* Since flags is shared between the "remoteSystemArgs",
		 * gsmArgs need to be checked first */
		if (arg == gsmArgs){

			 /* Check if a GSM connection has been active (registered addr) */
			if (gArgs->receivedMsg->senderAddr){
				sysIndex = SYS_IdentifySystem(gArgs->receivedMsg->senderAddr);
				nSystems = sysIndex + 1;
			}
			else continue;
		}

		else {
			sysIndex = 0;
			nSystems = SYS_GetnRegSystems();
		}

		for (;sysIndex < nSystems; sysIndex++){

			if (remoteSystemArgs[arg].flags[sysIndex]){

				while(!maskMatch && mask <= 0x80){

					maskMatch = true;

					switch(remoteSystemArgs[arg].flags[sysIndex] & mask){

					case BMP_TF_MASK: /* Data transfer reported finished from the lower level
					 	 	 	 	   * com module or log manager (no more data to transfer) */

						remoteSystemArgs[arg].flags[sysIndex] &= ~BMP_TF_MASK;

						if (arg == gsmArgs) state = state_transFinished_remote;
						else state = state_transFinished_local;

						break;

					case BMP_RF_MASK: /* A "transmission finished" packet has been received,
					 	 	 	 	   * and transmission is considered finished */

						remoteSystemArgs[arg].flags[sysIndex] &= ~BMP_RF_MASK;

						if (arg == gsmArgs) state = state_recvFinished_remote;
						else state = state_recvFinished_local;

						break;

					case BMP_TIP_MASK: /* Transfer in progress */

						/* Cancel transfer if message to transfer is not a data message
						 * ("write" or "update" (fail-safe feature)) */
						if (!(remoteSystemArgs[arg].msgToTransfer->mode & BMP_DATA_MASK)){

							remoteSystemArgs[arg].flags[sysIndex] &= ~BMP_TIP_MASK;
							maskMatch = false;
						}

						else if (arg == gsmArgs) state = state_transInProgress_remote;
						else state = state_transInProgress_local;

						break;

					case BMP_ACK_MASK: /* The receiver of the sent "transmission finished" packet
										* reported transmission OK */

						remoteSystemArgs[arg].flags[sysIndex] &= ~BMP_ACK_MASK;

						if (arg == gsmArgs) state = state_ackReceived_remote;
						else state = state_ackReceived_local;


						if (remoteSystemArgs[arg].writeMsg->cmd == bc_fromLast){

							/* Return the device log that has been accessed */
							*deviceLog = remoteSystemArgs[arg].writeMsg->deviceLog;

							/* Reset "write" message when ACK is received for the specific message,
							 * so that a future NAK may not be registered for the "write" message. */
							remoteSystemArgs[arg].writeMsg->cmd = 0;
						}

						break;

					case BMP_NAK_MASK: /* Receiver of last sent packet reported it to be faulty */

						remoteSystemArgs[arg].flags[sysIndex] &= ~BMP_NAK_MASK;

						/* Address the response message to the sender of the NAK message,
						 * so that a retransmit is only sent to the node that sent the NAK */
						remoteSystemArgs[arg].msgToTransfer->receiverAddr = *SYS_IdentifyIndex(sysIndex);

						if (arg == gsmArgs) state = state_nakReceived_remote;
						else state = state_nakReceived_local;


						if (remoteSystemArgs[arg].writeMsg->cmd == bc_fromLast){

							/* Return the device log that has been accessed */
							*deviceLog = remoteSystemArgs[arg].writeMsg->deviceLog;
						}

						break;

					case BMP_ACK_TIMEOUT_MASK: /* An expected ack message on a sent
												* "write" message was not received within the defined time */

						if (remoteSystemArgs[arg].flags[sysIndex] & BMP_AWAIT_ACK_MASK){

							if (arg == gsmArgs) state = state_recoverLogParam_remote;
							else state = state_recoverLogParam_local;

							/* Return the device log that has been accessed */
							*deviceLog = remoteSystemArgs[arg].writeMsg->deviceLog;
						}

						else maskMatch = false;

						remoteSystemArgs[arg].flags[sysIndex] &= ~(BMP_ACK_TIMEOUT_MASK | BMP_AWAIT_ACK_MASK);

						break;

					case BMP_SESSION_END_MASK: /* "Session finished" packet received  */

						remoteSystemArgs[arg].flags[sysIndex] &= ~BMP_SESSION_END_MASK;

						if (arg == gsmArgs) state = state_remoteSessionComplete;
						else state = state_localSessionComplete;
						break;

					default:

						maskMatch = false;
						break;
					}
					mask <<= 1;
				}

				if (maskMatch){
					goto exit; /* Break */
				}
			}
		}
		mask = 1; /* Reset mask counter */
	}

exit:
	return state;
}


/*******************************************************************************
 * @brief Initializes the comm module.
 *******************************************************************************/
void COM_Init(){

	rArgs = &remoteSystemArgs[radioArgs];
	gArgs = &remoteSystemArgs[gsmArgs];

	InitializeArgsRadio();
	InitializeArgsGsm();
}


/*******************************************************************************
 * @brief Receive data from the specified stream.
 *******************************************************************************/
bool COM_Receive(Stream stream){

	if (stream == stream_radio) return ReceiveRadio();

	else if (stream == stream_gsm) return ReceiveGsm();

	else return true;
}


/*******************************************************************************
 * @brief Send current "messageToTransfer" stored in the private
 * BMP_TransferArgs structure.
 *******************************************************************************/
bool COM_Send(Stream stream){

	if (stream == stream_radio) return TransferRadio();

	else if (stream == stream_gsm) return TransferGsm();

	else return true;
}


/*******************************************************************************
 * @brief Send specified message to the specified stream.
 *******************************************************************************/
bool COM_SendMessage(Stream stream, BMP_Message * msg){

	if (stream == stream_radio) return TransferSpecifiedRadio(msg);

	else if (stream == stream_gsm) return TransferSpecifiedGsm(msg);

	else return true;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void InitializeArgsRadio(){

	rArgs->flags = &flags[0];
	rArgs->includeHeaderInMsg = true;
	rArgs->streamArgs = &streamArgs[radioArgs];

	rArgs->msgToTransfer = CreateNewBmpMsg();
	rArgs->receivedMsg = CreateNewBmpMsg();
	rArgs->writeMsg = CreateNewBmpMsg();

	rArgs->fileReceived = SYS_FormatReceivedFileName;

	rArgs->streamArgs->getNextFile = SYS_GetNextFile;
	rArgs->streamArgs->transferComplete = OnTransferCompleteRadio;

	rArgs->streamArgs->stream = stream_radio;
	rArgs->streamArgs->nBytesToTransfer = RADIO_MAX_PACKET_LEN - BMP_HEADER_W_FILE_LEN;
	rArgs->streamArgs->waitAfterTransferEnd_ms = RADIO_DEFAULT_SERIAL_TIMEOUT  + 110;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void InitializeArgsGsm(){

	gArgs->flags = &flags[0];
	gArgs->streamArgs = &streamArgs[gsmArgs];

	gArgs->msgToTransfer = CreateNewBmpMsg();
	gArgs->receivedMsg = CreateNewBmpMsg();
	gArgs->writeMsg = CreateNewBmpMsg();

	gArgs->streamArgs->getNextFile = SYS_GetNextFile;
	gArgs->streamArgs->transferComplete = OnTransferCompleteGsm;

	gArgs->streamArgs->stream = stream_gsm;
	gArgs->streamArgs->logToAccess = type_local;
	gArgs->streamArgs->waitAfterTransferEnd_ms = 1;

	if (USE_PROTOCOL_GSM){
		gArgs->includeHeaderInMsg = true;
		gArgs->streamArgs->nBytesToTransfer = GSM_TCP_MAX_PACKET_LEN - BMP_HEADER_W_FILE_LEN;
	}
	else {
		gArgs->includeHeaderInMsg = false;
		gArgs->streamArgs->nBytesToTransfer = GSM_TCP_MAX_PACKET_LEN;
	}
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static BMP_Message * CreateNewBmpMsg(){

	static int count = 0;

	if (count >= N_MAX_MSG) return NULL;

	else {

		messages[count].fileName = &filesNames[count][0];

		return &messages[count++];
	}
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TransferSpecifiedRadio(BMP_Message * msg){

	bool res;

	/* Temporary store the current BMP message pointer */
	void * temp = rArgs->msgToTransfer->fileName;

	*rArgs->msgToTransfer = *msg;


	if (msg->cmd == bc_nak){
		/* Set receiver of message for NAK messages */
		rArgs->msgToTransfer->receiverAddr =
			rArgs->receivedMsg->senderAddr;
	}

	/* Set callback function pointers for receive finished timeout */

	if (msg->cmd == bc_systemStatus) rArgs->receiveFinished = SYS_AnalyzeSystemInfo;

	else if (msg->mode & BMP_FILE_MASK) rArgs->receiveFinished = SYS_DataAcqFinished;

	res = COM_Send(stream_radio);

	rArgs->msgToTransfer->fileName = temp;

	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TransferSpecifiedGsm(BMP_Message * msg){

	bool res;

	/* Temporary store the current BMP message pointer */
	void * temp = gArgs->msgToTransfer->fileName;

	/* Copy message */
	*gArgs->msgToTransfer = *msg;


	if (msg->cmd == bc_nak){
		/* Set receiver of message for NAK messages */
		gArgs->msgToTransfer->receiverAddr =
				gArgs->receivedMsg->senderAddr;

		/* Stop any ongoing data transfer */
		gArgs->flags[gArgs->sysIndex] &= ~BMP_TIP_MASK;
	}


	res = COM_Send(stream_gsm);

	gArgs->msgToTransfer->fileName = temp;

	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TransferRadio(){

	/* Read from local log when sending data via radio */
	rArgs->streamArgs->logToAccess = type_local;
	return BMP_TransmitMessage(rArgs);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool TransferGsm(){

	bool res;

	if ((gArgs->msgToTransfer->cmd & BMP_DATA_MASK) ||
			USE_PROTOCOL_GSM) res = BMP_TransmitMessage(gArgs);

	else res = GSM_TransmitTCP(gArgs->msgToTransfer);

	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool ReceiveRadio(){

	/* Write to combined log when receiving data via radio */
	rArgs->streamArgs->logToAccess = type_combined;
	return BMP_ReceiveTransmission(rArgs);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool ReceiveGsm(){

	bool res;

	if (USE_PROTOCOL_GSM){

		res = BMP_ReceiveTransmission(gArgs);
	}

	else {
		res = GSM_ReceiveTCP(gArgs);
	}
	return res;
}


/*******************************************************************************
 * @brief Callback for log manager. Called when EOF has been reached.
 *******************************************************************************/
static void OnTransferCompleteRadio(){

	/* Clear "transfer in progress" flag */
	rArgs->flags[rArgs->sysIndex] &= ~BMP_TIP_MASK;

	/* Only respond on "write" messages */
	if (rArgs->msgToTransfer->mode == acm_write){

		rArgs->flags[rArgs->sysIndex] |= BMP_TF_MASK;

		/* Define message to signal that transfer is complete */
		rArgs->msgToTransfer->mode = acm_response;
		rArgs->msgToTransfer->cmd = bc_transFinished;

		/* Set timer for ACK to be received for a "from last" message */
		if (rArgs->writeMsg->cmd == bc_fromLast){
			StartTimer_s(BMP_DATA_RECEIVE_TIMEOUT + 2, OnRadioAckTimeout);
		}
	}
}


/*******************************************************************************
 * @brief Callback for log manager. Called when EOF has been reached.
 *******************************************************************************/
static void OnTransferCompleteGsm(){


	static int filesAccessed = 0;

	if (gArgs->msgToTransfer->mode == acm_write){

		if (!filesAccessed){

			/* Enable file transfer from local log as well */
			gArgs->streamArgs->logToAccess = type_combined;

			/* Change to "combined" file name to read the combined log file as well */
			gArgs->msgToTransfer->fileName[0] = 'C';

			filesAccessed++;
		}

		else { /* Both local -and combined log transferred */

			/* Clear "transfer in progress" flag */
			gArgs->flags[gArgs->sysIndex] &= ~BMP_TIP_MASK;

			gArgs->flags[gArgs->sysIndex] |= BMP_TF_MASK;
			gArgs->streamArgs->logToAccess = type_local;

			/* Change back to "local" file name in case of re-transmit */
			gArgs->msgToTransfer->fileName[0] = 'L';

			filesAccessed = 0;

			/* Define message to issue to receiver to signal that transfer is complete */
			gArgs->msgToTransfer->mode = acm_response;
			gArgs->msgToTransfer->cmd = bc_transFinished;

			/* Set timer for ACK to be received for "fromLast" messages */
			if (gArgs->writeMsg->cmd == bc_fromLast){
				StartLETimer_s(BMP_DATA_RECEIVE_TIMEOUT + 20, OnGsmAckTimeout);
			}
		}
	}

	/* Clear "transfer in progress" flag */
	else gArgs->flags[gArgs->sysIndex] &= ~BMP_TIP_MASK;
}


/*******************************************************************************
 * @brief Callback if Ack not received within specified time.
 *******************************************************************************/
static void OnRadioAckTimeout(){

	if (rArgs->flags[rArgs->sysIndex] & BMP_AWAIT_ACK_MASK){
		rArgs->flags[rArgs->sysIndex] |= BMP_ACK_TIMEOUT_MASK;
	}
}


/*******************************************************************************
 * @brief Callback if Ack not received within specified time.
 *******************************************************************************/
static void OnGsmAckTimeout(){

	if (gArgs->flags[gArgs->sysIndex] & BMP_AWAIT_ACK_MASK){
		gArgs->flags[gArgs->sysIndex] |= BMP_ACK_TIMEOUT_MASK;
	}
}
