/******************************************************************************
 * @file:		device_handlers.c
 * @project:	Ove Nesse
 * @version:	1.00
 * @created:	20. mai 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include <sys/time.h>

#include "err.h"
#include "system.h"
#include "system_format.h"
#include "ee_printf.h"

#include "timer.h"

#include "device_handlers.h"
#include "stream.h"
#include "log_manager.h"

#include "gauge_driver.h"
#include "radio_driver.h"
#include "aqua_driver.h"
#include "gsm_driver.h"
#include "gps_driver.h"
#include "spi_driver.h"
#include "imu_driver.h"
#include "tbr_driver.h"
#include "sd_driver.h"
#include "gsm_inet.h"
#include "file.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static void 		SyncSystemTime(struct timeval * time, bool force);
static void 		SyncTbrTime(struct timeval * time, bool force);
static uint16_t 	FormatGpsData(char * buf, GeoPosition * position);
static uint16_t 	FormatGaugeData(char * buf);
static void 		UpdateTbrDetections(bool force);

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

static LOG_StreamArgs tbrStreamArgs = {
		.stream = stream_tbr700,
		.accessMode = am_appendToCurrentFile,
		.deviceLogToAccess = dev_tbr700,
		.logToAccess = type_local
};

static bool tbr700Enabled = false;
static DEV_Settings * gpsSettings;

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Collect data from GPS. Clock update and position save rate can
 * be set with SAVE_GPS_POS_RATE and UPDATE_SYSTEM_TIME_RATE in "shared.h"
 *******************************************************************************/
bool DEV_CollectGpsData(){

	GPS_PVTSolution sol;
	uint16_t nBytes;
	char buf[50];
	bool resOk = true;
	static uint32_t savePosCounter = 0;


	resOk = GPS_GetPVTSolution(&sol, !GPS_UsingTxReady());

	if (resOk){

		if(sol.time.valid || sol.position.valid){
			/* Valid data received -> clear await data flag */
			gpsSettings->flags &= ~DEV_AWAIT_DATA_MASK;
		}

		if (sol.time.valid){

			DEV_SyncGpsTime(&sol.time, false);

		#if DEBUG_OPERATION
			STREAM_Printf(stream_debug, "Clock update\n");
		#endif
		}


		if (sol.position.valid){

			/* Update position in system info */
			SYS_UpdateGeoPos(&sol.position);

			/* Save position at given rate */
			if (!(savePosCounter % SAVE_GPS_POS_RATE)){

				nBytes = FormatGpsData(buf, &sol.position);
				resOk = LOG_WriteToStorage(dev_gps, type_local, buf, nBytes);

			#if DEBUG_OPERATION
				STREAM_Printf(stream_debug, "Save GPS data\n");
			#endif
			}
			savePosCounter++;
		}
	}
	return resOk;
}


/*******************************************************************************
 * @brief Collect data from battery gas gauge and save to log file.
 *******************************************************************************/
bool DEV_CollectGaugeData(){

	char buf[50];
	bool resOk = GAUGE_ReadAllRegisters();

	if (resOk){

		/* Update system variables */
		SYS_UpdatePowerInfo(
				GAUGE_BatteryVoltage_mV(),
				GAUGE_AverageCurrent_mA(),
				GAUGE_StateOfCharge());

		/* Format and write data to storage */
		uint16_t nBytes = FormatGaugeData(buf);

		resOk = LOG_WriteToStorage(dev_gauge, type_local, buf, nBytes);
	}
	return resOk;
}


/*******************************************************************************
 * @brief Collect TBR700 data. If data available in stream, data will be saved.
 * If no data available, the TBR will be polled for data from the last registered
 * TBR flash code (tag memory position).
 *******************************************************************************/
bool DEV_CollectTbrData(){

	int i = 0;
	bool resOk = true;

	if (STREAM_DataAv(stream_tbr700)){

		resOk = LOG_StreamToStorage(&tbrStreamArgs);
		if (error.erRes == err_rxTimeout) resOk = true;

		UpdateTbrDetections(false);
	}

	/* Request data */
	else {

		while (i < 2){

			if (resOk) TBR_RequestLog(i, SYS_TbrFlashCode(NULL));

			/* Receive data */
			if (resOk) resOk = LOG_StreamToStorage(&tbrStreamArgs);

			if (error.erRes == err_rxTimeout) resOk = true;

			Delay_ms(100);
			i++;
		}
		UpdateTbrDetections(true);
		TBR_ExitCommandMode();
	}

	return resOk;
}


/*******************************************************************************
 * @brief Default implemented configuration
 *******************************************************************************/
bool DEV_DefaultConfigGps(DEV_Settings * settings){

	bool resOk;
	gpsSettings = settings;

	GPS_OpConfig config = GPS_DEFAULT_OP_CONFIG;
	//config.mode = op_psm_cyclic;
	//config.onPeriod_s = 1;

	if (settings->flags & DEV_ALWAYS_ON_MASK){

		config.updatePeriod_ms = settings->interval_s * 1000;

		if (settings->interval_s <= 10) config.mode = op_psm_cyclic;
		else config.mode = op_psm_onOff;
	}

	/* update period set so that GPS position will be saved one time each cycle */
	else if (settings->onPeriod_s) config.updatePeriod_ms =
			settings->onPeriod_s * 1000 / (SAVE_GPS_POS_RATE * 2);


	resOk = GPS_DefaultSPIPortConfig(GPS_USE_TX_READY);

	if (resOk) resOk = GPS_DefaultMsgConfig();
	if (resOk) resOk = GPS_DefaultNAV5Config();

	if (resOk) resOk = GPS_SetOperationalMode(&config);

	 /* Save new permanent configuration on first time startup */
	if(firstTimeOnly && resOk) resOk = GPS_SaveNewPermanentConfig();

	return resOk;
}


/*******************************************************************************
 * @brief Default implemented configuration
 *******************************************************************************/
bool DEV_DefaultConfigRadio(){

	uint32_t uid;
	bool resOk = true;

//	resOk = RADIO_MemReset();

	/* Check if this system is a gateway */
	if (*SYS_GetGateway() == *SYS_GetID()){

		if (resOk) resOk = RADIO_SetMode(om_gateway);
	}

	else if (resOk) resOk = RADIO_SetMode(om_router);



	if (resOk) resOk = RADIO_DefaultMemConfig();

	if (resOk) resOk = RADIO_UniqueID(false, &uid);

	if (resOk) SYS_SetID(&uid);

	return resOk;
}


/*******************************************************************************
 * @brief Default implemented configuration
 *******************************************************************************/
bool DEV_DefaultConfigGsm(){

	bool resOk = GSM_DefaultConfig();
	if (resOk) GSM_ConfigureSmsSim();
	return resOk;
}


/*******************************************************************************
 * @brief Default implemented configuration
 *******************************************************************************/
bool DEV_DefaultConfigGauge(){

	return true;
}


/*******************************************************************************
 * @brief Default implemented configuration
 *******************************************************************************/
bool DEV_DefaultConfigTbr700(){

	TBR_TempInt_int interval = int_30min;
	bool resOk;

	TBR_TempInt_interval(true, &interval);
//
//		if (resOk && interval != int_5min){
//			interval = int_5min;
//			resOk = TBR_TempInt_interval(true, &interval);
//		}

	resOk = TBR_TestConnection();
	tbr700Enabled = resOk;
	return resOk;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool DEV_DefaultResetRadio(){

	return RADIO_Reset();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool DEV_DefaultResetGps(){

	int retryCount = 0;
	GPS_Reset();

	while (!DEV_DefaultConfigGps(gpsSettings) && retryCount < 10);

	return retryCount < 10 ? true : false;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool DEV_DefaultResetTbr(){

	return TBR_Restart();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool DEV_DefaultResetGsm(){

	GSM_Reset();
	return GSM_PowerOn();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool DEV_DefaultResetSdCard(){

	SD_PowerOff();
	Sleep_ms(20, true);
	return FILE_InitFileSystem();
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool DEV_DefaultResetGauge(){

	GAUGE_DeInit();
	Sleep_ms(20, true);
	GAUGE_Init();
	return true;
}


/*******************************************************************************
 * @brief Update the system -and TBR time with the provided GPS UTC time.
 * If forceSync is set, system time will be updated independent of the set
 * update rate.
 *******************************************************************************/
bool DEV_SyncGpsTime(GPS_Time * currentTime, bool forceSync){

	bool res = true;
	struct tm calendar;
	struct timeval time;

	calendar.tm_year 	= (int)currentTime->year - 1900;
	calendar.tm_mon 	= (int)currentTime->month - 1;
	calendar.tm_mday	= (int)currentTime->day;
	calendar.tm_hour 	= (int)currentTime->hour;
	calendar.tm_min 	= (int)currentTime->min;
	calendar.tm_sec 	= (int)currentTime->sec;
	calendar.tm_isdst	= -1;

	time.tv_sec = mktime(&calendar);
	time.tv_usec = currentTime->nano / 1000;

	SyncSystemTime(&time, forceSync);
	SyncTbrTime(&time, forceSync);

	return res;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void SyncSystemTime(struct timeval * time, bool force){

	static uint16_t timeUpdateCounter = 0;

	/* Update the epoch */
	if (!(timeUpdateCounter % UPDATE_SYSTEM_TIME_RATE) || force) settimeofday(time, NULL);
	timeUpdateCounter++;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void SyncTbrTime(struct timeval * time, bool force){

	static uint16_t timeUpdateCounter = 0;

	 /* Only sync on last digit = 0 */
	if (tbr700Enabled && !(time->tv_sec % 10)){

	#if DEBUG_OPERATION
		STREAM_Printf(stream_debug, "TBR time update\n");
	#endif

		if (!(timeUpdateCounter % UPDATE_TBR_CLOCK_RATE) || force) TBR_SyncTime(&time->tv_sec);
		else TBR_SyncTime(NULL);

		timeUpdateCounter++;
	}
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static uint16_t FormatGpsData(char * buf, GeoPosition * position){

	uint16_t nBytes;

	nBytes  = sprintf(buf, "UID:%lu,", *SYS_GetID());
	nBytes += SYS_FormatTimeStamp(buf + nBytes);

	nBytes += ee_sprintf(buf + nBytes, ",@%.5f,", (double)position->latitude / 10000000);
	nBytes += ee_sprintf(buf + nBytes, "%.5f,", (double)position->longitude / 10000000);
	nBytes += ee_sprintf(buf + nBytes, "%.1fz\r\n", (double)position->heightAboveMeanSea_mm / 1000);

	return nBytes;
}


/*******************************************************************************
 * @brief Formats selected dev_gauge data and writes it to the provided buffer.
 *******************************************************************************/
static uint16_t FormatGaugeData(char * buf){

	uint16_t nBytes;

	nBytes  = sprintf(buf, "UID:%lu,", *SYS_GetID());
	nBytes += SYS_FormatTimeStamp(buf + nBytes);

	nBytes += ee_sprintf(buf + nBytes, ",%.2fV,", (double)GAUGE_BatteryVoltage_mV() / 1000);
//	nBytes += sprintf(buf + nBytes, "%dmA,", GAUGE_InstCurrent_mA());
	nBytes += sprintf(buf + nBytes, "%dmA,", GAUGE_AverageCurrent_mA());
	nBytes += sprintf(buf + nBytes, "%umAh\r\n", GAUGE_RemainingCapacity_mAh());

	return nBytes;
}


/*******************************************************************************
 * @brief Update the system info with the last registered TBR detection data.
 * The values are saved to param file with the rate SAVE_TBR_DET_RATE, if
 * "force" not set. If set, the parameters will be saved independent on the rate.
 *******************************************************************************/
static void UpdateTbrDetections(bool force){

	static uint16_t updateCounter = 0;

	updateCounter++;

	SYS_UpdateSeaTemp(TBR_GetLastTempRead());
	SYS_AddTbrDetections(TBR_GetTagDetections());
	SYS_TbrFlashCode(TBR_GetLastFlashCode());

	if (!(updateCounter % SAVE_TBR_DET_RATE) || force){

		LOG_SaveMiscOpParam(param_lastReadTbrFlashCode);
		LOG_SaveMiscOpParam(param_tbrTagDetections);
		LOG_CloseOpParamFile();
	}
}


