/******************************************************************************
 * @file:		device_manager.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	4. feb. 2016
 ******************************************************************************
 *	Handles the direct communication with the different peripherals;
 *	initialization, sleep/wake up, configuration, data acquisition, etc.
 *	Also manages all of the GPIO interrupts, peripheral clocks,
 *	user configuration.
 ******************************************************************************/

#include <sys/time.h>

#include "em_int.h"

#include "system.h"
#include "timer.h"
#include "err.h"

#include "device_manager.h"
#include "device_config.h"
#include "device_handlers.h"
#include "device_settings.h"

#include "stream.h"
#include "log_manager.h"
#include "log_param.h"

#include "gauge_driver.h"
#include "radio_driver.h"
#include "aqua_driver.h"
#include "gsm_driver.h"
#include "gps_driver.h"
#include "spi_driver.h"
#include "imu_driver.h"
#include "tbr_driver.h"
#include "sd_driver.h"
#include "gsm_inet.h"
#include "file.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static void SetMiscUserConfigParam(LOG_UserParam_user paramType, char * param);
static void AttachFunctionPointers();
static void Enable12V(bool enable);
static void EnablePeripheralClock(Device_dev device, bool enable);
static void EnablePeripheralClocks(bool enable);
static void InterruptHandler_GPIO();

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

static DEV_Settings devices[] = {
		INIT_DEVICE_TBR,
		INIT_DEVICE_AQUA,
		{0}, /* INIT_DEVICE_WS */
		INIT_DEVICE_GPS,
		INIT_DEVICE_IMU,
		INIT_DEVICE_GAUGE,
		INIT_DEVICE_SD,
		INIT_DEVICE_RADIO,
		INIT_DEVICE_GSM
};

static bool gpsConfigFinished;
static bool gpsDataAv = false;
static bool gsmDataAv = false;

/* Set via configuration file parameter "First time" */
bool firstTimeOnly = false;

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Initializes all system clocks, device related pins, and GPIO
 * interrupts.
 *******************************************************************************/
bool DEV_Init(){

	/* Enable GPIO clock */
	CMU_ClockEnable(cmuClock_GPIO, true);
	/* Init power-related pins */

	GPIO_PinModeSet(V12_PWR_PORT, V12_PWR_PIN, gpioModePushPull, 0);	/* 12V enable */

	GPIO_PinModeSet(CH_FAULT_PORT, CH_FAULT_PIN, gpioModeInputPull, 1);	/* Charger fault pin */
	GPIO_PinModeSet(CH_CHRG_PORT, CH_CHRG_PIN, gpioModeInputPull, 1);	/* Charger charge pin */

	GPIO_PinModeSet(LED_PORT, LED_GR_PIN, gpioModePushPull, 0);
	GPIO_PinModeSet(LED_PORT, LED_YL_PIN, gpioModePushPull, 0);
	GPIO_PinModeSet(LED_PORT, LED_RED_PIN, gpioModePushPull, 0);

	/* Enable interrupts on charger pins
	 * Interrupt on both falling and rising edge */
	GPIO_IntConfig(CH_FAULT_PORT, CH_FAULT_PIN, true, true, true);
	GPIO_IntConfig(CH_CHRG_PORT, CH_CHRG_PIN, true, true, true);

	NVIC_EnableIRQ(GPIO_ODD_IRQn);
	NVIC_EnableIRQ(GPIO_EVEN_IRQn);

	/* Attach pointers to the DEV_Settings structures */
	AttachFunctionPointers();

	return true;
}


/*******************************************************************************
 * @brief Initializes all devices. Devices that do not respond will be disabled.
 *******************************************************************************/
bool DEV_InitAllDevices(){

	bool resOk = true;

	/* Enable all device-related clocks */
	EnablePeripheralClocks(true);
	Enable12V(true);
	SPI_Init();

	for (int i = 0; i < N_DEVICES; i++){

		if (~devices[i].flags & DEV_ACTIVE_MASK){

			if (!devices[i].Init){
				ERR_RAISE(err_connectionErr, i);
				resOk = false;
			}

			else if ((resOk = devices[i].Init())){

				/* Flag device active if init OK */
				devices[i].flags |= DEV_ACTIVE_MASK;

				/* Flag device as enabled in system info */
				if (devices[i].flags & DEV_ENABLED_MASK){
					SYS_SetEnabledDevice(i);
				}
			}

			/* In case init was not OK */
			if (!resOk) DEV_Error_Handler();
		}
	}

	return true;
}


/*******************************************************************************
 * @brief Collects data from the specified device and append it to the
 * current log file on the SD card.
 *******************************************************************************/
bool DEV_CollectData(Device_dev device){

	bool resOk = devices[device].CollectData ? devices[device].CollectData(): false;

	if (!resOk) DEV_Error_Handler();

	return resOk;
}


/*******************************************************************************
 * @brief Configure selected device with the defined configuration.
 * If device is disabled, configuration will be omitted.
 *******************************************************************************/
bool DEV_ConfigureDevice(Device_dev device){

	bool resOk = true;

	if (devices[device].flags & DEV_ENABLED_MASK){

		if (device == dev_gps) resOk = DEV_DefaultConfigGps(&devices[device]);

		else resOk = devices[device].Configure ? devices[device].Configure() : true;

		if (!resOk) DEV_Error_Handler();
	}

	return resOk;
}


/*******************************************************************************
 * @brief Reset selected device. Either hard or soft reset depending on device
 * and implementation.
 * Current settings in volatile memory may get lost for some devices.
 *******************************************************************************/
bool DEV_ResetDevice(Device_dev device){

	bool resOk = devices[device].Reset ? devices[device].Reset(): false;

//	if (!resOk) resOk = DEV_Error_Handler();

	return resOk;
}


/*******************************************************************************
 * @brief Wake up selected device. Depending on implementation, it may have
 * been in sleep mode, or completely shut off.
 *******************************************************************************/
bool DEV_WakeUpDevice(Device_dev device){

	bool resOk = true;

	/* Return true if device is already awake */
	if (devices[device].flags & DEV_ACTIVE_MASK) goto end;

	/* Return false if device is not enabled */
	else if (~devices[device].flags & DEV_ENABLED_MASK){
		resOk = false;
		goto end;
	}


	/* Enable device clock */
	EnablePeripheralClock(device, true);

	if (device == dev_tbr700 || device == dev_aquaTroll){
		Enable12V(true);
		Delay_ms(2);
	}

	/* Wake up device if function pointer registered */
	if (devices[device].WakeUp) resOk = devices[device].WakeUp();
	else resOk = false;

	if (resOk){
		/* flag device as active */
		devices[device].flags |= DEV_ACTIVE_MASK;

		/* Set "await data" flag if "use await data" flag is set,
		 * meaning that data device cannot be shut down before "await
		 * data" flag is cleared */
		if (devices[device].flags & DEV_USE_AWAIT_DATA_MASK){

			devices[device].flags |= DEV_AWAIT_DATA_MASK;
		}
	}

end:
	if (!resOk) DEV_Error_Handler();
	return resOk;
}


/*******************************************************************************
 * @brief Shut down or sleep selected device, depending on the device implements
 * a sleep mode. If not, a power switch shuts the device off.
 *******************************************************************************/
bool DEV_ShutdownDevice(Device_dev device){

	bool resOk = true;

	/* Return true if the device is enabled and the "always on" flag is set */
	if ((devices[device].flags & DEV_ENABLED_MASK) &&
			(devices[device].flags & DEV_ALWAYS_ON_MASK)) goto end;

	/* Return true if device has already been shut down */
	else if (~devices[device].flags & DEV_ACTIVE_MASK) goto end;

	/* Do not turn off device if awaiting data */
	else if (devices[device].flags & DEV_AWAIT_DATA_MASK){
		devices[device].flags |= DEV_SHDWN_MASK;
		goto end;
	}

	else devices[device].flags &= ~DEV_SHDWN_MASK;


	switch(device){

	case dev_sdCard:

		SD_PowerOff();

		/* Disable clock if this is the only active SPI device */
		if (DEV_ACTIVE_MASK & ~(
				devices[dev_gps].flags |
				devices[dev_imu].flags)){

			EnablePeripheralClock(device, false);
		}
		break;

	case dev_gps:

		GPS_PowerOff();

		/* Disable clock if this is the only active SPI device */
		if (DEV_ACTIVE_MASK & ~(
				devices[dev_imu].flags |
				devices[dev_sdCard].flags)){

			EnablePeripheralClock(device, false);
		}
		break;

	case dev_imu:
		resOk = IMU_EnterSleepMode();

		/* Disable clock if this is the only active SPI device,
		 * and if shutdown was OK */
		if (resOk && (
				DEV_ACTIVE_MASK & ~(
				devices[dev_gps].flags |
				devices[dev_sdCard].flags))){

			EnablePeripheralClock(device, false);
		}
		break;

	case dev_radio:
		resOk = RADIO_EnterSleepMode();
		break;

	case dev_tbr700:

		TBR_DisablePower(true);

		/* If Aqua troll is inactive, shut down 12V as well */
		if (~devices[dev_aquaTroll].flags &
				DEV_ACTIVE_MASK) Enable12V(false);
		break;

	case dev_aquaTroll:

		AQUA_PowerOff();

		/* If TBR 700 is inactive, shut down 12V as well */
		if (~devices[dev_tbr700].flags &
				DEV_ACTIVE_MASK) Enable12V(false);
		break;

	case dev_gsm:

		/* Disconnect server/client if active */
		switch (GSM_TcpStatus()){
		case tcp_client: GSM_DisconnectFromServer(); break;
		case tcp_server: GSM_CloseTCPServer(); break;
		default: break;
		}

		GSM_PowerOff();
		break;

	case dev_ws: /* not implemented */
		break;

	case dev_gauge: /* Disable i2c and GPIO pins. Device has auto sleep */

		GAUGE_DeInit();
		break;

	default:
		break;
	}
	/* Sleep or shutdown OK */
	if (resOk){
		/* flag device as inactive */
		devices[device].flags &= ~DEV_ACTIVE_MASK;

		/* Disable clock if the device is not on SPI */
		if (devices[device].clock != SPI_CMUCLOCK){
			EnablePeripheralClock(device, false);
		}
	}

end:
	if (!resOk) DEV_Error_Handler();
	return resOk;
}


/*******************************************************************************
 * @brief Check if data is available from the selected device, by reading the
 * TX ready / RI pin. Implemented for the GPS and the GSM/GPRS modules.
 *******************************************************************************/
bool DEV_DataAv(Device_dev device){

	bool res = false;

	/* Only regard if device is active */
	if (devices[device].flags & DEV_ENABLED_MASK){

		if (device == dev_gps &&
				(devices[device].flags & DEV_ACTIVE_MASK)){

			res = GPIO_PinInGet(GPS_GPIO_PORT, GPS_EXTINT_PIN) ? true : false;
		}


		else if (device == dev_gsm){

			if (GPIO_PinInGet(GSM_RI_PORT, GSM_RI_PIN)) res = gsmDataAv;
			else res = true;

			gsmDataAv = false;
		}
	}

	return res;
}


/*******************************************************************************
 * @brief Carries out a time synchronization of the system clock and the
 * TBR700 clock with the GPS UTC time.
 * Allows up to 60 seconds for the GPS to acquire a valid time.
 *******************************************************************************/
bool DEV_SyncTime(){

	GPS_PVTSolution sol;

	sol.time.valid = false;
	bool resOk = true;
	bool waitingForFix = true;
	int secWait = 0;

	resOk = DEV_WakeUpDevice(dev_gps);

	/* Allow up to 60 sec wait for position fix */
	while (waitingForFix){

		resOk = GPS_GetTimeUTC(&sol.time);

		if (sol.time.valid) waitingForFix = false;

		else {

			Sleep_ms(1000, true);
			secWait++;
		}

		if (secWait > TIMEOUT_GPS_FIX){

			ERR_RAISE(err_gpsFixTimeout, or_deviceManager);
			resOk = false;
			goto end;
		}
	}
	/* Clear await data flag so that GPS can be turned off */
	devices[dev_gps].flags &= ~DEV_AWAIT_DATA_MASK;

	resOk = DEV_SyncGpsTime(&sol.time, true);
	DEV_ShutdownDevice(dev_gps);

end:
	if (!resOk) DEV_Error_Handler();
	return resOk;
}


/*******************************************************************************
 * @brief Turn on or off one of the on-board LEDs.
 *******************************************************************************/
void DEV_SetLED(bool active, LED_Color_led color){

	switch(color){

	case led_red:
		if (active)
			GPIO_PinOutSet(LED_PORT, LED_RED_PIN);
		else
			GPIO_PinOutClear(LED_PORT, LED_RED_PIN);
		break;

	case led_yellow:
		if (active)
			GPIO_PinOutSet(LED_PORT, LED_YL_PIN);
		else
			GPIO_PinOutClear(LED_PORT, LED_YL_PIN);
		break;

	case led_green:
		if (active)
			GPIO_PinOutSet(LED_PORT, LED_GR_PIN);
		else
			GPIO_PinOutClear(LED_PORT, LED_GR_PIN);
		break;
	}
}


/*******************************************************************************
 * @brief Update the device settings specified in the DEV_Config structure.
 *******************************************************************************/
void DEV_SetDeviceConfig(DEV_Config * config){

	devices[config->device].interval_s = config->interval_s;
	devices[config->device].onPeriod_s = config->onPeriod_s;

	if (config->enable) 		devices[config->device].flags |= DEV_ENABLED_MASK;
	else 						devices[config->device].flags &= ~DEV_ENABLED_MASK;

	if (config->alwaysOn)		devices[config->device].flags |= DEV_ALWAYS_ON_MASK;
	else 						devices[config->device].flags &= ~DEV_ALWAYS_ON_MASK;

	if (config->setLightSleep){

		if (config->lightSleep)	devices[config->device].flags |= DEV_NO_DEEP_SLEEP_MASK;
		else 					devices[config->device].flags &= ~DEV_NO_DEEP_SLEEP_MASK;
	}
}


/********************************************************************************************
 * Returns the flags for the selected device.
 * Bit explanation:
 * bit7		bit6		bit5		bit4		bit3		bit2		bit1		bit0
 * 					  wakeupErr   isActive  noDeepSleep  dataOnWake   alwaysOn	  enabled
 *
 * Explanations:
 * wakeupErr:	Asserted when wake up of the device somehow failed.
 * TTA: 		Time to activate, meaning that the RTC has triggered on the defined
 * 				cycle time, and it is time to read data etc. from the device.
 * isActive:	Is asserted when the device is active, meaning it is not in sleep or powered
 * 				off.
 * enabled:		Defines whether the device is enabled (in use/available) on this
 * 				particular controller.
 ********************************************************************************************/
uint8_t * DEV_GetFlags(Device_dev device){

	return &devices[device].flags;
}


/*******************************************************************************
 * @brief Set the flags specified by mask.
 *******************************************************************************/
void DEV_SetFlags(Device_dev device, uint8_t mask){

	devices[device].flags |= mask;
}


/*******************************************************************************
 * @brief Clear the flags specified by mask.
 *******************************************************************************/
void DEV_ClearFlags(Device_dev device, uint8_t mask){

	devices[device].flags &= ~mask;
}


/*******************************************************************************
 * @brief Checks if there are any currently active devices that
 * does not allow deep sleep.
 *******************************************************************************/
bool DEV_LightSleepDeviceActive(){

	bool res = false;

	for (int i = 0; i < N_DEVICES; i++){

		/* Check if any device that do not allow deep sleep is active */
		if ((devices[i].flags & DEV_ACTIVE_MASK) &&
				(devices[i].flags & DEV_NO_DEEP_SLEEP_MASK)){

			res = true;
		}
	}
	return res;
}


/*******************************************************************************
 * @brief Accessible trough "device_settings.h".
 * @return true if device is enabled
 *******************************************************************************/
bool DEV_GetEnabledDevice(Device_dev device){

	return devices[device].flags & DEV_ENABLED_MASK;
}


/*******************************************************************************
 * @brief Accessible trough "device_settings.h".
 * @return the "on" time for the specified device.
 *******************************************************************************/
uint16_t * DEV_GetDeviceOnPeriod(Device_dev device){

	return &devices[device].onPeriod_s;
}


/*******************************************************************************
 * @brief Accessible trough "device_settings.h".
 * @return the "interval" time for the specified device.
 *******************************************************************************/
uint32_t * DEV_GetDeviceInterval(Device_dev device){

	return &devices[device].interval_s;
}


/*******************************************************************************
 * @brief Reads the user configuration file from storage and configures the
 * parameters set in the file.
 *******************************************************************************/
bool DEV_SetUserConfig(){

	static bool fileLoaded = false;

	bool res = true;
	LOG_UserConfig config;
	DEV_Config devConfig;


	if (!fileLoaded){

		fileLoaded = true;

		/* Allocate memory fo misc items */
		for (int i = 0; i < LOG_USER_CONFIG_MISC_ITEMS; i++){
			config.misc[i] = calloc(1, LOG_USER_CONFIG_MISC_ITEMS_LEN);

		}

		bool fileFound = LOG_LoadUserConfig(&config);

		if (fileFound){

			for (int i = 0; i < N_DEVICES; i++){

				if (config.configFound[i]){ /* Device operation configuration parameters */

					/* Set device configuration */
					devConfig.device = i;
					devConfig.enable = (bool)config.deviceOpConfig[0][i];
					devConfig.alwaysOn = (bool)config.deviceOpConfig[1][i];
					devConfig.setLightSleep = false;
					devConfig.interval_s = config.deviceOpConfig[2][i];
					devConfig.onPeriod_s = config.deviceOpConfig[3][i];

					DEV_SetDeviceConfig(&devConfig);
				}
			}

			for (int paramType = 0; paramType < LOG_USER_CONFIG_MISC_ITEMS; paramType++){

				if (config.configFound[paramType + N_DEVICES]){ /* Other configuration parameters */

					SetMiscUserConfigParam(paramType, config.misc[paramType]);
				}

				else if (paramType == user_apn || paramType == user_tcpPort){ /* APN and TCP port must be set */
					ERR_RAISE(err_majorFailure, or_deviceManager);
					res = false;
					break;
				}
			}
		}

		/* File not found */
		else {
			ERR_RAISE(err_majorFailure, or_deviceManager);
			res = false;
		}
	}
	return res;
}


/*******************************************************************************
 * @brief Handler for most device related errors that may be raised.
 *******************************************************************************/
void DEV_Error_Handler(){


	switch(error.erRes){

	case err_connectionErr:

		if (!DEV_ResetDevice(error.origin)){

		#if DEBUG_OPERATION
			STREAM_Printf(stream_debug,"%s: %s\n", ERROR_TO_STRING(0),
					DEVICE_TO_STRING((Device_dev)error.origin));
		#endif

			/* Set device faulty if no contact was established after reset */
			SYS_SetFaultyDevice(error.origin);
			DEV_ClearFlags(error.origin, DEV_ENABLED_MASK);
			DEV_SetFlags(error.origin, DEV_ACTIVE_MASK);

			DEV_ShutdownDevice(error.origin);
			DEV_ClearFlags(error.origin, DEV_ACTIVE_MASK);

			error.retVal = ret_resolved;
		}

		else {

			DEV_SetFlags(error.origin, DEV_ACTIVE_MASK);
			SYS_SetEnabledDevice(error.origin);

			error.erRes = ret_resolvedRetry; /* Error handled, retry operation */
		}

		break;

	case err_notEnabled:
		/* Device not enabled / available. Enable flag must
		 * be set to allow access.
		 */
		error.retVal = ret_resolved;
		break;

	case err_gpsFixTimeout:
		/* Make sure GPS is enabled */
		DEV_SetFlags(dev_gps, DEV_ENABLED_MASK);
		//DEV_ResetDevice(dev_gps);

		error.erRes = ret_resolved;
		break;

	/* Do nothing, bad data from GPS */
	case err_nakRecv:
	case err_badCheckSum:

		error.retVal = ret_resolved;
		break;

	case err_rxTimeout: /* Expected data from device not received within timeout */
	case err_dataNotValid: /* Data received from device not valid */

		if (error.origin == (ERR_ExceptionOrigin_or)dev_gps) error.retVal = ret_resolved;
		break;

	case err_sdCardNotDetected:

		/* Signal SD card not detected */
		DEV_SetLED(true, led_red);
		Sleep_ms(2000, true);
		DEV_SetLED(false, led_red);
		Sleep_ms(2000, true);

		/* Try to initialize file system */
		FILE_CloseAll();

		SD_PowerOff();

		if (error.errCount > 5){ /* Set SD card faulty */ // does not work (errcount)!
			SYS_SetFaultyDevice(dev_sdCard);
			DEV_ClearFlags(dev_sdCard, DEV_ENABLED_MASK);
			error.erRes = ret_resolved;
		}

		else {
			Sleep_ms(10, true);
			FILE_InitFileSystem();
			error.erRes = ret_resolvedRetry;
		}

		break;

	/* Cannot handle error -> return to higher level error handler */
	default:
		error.retVal = ret_notResolved;
		break;
	}

	if (error.retVal == ret_resolved) ERR_RESET();
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Set the specified user configuration parameter, which has been
 * read from user configuration file. The "param" memory not needed will be freed.
 *******************************************************************************/
static void SetMiscUserConfigParam(LOG_UserParam_user paramType, char * param){

	uint32_t value;

	switch (paramType){

	case user_sId:

		value = (uint32_t)atoi(param);
		RADIO_SetSID(&value);
		free(param);
		break;

	case user_uId:

		value = (uint32_t)atoi(param);
		RADIO_UniqueID(true, &value);
		free(param);
		break;

	case user_resetTbr: /* Clear TBR memory */

		value = atoi(param);

		if (value == 1) TBR_ClearMemory();

		free(param);
		LOG_ClearUserParam(user_resetTbr);
		break;

	case user_apn:
		GSM_SetApn(param);
		break;

	case user_simPin:
		GSM_SetPin(param);
		break;

	case user_firstTimeOnly:
		if (atoi(param) == 1) firstTimeOnly = true;
		free(param);
		LOG_ClearUserParam(user_firstTimeOnly);
		break;

	case user_tcpPort:
		GSM_SetStandardTcpPort(param);
		break;

	case user_remoteIp:
		// Set a standard remote IP (To for example a TCP server) to use.
		break;
	}

}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void Enable12V(bool enable){

	if (enable) GPIO_PinOutSet(V12_PWR_PORT, V12_PWR_PIN);
	else GPIO_PinOutClear(V12_PWR_PORT, V12_PWR_PIN);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void EnablePeripheralClock(Device_dev device, bool enable){

	CMU_ClockEnable(devices[device].clock , enable);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void EnablePeripheralClocks(bool enable){

	/* SPI, UART and I2C clocks */
	for (int i = 0; i < N_DEVICES; i++){

		CMU_ClockEnable(devices[i].clock , enable);
	}
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void AttachFunctionPointers(){

	devices[dev_radio].WakeUp =  	RADIO_ExitSleepMode;
	devices[dev_tbr700].WakeUp =  	TBR_PowerOn;
	devices[dev_aquaTroll].WakeUp = AQUA_PowerOn;
	devices[dev_gsm].WakeUp =  	 	GSM_PowerOn;
	devices[dev_gps].WakeUp =   	GPS_PowerOn;
	devices[dev_sdCard].WakeUp =	FILE_InitFileSystem;
	devices[dev_imu].WakeUp =   	IMU_ExitSleepMode;
	devices[dev_gauge].WakeUp =   	GAUGE_Init;
	devices[dev_ws].WakeUp =    	NULL;

	devices[dev_radio].Init =  		RADIO_Init;
	devices[dev_tbr700].Init =  	TBR_Init;
	devices[dev_aquaTroll].Init = 	AQUA_Init;
	devices[dev_gsm].Init =  	 	GSM_Init;
	devices[dev_gps].Init =   		GPS_Init;
	devices[dev_sdCard].Init =		FILE_InitFileSystem;
	devices[dev_imu].Init =   		IMU_Init;
	devices[dev_gauge].Init =   	GAUGE_Init;
	devices[dev_ws].Init =    	 	NULL;

	devices[dev_tbr700].CollectData =  		DEV_CollectTbrData;
	devices[dev_gps].CollectData =   		DEV_CollectGpsData;
	devices[dev_gauge].CollectData =   		DEV_CollectGaugeData;
	devices[dev_ws].CollectData =    	 	NULL;
	devices[dev_aquaTroll].CollectData = 	NULL;
	devices[dev_gsm].CollectData =  	 	NULL;
	devices[dev_sdCard].CollectData =		NULL;
	devices[dev_imu].CollectData =   		NULL;
	devices[dev_radio].CollectData =  		NULL;

	devices[dev_radio].Configure =  		DEV_DefaultConfigRadio;
	devices[dev_tbr700].Configure =  		DEV_DefaultConfigTbr700;
	devices[dev_gsm].Configure =  	 		DEV_DefaultConfigGsm;
	devices[dev_gps].Configure =   			DEV_DefaultConfigGps;
	devices[dev_gauge].Configure =   		DEV_DefaultConfigGauge;
	devices[dev_aquaTroll].Configure = 		NULL;
	devices[dev_ws].Configure =    	 		NULL;
	devices[dev_imu].Configure =   			NULL;
	devices[dev_sdCard].Configure =			NULL;

	devices[dev_radio].Reset =  		DEV_DefaultResetRadio;
	devices[dev_tbr700].Reset =  		DEV_DefaultResetTbr;
	devices[dev_gsm].Reset =  	 		DEV_DefaultResetGsm;
	devices[dev_gps].Reset =   			DEV_DefaultResetGps;
	devices[dev_sdCard].Reset =			DEV_DefaultResetSdCard;
	devices[dev_gauge].Reset =   		DEV_DefaultResetGauge;
	devices[dev_ws].Reset =    	 		NULL;
	devices[dev_imu].Reset =   			NULL;
	devices[dev_aquaTroll].Reset = 		NULL;
}


/*******************************************************************************
 * 							Interrupt handlers
 *******************************************************************************/


/*******************************************************************************
 * @brief Handler for all GPIO interrupts, both from odd and even pins.
 *******************************************************************************/
static void InterruptHandler_GPIO(){

	uint32_t flags = GPIO->IF;
	GPIO->IFC = flags;

	/* Radio CTS pin asserted or de-asserted */
	if (flags & (1 << RADIO_CTS_PIN)){

		/* Asserted */
		if (!GPIO_PinInGet(RADIO_CTS_PORT, RADIO_CTS_PIN)){

			/* Flag as clear to send */
			RADIO_SetCTS(true);
		}
		/* Flag as not clear to send */
		else RADIO_SetCTS(false);
	}

	/* Charger charge pin asserted or de-asserted */
	else if (flags & (1 << CH_CHRG_PIN)){

		/* Asserted */
		if (!GPIO_PinInGet(CH_CHRG_PORT, CH_CHRG_PIN)){

			SYS_SetFlags(SYS_CHARGING_MASK);
		}
		else SYS_ClearFlags(SYS_CHARGING_MASK);
	}

	/* Charger fault pin asserted or de-asserted */
	else if (flags & (1 << CH_FAULT_PIN)){

		/* Asserted */
		if (!GPIO_PinInGet(CH_CHRG_PORT, CH_CHRG_PIN)){

			SYS_SetFlags(SYS_CHRGERR_MASK);
		}
		else SYS_ClearFlags(SYS_CHRGERR_MASK);
	}

	/* GPS time pulse pin rising edge */
	else if (flags & (1 << GPS_TIMEPULSE_PIN)){

		/* For accurate time synchronization.
		 * Needs to be enabled in GPS driver */
	}

	/* Tx ready interrupt when feature enabled */
	else if (flags & (1 << GPS_EXTINT_PIN)){

		if (gpsConfigFinished){

			/* Asserted */
			if (GPIO_PinInGet(GPS_GPIO_PORT, GPS_EXTINT_PIN))
				gpsDataAv = true;

			else gpsDataAv = false;
		}
	}

	/* Incoming SMS/data/call through GPRS modem */
	else if (flags & (1 << GSM_RI_PIN)){

		if (!GPIO_PinInGet(GSM_RI_PORT, GSM_RI_PIN)) gsmDataAv = true;
	}

	/* GPRS modem CTS pin asserted or de-asserted */
	else if (flags & (1 << GSM_CTS_PIN)){

		if (!GPIO_PinInGet(GSM_CTS_PORT, GSM_CTS_PIN)){

			/* Flag as clear to send */
			GSM_SetCTS(true);
		}

		/* Flag as not clear to send */
		else GSM_SetCTS(false);
	}
}


void GPIO_ODD_IRQHandler(){

	INT_Disable();
	InterruptHandler_GPIO();
	INT_Enable();
}


void GPIO_EVEN_IRQHandler(){

	INT_Disable();
	InterruptHandler_GPIO();
	INT_Enable();
}
