/******************************************************************************
 * @file:		fifo_rx.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include "fifo_rx.h"
#include "fifo_config.h"


static uint8_t 		rxDataAvFlags = 0;

static uint8_t 		RxQueueRadio[FIFO_RX_SIZE_RADIO];
static uint8_t 		RxQueueTbr[FIFO_RX_SIZE_TBR];
static uint8_t 		RxQueueAqua[FIFO_RX_SIZE_AQUA];
static uint8_t 		RxQueueWs[FIFO_RX_SIZE_WS];
static uint8_t 		RxQueueGprs[FIFO_RX_SIZE_GSM];

static uint16_t 	FifoRxSize[N_STREAM_DEVICES];

static uint8_t * 	RxQueuePtr[N_STREAM_DEVICES];

volatile static int RxQueueInPos[N_STREAM_DEVICES] = {0};
volatile static int RxQueueOutPos[N_STREAM_DEVICES] = {0};

static FIFO_RtsCallback RtsCallback[N_STREAM_DEVICES];


/*******************************************************************************
 * @brief
 *******************************************************************************/
void FIFO_RX_Init(){

	RxQueuePtr[0] = &RxQueueRadio[0];
	RxQueuePtr[1] = &RxQueueTbr[0];
	RxQueuePtr[2] = &RxQueueAqua[0];
	RxQueuePtr[3] = &RxQueueWs[0];
	RxQueuePtr[4] = &RxQueueGprs[0];

	FifoRxSize[0] = FIFO_RX_SIZE_RADIO;
	FifoRxSize[1] = FIFO_RX_SIZE_TBR;
	FifoRxSize[2] = FIFO_RX_SIZE_AQUA;
	FifoRxSize[3] = FIFO_RX_SIZE_WS;
	FifoRxSize[4] = FIFO_RX_SIZE_GSM;
}


/*******************************************************************************
 * @brief Register a RTS callback function to be called when the FIFO buffer
 * becomes ready to receive data.
 *******************************************************************************/
void FIFO_RX_SetRtsCallBack(Stream stream, FIFO_RtsCallback callBack){

	RtsCallback[stream] = callBack;
}


/*******************************************************************************
 * @brief Reset the specified stream FIFO buffer.
 *******************************************************************************/
void FIFO_RX_Reset(Stream stream){

	RxQueueInPos[stream] = 0;
	RxQueueOutPos[stream] = 0;

	/* Clear data available flag */
	rxDataAvFlags &= ~(1 << stream);
}


/*******************************************************************************
 * @brief
 * @return true if the buffer is empty, otherwise false.
 *******************************************************************************/
bool FIFO_RX_Empty(Stream stream){

	return (RxQueueInPos[stream] == RxQueueOutPos[stream]);
}


/*******************************************************************************
 * @brief
 * @return true if the buffer is full, otherwise false.
 *******************************************************************************/
bool FIFO_RX_Full(Stream stream){

	return (RxQueueInPos[stream] ==
    		((RxQueueOutPos[stream] - 1 + FifoRxSize[stream]) % FifoRxSize[stream]));
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool FIFO_RX_Put(Stream stream, uint8_t data){

	bool res = true;

    if (RxQueueInPos[stream] ==
    		((RxQueueOutPos[stream] - 1 + FifoRxSize[stream]) % FifoRxSize[stream])){

    	res = false; /* Queue Full */
    }
    else {

    	*(RxQueuePtr[stream] + RxQueueInPos[stream]) = data;

    	RxQueueInPos[stream] = (RxQueueInPos[stream] + 1) % FifoRxSize[stream];

    	/* Set data available flag */
    	rxDataAvFlags |= 1 << stream;
    }

    return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool FIFO_RX_Get(Stream stream, uint8_t * data){

	bool res = true;

    if (RxQueueInPos[stream] == RxQueueOutPos[stream]){

        res = false; /* Queue empty */

        /* Clear data available flag */
        rxDataAvFlags &= ~(1 << stream);

        /* Callback function to set RTS for the specified stream,
         * if activated/supported */
        if (RtsCallback[stream]) RtsCallback[stream](true);

    }
    else {

    	*data = *(RxQueuePtr[stream] + RxQueueOutPos[stream]);

    	RxQueueOutPos[stream] = (RxQueueOutPos[stream] + 1) % FifoRxSize[stream];

    	if (RxQueueInPos[stream] == RxQueueOutPos[stream]){

    		/* Queue empty -> clear data available flag */
    		rxDataAvFlags &= ~(1 << stream);
    	}
    }
    return res;
}
