/******************************************************************************
 * @file:		file.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	The file module act as a interface to the FatFs module. Basic functionality
 *	as changing directories, creating, opening and closing of files is
 *	implemented. In addition, the module enables several files to be open
 *	simultaneously, by storing the file information in static FIL structures,
 *	and the file names in static char arrays.
 *	This allows for fast access to files when reading and writing from / to files.
 *	The module also enables stream functionality to -and from the files.
 *
 *	The current implementation depends on all file names being unique, even if
 *	they are placed in different directories. The file names must also be kept
 *	less than 13 characters, as specified in the FatFs module.
 *
 *	An error handler is included in the module, which transform the FRESULT
 *	enumeration to the ERR_Result_err defined in err.h.
 ******************************************************************************/

#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>

#include "file.h"
#include "timer.h"
#include "err.h"

/*******************************************************************************
 * 							Private variable declarations
 *******************************************************************************/

static FATFS 		sdFatSystem;	 /* file system object */
static const char * currentDirectory;

static FIL 			fileObj[MAX_OPEN_FILES];
static char  		openFiles[MAX_OPEN_FILES][FILENAME_MAX_SIZE];

static uint8_t 		openFilesCounter = 0;

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static FIL * 	GetNewFileObj(const char * fileName);
static bool 	ReleaseFileObj(const char * fileName);
static FIL * 	IdentifyFileObj(const char * fileName);
static void 	Error_Handler(FRESULT fileErr);

/******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *   This function is required by the FAT file system in order to provide
 *   timestamps for created files.
 *
 *   Refer to drivers/fatfs/doc/en/fattime.html for the format of this DWORD.
 * @return
 *    A DWORD containing the current time and date as a packed datastructure.
 ******************************************************************************/
uint32_t get_fattime(){

	uint32_t fattime;
	struct timeval currentTime;
	gettimeofday(&currentTime, NULL);	/* Get current time */

	struct tm * calendar = localtime(&currentTime.tv_sec); /* Get calendar format of current time */

	fattime = 	((calendar->tm_year - 80) << 25) |
				((calendar->tm_mon + 1) << 21) |
				(calendar->tm_mday << 16) |
				(calendar->tm_hour << 11) |
				(calendar->tm_min << 5) |
				(calendar->tm_sec / 2);

	return fattime;
}


/*******************************************************************************
 * @brief Initializes the specified file system. Returns false if the
 * file system could not be found /SD card not detected.
 * Power and SPI controlled in the lower layers.
 *******************************************************************************/
bool FILE_InitFileSystem(){

	TCHAR drive = 0;
	/* Initialize file system */
	f_mount(&sdFatSystem, &drive, 0);

	/* Create/open test file to test communication */
	FIL * obj = FILE_OpenNew("c_test.txt");

	currentDirectory = NULL; /* Reset current directory on power on */

	return obj ? FILE_Close("c_test.txt") : false;
}


/*******************************************************************************
 * @brief Get number of MB free storage on the SD card.
 * @return Number of MB.
 *******************************************************************************/
int16_t FILE_GetFree_MB(){

	FATFS * fs;
	FRESULT fRes;
	DWORD freeClust, freeSect/*, totSect*/;
	int16_t freeCap;

	fs = &sdFatSystem;
	fRes = f_getfree(&freeClust, &fs);

	if (fRes){
		Error_Handler(fRes);
		freeCap = -1;
		goto end;
	}

	//totSect = (fs->n_fatent - 2) * fs->csize;
	freeSect = freeClust * fs->csize;

	/* number of free sectors * 512 bytes per sector / number of bytes per MB (2^20 -1) */
	freeCap = (freeSect * 512) / 1048575;

end:
	return freeCap;
}


/*******************************************************************************
 * @brief
 * @return Number of currently open files.
 *******************************************************************************/
uint8_t FILE_GetNOpenFiles(){

	return openFilesCounter;
}


/*******************************************************************************
 * @brief Delete the specified file. The file will be closed if already open.
 *******************************************************************************/
bool FILE_Delete(const char * fName){

	FRESULT fRes;
	FIL * fObj = IdentifyFileObj(fName);

	if (fObj){
		/* Close the file if open */
		f_close(fileObj);
		ReleaseFileObj(fName);
	}

	/* Delete file */
	fRes = f_unlink(fName);

	if (fRes) Error_Handler(fRes);
	return !fRes;
}


/*******************************************************************************
 * @brief Opens read/write access on the specified file. If it does not already
 * exist, a new file will be created.
 * @return the FIL object structure pointer, or a null pointer if
 * something went wrong.
 *******************************************************************************/
FIL * FILE_OpenNew(const char * fName){

	FRESULT fRes = FR_OK;

	/* Check if file is already open */
	 FIL * fileObjPtr = IdentifyFileObj(fName);

	if (!fileObjPtr){

		fileObjPtr = GetNewFileObj(fName);

		if (fileObjPtr) fRes = f_open(fileObjPtr, fName,
					FA_OPEN_ALWAYS | FA_WRITE | FA_READ);

		if (fRes){
			Error_Handler(fRes);
			ReleaseFileObj(fName);
		}
	}

	return fRes ? NULL : fileObjPtr;
}


/*******************************************************************************
 * @brief Opens read/write access on the specified file.
 * @return the FIL object structure pointer, or a null pointer if
 * something went wrong.
 *******************************************************************************/
FIL * FILE_OpenExisting(const char * fName){

	FRESULT fRes = FR_OK;

	/* Check if file is already open */
	FIL * fileObjPtr = IdentifyFileObj(fName);

	if (!fileObjPtr){

		fileObjPtr = GetNewFileObj(fName);

		if(fileObjPtr) fRes = f_open(fileObjPtr, fName,
				FA_OPEN_EXISTING | FA_READ | FA_WRITE);


		if (fRes){
			Error_Handler(fRes);
			ReleaseFileObj(fName);
		}
	}

	return fRes ? NULL : fileObjPtr;
}


/*******************************************************************************
 * @brief Opens read/write access on the specified file, and moves the
 * file pointer to the end of the file. If file does not exist,
 * a new file will be created.
 * @return the FIL object structure pointer, or a null pointer if
 * something went wrong.
 *******************************************************************************/
FIL * FILE_OpenAppend(const char * fName){

	FRESULT fRes = FR_OK;

	/* Check if file is already open */
	FIL * fileObjPtr = IdentifyFileObj(fName);

	if (!fileObjPtr){

		fileObjPtr = GetNewFileObj(fName);

		if (fileObjPtr) fRes = f_open(fileObjPtr, fName,
				FA_WRITE | FA_READ | FA_OPEN_ALWAYS);
	}

	/* Move file pointer to end of file */
	if (!fRes && !f_eof(fileObjPtr)) fRes = f_lseek(fileObjPtr, f_size(fileObjPtr));


	if (fRes){
		Error_Handler(fRes);
		ReleaseFileObj(fName);
	}

	return fRes ? NULL : fileObjPtr;
}


/*******************************************************************************
 * @brief Opens the provided file and searches for the string pattern.
 * Returns a pointer to the next char after the found string pattern in the buffer.
 * The file pointer points to corresponding place in file.
 *******************************************************************************/
char * FILE_FindString(
		FIL * fObj,
		const char * string,
		char * buf,
		int bufLen,
		uint32_t * filePointer){

	bool res = true;
	bool stringFound = false;
	char * stringPointer = NULL;

	fObj->fptr = 0;

	StartTimer_ms(50);

	while (!TimeElapsed() && !stringFound && !f_eof(fObj)){

		/* Reached end of file */
		if (fObj->fsize == fObj->fptr) break;

		/* Return file pointer to start of read string */
		if (filePointer) *filePointer = fObj->fptr;

		/* Read a string of max length "FINDSTRING_LEN" into buffer */
		if (res) res = FILE_ReadString(fObj, buf, bufLen);

		/* Search for provided string pattern in the buffer */
		if (res) stringPointer = strstr(buf, string);

		/* String found */
		if (stringPointer){
			stringFound = true;

			/* Set pointer to end of found string */
			stringPointer += strlen(string);
			if(filePointer) *filePointer += strlen(string);
		}
	}

	return stringPointer;
}


/*******************************************************************************
 * @brief Closes the specified file.
 *******************************************************************************/
bool FILE_Close(const char * fileName){

	FRESULT fRes = FR_OK;

	if (fileName){

		FIL * fileObj = IdentifyFileObj(fileName);

		if (fileObj){

			/* Close the file */
			fRes = f_close(fileObj);

			if (fRes) Error_Handler(fRes);

			ReleaseFileObj(fileName);
		}
	}
	return !fRes;
}


/*******************************************************************************
 * @brief Closes all open files.
 *******************************************************************************/
bool FILE_CloseAll(){

	for (int i = 0; i < MAX_OPEN_FILES; i++){
		FILE_Close(openFiles[i]);
	}

	return openFilesCounter == 0;
}


/*******************************************************************************
 * @brief Changes the current directory.
 *******************************************************************************/
bool FILE_ChangeDir(const char * dirName){

	FRESULT fRes = FR_OK;

	if (currentDirectory != dirName){

		fRes = f_chdir(dirName);

		if (fRes) Error_Handler(fRes);

		else currentDirectory = dirName;

	}
	return !fRes;
}


/*******************************************************************************
 * @brief If directory does not exist, it will be created.
 *******************************************************************************/
bool FILE_CreateDir(const char * dirName){


	FRESULT fRes = f_mkdir(dirName);

	if (fRes && fRes != FR_EXIST) Error_Handler(fRes);

	return fRes ? fRes == FR_EXIST : true;
}


/*******************************************************************************
 * @brief Opens the specified directory.
 *******************************************************************************/
bool FILE_OpenDir(
		DIR * dirObj,
		const char * dirName){

	FRESULT fRes = f_opendir(dirObj, dirName);

	if (fRes) Error_Handler(fRes);
	return !fRes;
}


/*******************************************************************************
 * @brief Closes the specified directory.
 *******************************************************************************/
bool FILE_CloseDir(DIR * dirObj){

	FRESULT fRes = f_closedir(dirObj);

	if (fRes) Error_Handler(fRes);
	return !fRes;
}


/*******************************************************************************
 * @brief Transfers the specified file to a stream using the provided function
 * streamFunc. The file to transfer must already be open.
 *******************************************************************************/
bool FILE_TransferToStream(
		FIL * fObj,
		FILE_OutStreamFunc streamFunc,
		uint nBytesToRead){

	FRESULT fRes = FR_OK;
	uint	bytesSent;


	if (streamFunc && !f_eof(fObj)){ /* Only initiate transfer if file pointer
									  * is not at the end of file */

		/* Forward data to the defined function */
		fRes = f_forward(fObj, streamFunc, nBytesToRead, &bytesSent);
	}

	if (fRes) Error_Handler(fRes);
	return !fRes;
}


/*******************************************************************************
 * @brief  Writes the incoming data stream to the specified file,
 * using the provided function streamFunc. The file must already be open.
 * Temporarily stores data in a buffer before writing to file.
 *******************************************************************************/
bool FILE_WriteFromStream(
		FIL * fObj,
		FILE_InStreamFunc streamFunc){

	FRESULT fRes;
	bool res = true, transFinished = false;
	uint8_t recvBytes[FILE_STREAM_BUF_LEN];
	uint8_t dataAvailable = 0;
	uint nBytesWritten, nBytesRead = 0;

	if (!streamFunc) goto exit;


	while (!transFinished){

		StartTimer_ms(FILE_STREAM_RX_TIMEOUT_MS);

		/* Wait for data to be available */
		while (!dataAvailable){

			/* Timeout error if data has not become
			 * available within the specified time */
			if (TimeElapsed()){

				res = false;
				ERR_RAISE(err_rxTimeout, or_file);
				goto exit; /* Unwritten buffered data from the
				 	 	 	*current message will be lost! */
			}

			/* Poll to check if data is available in stream */
			dataAvailable = streamFunc(&recvBytes[nBytesRead], 1);
		}

		while (dataAvailable && !transFinished){

			/* Receive data from stream */
			dataAvailable = streamFunc(&recvBytes[nBytesRead], 0);

			/* Signals that the "end of transmission"
			 * char has been received */
			if (dataAvailable == 0xFF){
				transFinished = true;
			}


			if (dataAvailable){

				if (dataAvailable != 0xFF) nBytesRead++;

				/* Write data to file when buffer is full, or transmission has ended */
				if (nBytesRead >= FILE_STREAM_BUF_LEN ||
						dataAvailable == 0xFF){

					fRes = f_write(fObj, recvBytes, nBytesRead, &nBytesWritten);

					if (fRes){

						Error_Handler(fRes);
						res = false;
						goto exit;
					}

					else if (nBytesWritten != nBytesRead){

						res = false;
						ERR_RAISE(err_storageFull, or_file);
						goto exit;
					}

					else nBytesRead = 0;
				}
			}
		}
	}
exit:
	return res;
}


/*******************************************************************************
 * @brief Writes the provided buffer to the specified file.
 *******************************************************************************/
bool FILE_Write(
		FIL *fObj,
		const char * buffer,
		uint nBytesToWrite){

	bool res = true;
	uint bw;
	FRESULT fRes;

	/* Write a buffer to file */
	fRes = f_write(fObj, buffer, nBytesToWrite, &bw);

	if (fRes){
		Error_Handler(fRes);
		res = false;
	}

	else if (nBytesToWrite != bw){
		res = false;
		ERR_RAISE(err_storageFull, or_file);

	}
//	f_truncate(fObj);
	return res;
}


/*******************************************************************************
 * @brief Reads the specified number of bytes from the file.
 *******************************************************************************/
bool FILE_Read(
		FIL *fObj,
		char * buffer,
		uint nBytesToRead){

	uint br;
	FRESULT fRes;

	/* Read some data from file */
	fRes = f_read(fObj, buffer,  nBytesToRead, &br);

	if (fRes) Error_Handler(fRes);
	return !fRes;
}


/*******************************************************************************
 * @brief Reads the specified number of bytes from the file. Returns on line
 * shift (\n) or when it reaches the buffer length. Null terminates the buffer.
 *******************************************************************************/
bool FILE_ReadString(
		FIL *fObj,
		char * buffer,
		int bufferLen){

	bool res = true;

	/* Read data from file */
	if (!f_gets(buffer, bufferLen, fObj) && fObj->err){
		/* Some error occurred */
		Error_Handler(fObj->err);
		res = false;
	}
	return res;
}


/*******************************************************************************
 * @brief Moves to the specified position in the file.
 *******************************************************************************/
bool FILE_MoveToPosition(
		FIL * fObj,
		uint32_t index){

	FRESULT fRes = FR_OK;

	if (fObj->fptr != index) fRes = f_lseek(fObj, index);

	if (fRes) Error_Handler(fRes);
	return !fRes;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 * @return FIL structure pointer. If there are too many open files,
 * a file will be closed and the memory will be given to the new file.
 *******************************************************************************/
static FIL * GetNewFileObj(const char * fileName){

	bool allocOk = false;
	FIL * objPtr;

	if (openFilesCounter < MAX_OPEN_FILES){

		for (int i = 0; i < MAX_OPEN_FILES; i++){

			if (!openFiles[i][0]){

				allocOk = true;
				objPtr = &fileObj[i];

				strcpy(&openFiles[i][0], fileName);	/* Copy file name into new memory location */

				openFilesCounter++;

				break;
			}
		}
	}
	/* Too many open files */
	if (!allocOk){

		if (openFilesCounter > 1){

			/* Close an already open file to allocate its memory space */
			FILE_Close(openFiles[MAX_OPEN_FILES - openFilesCounter]);
			objPtr = GetNewFileObj(fileName);
		}
		else objPtr = NULL;
	}

	return objPtr;
}


/*******************************************************************************
 * @brief Releases the FIL object structure connected to the specified file name.
 *******************************************************************************/
static bool ReleaseFileObj(const char * fileName){

	bool fileFound = false;

	for (int i = 0; i < MAX_OPEN_FILES; i++){

		if (openFiles[i][0]){

			if (!strcmp(&openFiles[i][0], fileName)){

				fileFound = true;
				openFiles[i][0] = 0;

				if (openFilesCounter > 0) openFilesCounter--;
				break;
			}
		}
	}
	return fileFound;
}


/*******************************************************************************
 * @brief Compares the file name to currently open files. Returns the FIL
 * object structure if the file is open.
 *******************************************************************************/
static FIL * IdentifyFileObj(const char * fileName){


	for (int i = 0; i < MAX_OPEN_FILES; i++){

		if (openFiles[i][0]){

			if (!strcmp(&openFiles[i][0], fileName)){

				/* File found */
				return &fileObj[i];
			}
		}
	}
	return NULL;
}


/*******************************************************************************
 * @brief Analyzes the file error and raises a new error through ERR_RAISE.
 *******************************************************************************/
static void Error_Handler(FRESULT fileErr){

	switch(fileErr){

	case FR_DENIED:
	case FR_MKFS_ABORTED:
		ERR_RAISE(err_storageFull, or_file);
		break;

	case FR_NO_FILESYSTEM:
	case FR_INVALID_DRIVE:
	case FR_NOT_READY:
	case FR_WRITE_PROTECTED:
	case FR_INVALID_OBJECT:
	case FR_DISK_ERR:
		ERR_RAISE(err_sdCardNotDetected, or_file);
		break;

	case FR_INVALID_NAME:
	case FR_INVALID_PARAMETER:
	case FR_NO_PATH:
		ERR_RAISE(err_badFileName, or_file);
		break;

	case FR_NO_FILE:
		ERR_RAISE(err_noFile, or_file);
		break;

	case FR_NOT_ENABLED:
		break;

	case FR_INT_ERR:
		break;

	default:
		break;
	}
}
