/******************************************************************************
 * @file:		gsm_sms.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	24. apr. 2016
 ******************************************************************************
 *	Handler for all SMS related communicaton with the GSM/GPRS module.
 ******************************************************************************/

#include <string.h>

#include "system.h"
#include "err.h"
#include "timer.h"

#include "gsm_sms.h"
#include "gsm_driver.h"

#include "stream.h"

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Send SMS to specified number with the specified message.
 * @return true if SMS sent, otherwise false. Implementation does not wait for
 * the SMS Ack since it takes up to several seconds!
 *******************************************************************************/
bool GSM_SendSMS(char * number, const char * message){

	char end[2] = {0x1A, 0}; /* Ctrl+Z = Send GSM */

	/* Transfer number of receiver */
	int fRes = STREAM_PrintfWaitResp(stream_gsm, 2000, ">", "AT+CMGS=\"%s\"\r", number);

	/* Send SMS text. Does not wait for "+CMGS" since it blocks for several seconds until
	 * the SMS is transmitted */
	STREAM_Printf(stream_gsm, "%s%s\r", message, end);
	Delay_ms(100);
	//if (fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 7000, "+CMGS", "%s%s\r", message, end);

	STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_gsm);

	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Send SMS to the number stored in the specified phone book position.
 * @return = 1: OK.
 * return = 0: No number in phone book position.
 * return = -1: An error occurred.
 *******************************************************************************/
int GSM_Send_SMS_PhB(uint8_t phonebookPos, const char * message){

	char simPhoneNumber[GSM_PHONE_NUMBER_LEN];

	int fRes = GSM_GetPhoneNumber(phonebookPos, simPhoneNumber);

	if (fRes > 0) fRes = GSM_SendSMS(simPhoneNumber, message) == true ? 1 : -1;

	return fRes;
}


/*******************************************************************************
 * @brief Send short system status SMS to the number stored in the
 * specified phone book position.
 * @return = 1: OK.
 * return = 0: No number in phone book position.
 * return = -1: An error occurred.
 *******************************************************************************/
int GSM_Send_SysInfo(uint8_t phonebookPos, const char * phoneNumber){

	int fRes;
	char end[2] = {0x1A, 0}; /* Ctrl+Z = Send GSM */
	char simPhoneNumber[GSM_PHONE_NUMBER_LEN];
	const char * phoneNumberPtr;


	if (phoneNumber) phoneNumberPtr = phoneNumber;

	else {
		fRes = GSM_GetPhoneNumber(phonebookPos, simPhoneNumber);
		phoneNumberPtr = &simPhoneNumber[0];
	}

	/* Transfer number of receiver */
	if (fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, ">", "AT+CMGS=\"%s\"\r", phoneNumberPtr);

	/* Send system info  */
	if (fRes > 0) SYS_PrintShortSysInfo(stream_gsm);

	/* Does not wait for "+CMGS" since it blocks for several seconds until
	 * the SMS is transmitted */
	STREAM_Printf(stream_gsm, "%s\r", end);
	Delay_ms(100);
	//if (fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "+CMGS", "%s\r", end);

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_gsm);

	return fRes;
}


/*******************************************************************************
 * @brief Check if specified SMS is available.
 * @return > 0: valid position found.
 * return = 0: No GSM found.
 * return < 0: Something went wrong.
 *******************************************************************************/
int GSM_IsSMSPresent(GSM_SMSItem item){

	bool res;
	int smsPos;
	int fRes = 0;

	switch (item){

    case item_unread:
    	res = STREAM_Printf(stream_gsm, "AT+CMGL=\"REC UNREAD\"\r");
    	break;

    case item_read:

    	res = STREAM_Printf(stream_gsm, "AT+CMGL=\"REC READ\"\r");
    	break;

    case item_all:
    	res = STREAM_Printf(stream_gsm, "AT+CMGL=\"ALL\"\r");
    	break;
	}

	/* Response format;
	 * +CMGL: <index>,<stat>,<oa/da>,,[,<tooa/toda>,<length>]
     * <CR><LF> <data> <CR><LF>OK<CR><LF>
	 */
	if (res) fRes = STREAM_WaitAltResponse(stream_gsm, 2000, 2, "OK", "+CMGL:");

	if (fRes == 1) smsPos = 0;

	else if (fRes == 2){

		/* Get SMS position in SIM memory */
		smsPos = STREAM_GetValue(stream_gsm, 1000);
		STREAM_WaitAltResponse(stream_gsm, 1000, 1, "OK");
	}
	else smsPos = -1;
	
	return smsPos;
}


/*******************************************************************************
 * @brief Read SMS from specified position in memory.
 * @return SMS result.
 *******************************************************************************/
GSM_SMSRes_sr GSM_GetSMS(
		int position,
		char * phoneNumber,
		char * smsText,
		int maxSmsLen){


	GSM_SMSRes_sr res;
	int fRes = 0;


	STREAM_Printf(stream_gsm, "AT+CMGR=%d\r", position);

	/* SMS response format:
	 * <CR><LF>+CMGR: "REC UNREAD","+XXXXXXXXXXXX",,
	 * "02/03/18,09:54:28+40"<CR><LF>GSM text<CR><LF>OK<CR><LF>
	 */

	/* Wait for either "OK" or "+CMGR" */
	fRes = STREAM_WaitAltResponse(stream_gsm, 3000, 2, "OK", "+CMGR:");

	if (fRes == 1) res = sr_noSms;

	else if (fRes == 2){ /* GSM exist */

		fRes = STREAM_WaitAltResponse(stream_gsm, 1000, 2, "\"REC UNREAD\"", "\"REC READ\"");


		if (fRes == 1) res = sr_unread;
		else if (fRes == 2) res = sr_read;
		else if (!fRes) res = sr_other;
		else res = sr_err;

		if (fRes > 0){

			/* Read phone number */
			fRes = STREAM_GetString(stream_gsm, 1000, '"', '"', phoneNumber, GSM_PHONE_NUMBER_LEN);

			/* Read SMS message */
			if (fRes > 0) fRes = STREAM_GetString(stream_gsm, 1000, '\n', '\r', smsText, maxSmsLen);

			STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");

			if (fRes < 1) res = sr_err;
		}
	}
	else res = sr_err;

	return res;
}


/*******************************************************************************
 * @brief Reads the SMS in the provided GSM position and compares it to the
 * GSM commands stored in memory.
 * @return GSM query result.
 *******************************************************************************/
GSM_Query_sq GSM_ReadSMS(uint8_t position, char * phoneNumber){


	int fRes;
	GSM_Query_sq res = sq_error;


	STREAM_Printf(stream_gsm, "AT+CMGR=%u\r", position);

	/* SMS response format:
	 * <CR><LF>+CMGR: "REC UNREAD","+XXXXXXXXXXXX",,
	 * "02/03/18,09:54:28+40"<CR><LF>GSM text<CR><LF>OK<CR><LF>
	 */

	/* Wait for either "OK" or "+CMGR" */
	fRes = STREAM_WaitAltResponse(stream_gsm, 3000, 2, "OK", "+CMGR:");

	if (fRes == 2){ /* SMS exist */

		fRes = STREAM_WaitAltResponse(stream_gsm, 1000, 2, "\"REC UNREAD\"", "\"REC READ\"");

		if (fRes > 0){

			/* Read phone number */
			fRes = STREAM_GetString(stream_gsm, 1000, '"', '"', phoneNumber, GSM_PHONE_NUMBER_LEN);

			/* Read SMS message */
			if (fRes){
				fRes = STREAM_WaitAltResponse(
						stream_gsm, 1000, 7, GSM_CMD_Register, GSM_CMD_Unreg, GSM_CMD_Connect,
						GSM_CMD_StartTCPServer, GSM_CMD_Info, GSM_CMD_GetStatus, GSM_CMD_StopTCP);

				res = GSM_SMS_MSG_TO_QUERY(fRes);

				if (res != sq_connectTCP) STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");
			}
		}
	}
	else if (fRes == 1) res = sq_noSms;

	return res;
}


/*******************************************************************************
 * @brief Delete SMS at specified position in memory.
 * @return true if OK, false if something went wrong.
 *******************************************************************************/
bool GSM_DeleteSMS(uint8_t position){

	int fRes = STREAM_PrintfWaitResp(
			stream_gsm, 5000, "OK", "AT+CMGD=%u\r", position);

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_gsm);

	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Return the phone number from the provided position.
 * @return = 1: OK.
 * return 0: Number not found.
 * return < 0: Something went wrong.
 *******************************************************************************/
int GSM_GetPhoneNumber(uint8_t position, char * phoneNumber){

	int fRes = 0;
	// response in case valid phone number stored:
	// <CR><LF>+CPBR: <index>,<number>,<type>,<text><CR><LF>
	// <CR><LF>OK<CR><LF>

	// response in case there is not phone number:
	// <CR><LF>OK<CR><LF>

	STREAM_Printf(stream_gsm, "AT+CPBR=%u\r", position);

	fRes = STREAM_WaitAltResponse(
			stream_gsm, 2000, 2, "OK", "+CPBR");


	/* Phone number found */
	if (fRes == 2){

		fRes = STREAM_GetString(
				stream_gsm, 2000, '"', '"', phoneNumber, GSM_PHONE_NUMBER_LEN);

		STREAM_WaitAltResponse(
					stream_gsm, 100, 1, "OK");

		if (!fRes){
			ERR_RAISE(err_dataNotValid, dev_gsm);
			fRes = -1;
		}
	}

	/* Phone number not found */
	else if (fRes == 1) fRes = 0;


	return fRes;
}


/*******************************************************************************
 * @brief Write provided number to position in SIM memory.
 * @return true if OK, false if something went wrong.
 *******************************************************************************/
bool GSM_WritePhoneNumber(uint8_t position, char * phoneNumber){

	/* Format: AT+CPBW=XY,"00420123456789"
	 * where XY = position,
	 * "00420123456789" = phone number string */

	int fRes = STREAM_PrintfWaitResp(
			stream_gsm, 5000, "OK", "AT+CPBW=%02u,\"%s\"\r", position, phoneNumber);

	if (fRes < 1){
		ERR_RAISE(err_dataNotValid, dev_gsm);
		fRes = -1;
	}
	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Delete phone number from position.
 * @return true if OK, false if something went wrong.
 *******************************************************************************/
bool GSM_DelPhoneNumber(uint8_t position){

	int fRes = STREAM_PrintfWaitResp(
			stream_gsm, 5000, "OK", "AT+CPBW=%u\r", position);

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_gsm);

	//else if (fRes > 0) mem
	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Finds the next free phone book position.
 * @return > 0: Valid position.
 * return 0: Max phone number limit reached.
 *******************************************************************************/
uint8_t	GSM_NextPhbPos(){

	char simPhoneNumber[GSM_PHONE_NUMBER_LEN];

	uint8_t pos = 0;

	while (++pos <= GSM_MAX_REG_PHONE_NUM && GSM_GetPhoneNumber(pos, simPhoneNumber) > 0);

	return pos > GSM_MAX_REG_PHONE_NUM ? 0 : pos;
}


/*******************************************************************************
 * @brief Finds the current phone book position of the provided phone number.
 * @return > 0: Valid position.
 * return 0: Number not found.
 *******************************************************************************/
uint8_t	GSM_GetPhbPos(char * phoneNumber){

	char simPhoneNumber[GSM_PHONE_NUMBER_LEN];
	uint8_t pos = 0;
	int fRes;

	while (++pos <= GSM_MAX_REG_PHONE_NUM){

		fRes = GSM_GetPhoneNumber(pos, simPhoneNumber);

		if (fRes > 0 && !strcmp(phoneNumber, simPhoneNumber)) return pos;
	}

	return 0;
}
