/******************************************************************************
 * @file:		gsm_com.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	12. mai 2016
 ******************************************************************************
 *	The module encapsulates the SMS and INET modules, and provides
 *	functionality to convert a GSM command to a BMP message.
 ******************************************************************************/

#include <string.h>

#include "shared.h"
#include "err.h"
#include "system.h"

#include "gsm_com.h"

#include "gsm_sms.h"
#include "gsm_inet.h"
#include "gsm_messages.h"

#include "bmp_com.h"

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static void Error_Handler();
static bool AnalyzeTCPRequest(GSM_TCPRequest * req, BMP_TransferArgs * args);

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Receive incoming data from GSM/GPRS modem. Analyzes the data and
 * converts the request to a BMP message provided in the BMP_TransferArgs
 * argument.
 *******************************************************************************/
bool GSM_ReceiveTCP(BMP_TransferArgs * args){

	bool res = true;
	GSM_TCPRequest req;

	/* Set file name pointer to received message */
	req.file = args->receivedMsg->fileName;

	if (GSM_InTransparentMode()) res = GSM_IdentifyTCPRequest(&req);

	/* Identify data when not in transparent mode */
	else req.cmd = GSM_IdentifyData();

	if (res){

		bool bmpMsg = AnalyzeTCPRequest(&req, args);

		if (bmpMsg) BMP_RegisterNewMessage(args);
	}

	if (!res) Error_Handler();
	return res;
}


/*******************************************************************************
 * @brief Converts the provided BMP message to a GSM command and sends it to
 * the GSM/GPRS module. Not all messages are supported.
 *******************************************************************************/
bool GSM_TransmitTCP(BMP_Message * msg){

	bool res = true;

	switch(msg->cmd){

	case bc_nak:
		GSM_TransferData(GSM_CMD_Unknown);
		break;

	case bc_transFinished:
		GSM_TransferData(GSM_CMD_Finished);
		break;

	default:
		break;
	}

	if (!res) Error_Handler();
	return res;
}


/*******************************************************************************
 * @brief Sends a status SMS to all registered numbers.
 *******************************************************************************/
bool GSM_SendStatusSMS(){

	int res;
	uint8_t phoneBookPos = 0;

	if (!GSM_InTransparentMode()){

		while (++phoneBookPos <=
				GSM_MAX_REG_PHONE_NUM) res = GSM_Send_SysInfo(phoneBookPos, NULL);

	}

	if (res < 0) Error_Handler();
	return res < 0 ? false : true;
}


/*******************************************************************************
 * @brief Receive any incoming SMS triggered by the RI (ring indicator) input.
 * @return Com SMS result.
 *******************************************************************************/
GSM_SmsRes_smsr GSM_ReceiveSms(){

	bool resOk = true;
	uint8_t phbPos;

	GSM_Query_sq readRes = sq_error;
	GSM_SmsRes_smsr smsRes = smsr_ok_shutdown;
	char phoneNumber[GSM_PHONE_NUMBER_LEN];


	/* Exit transparent mode if connected to a TCP client or server,
	 * to be able to send commands to the modem */
	if (GSM_InTransparentMode()) GSM_ExitTransparentMode();

	int smsPos = GSM_IsSMSPresent(item_unread);
//	int smsPos = 1; readRes = sq_startTcpServer;
	if(smsPos > 0) readRes = GSM_ReadSMS(smsPos, phoneNumber);
	else if (!smsPos) readRes = sq_noSms;

	/* Delete SMS from SIM card if received */
	if(smsPos > 0) GSM_DeleteSMS(smsPos);

	switch(readRes){

	case sq_register:

		phbPos = GSM_NextPhbPos();

		if (phbPos){
			resOk = GSM_WritePhoneNumber(phbPos, phoneNumber);
			resOk = GSM_SendSMS(phoneNumber, resOk ? GSM_CMD_Welcome_SMS : GSM_CMD_Nak);
		}
		else resOk = GSM_SendSMS(phoneNumber, GSM_CMD_MaxReg);

		break;

	case sq_unRegister:

		phbPos = GSM_GetPhbPos(phoneNumber);

		if (phbPos) resOk = GSM_DelPhoneNumber(phbPos);
		resOk = GSM_SendSMS(phoneNumber, resOk ? GSM_CMD_Ack : GSM_CMD_Nak);
		break;

	case sq_startTcpServer:{

		char message[40] = {0};
		strcpy(message, GSM_CMD_Connect);
		strcat(message, ":");

		if (GSM_TcpStatus() != tcp_client){
			resOk = GSM_EstablishTCPServer(true, message + strlen(GSM_CMD_Connect) + 1, 40);
		}

		else GSM_SendSMS(phoneNumber, GSM_CMD_Nak);

		/* Send "Connect" message via SMS */
		if (resOk && GSM_TcpStatus() != tcp_client){

			resOk = GSM_SendSMS(phoneNumber, message);
			smsRes = smsr_ok_stayAwake;
		}
	}
		break;

	case sq_connectTCP:

		if (GSM_TcpStatus() == tcp_notConnected){
			resOk = GSM_ConnectToServer(true, NULL);
		}

		if (!resOk) GSM_SendSMS(phoneNumber, GSM_CMD_Nak);

		else {
			GSM_TransferData(GSM_CMD_Welcome_TCP);
			smsRes = smsr_ok_stayAwake;
		}

		break;

	case sq_endSession:

		switch (GSM_TcpStatus()){
		case tcp_client: resOk = GSM_DisconnectFromServer(); break;
		case tcp_server: resOk = GSM_CloseTCPServer(); break;
		default: break;
		}

		resOk = GSM_SendSMS(phoneNumber, resOk ? GSM_CMD_Ack : GSM_CMD_Nak);

		break;

	case sq_getStatus:

		resOk = GSM_Send_SysInfo(0, phoneNumber);
		break;

	case sq_info:
		resOk = GSM_SendSMS(phoneNumber, GSM_CMD_Welcome_SMS);
		break;

	case sq_unknown:

		resOk = GSM_SendSMS(phoneNumber, GSM_CMD_Unknown);
		break;

	case sq_error:
		resOk = false;
		break;

	default:
		/* No SMS: expected to be a voice call */
		GSM_CloseCall();
		break;
	}

	// while loop to check if any more SMS are present

	GSM_ReEnterTransparentMode();
	if (GSM_InTransparentMode()) smsRes = smsr_ok_stayAwake;


	if (!resOk) Error_Handler();

	return resOk ? smsRes : smsr_error;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Error handler for the GSM Com module.
 *******************************************************************************/
static void Error_Handler(){

	switch(error.erRes){

	case err_incompleteCmd:

		/* Send NAK to remote host */
		GSM_TransferData(GSM_CMD_Unknown);
		error.retVal = ret_resolved;
		break;

	case err_dataNotValid: /* Data received from GSM not valid */
	case err_rxTimeout: /* Data not received from GSM modem within timeout */
		// Reset module ?
		error.retVal = ret_resolved;
		break;

	default:
		error.retVal = ret_notResolved;
		break;
	}

	if (error.retVal == ret_resolved) ERR_RESET();
}


/*******************************************************************************
 * @brief Analyzes the provided TCP request and converts it to a BMP message
 * if possible.
 * @return true if converted to BMP message, otherwise false.
 *******************************************************************************/
static bool AnalyzeTCPRequest(
		GSM_TCPRequest * req,
		BMP_TransferArgs * args){

	bool isBmpMsg = true;

	/* Register the GSM UID with the system */
	args->sysIndex = SYS_IdentifySystem(STANDARD_UID_GSM);

	args->receivedMsg->senderAddr = STANDARD_UID_GSM;
	args->receivedMsg->receiverAddr = *SYS_GetID();

	switch(req->cmd){

	case sq_readNewData:

		args->receivedMsg->cmd = bc_fromLast;
		args->receivedMsg->mode = acm_poll;
		args->receivedMsg->deviceLog = req->deviceLog;
		break;

	case sq_readSpecifiedFile:

		args->receivedMsg->cmd = bc_specifiedLog;
		args->receivedMsg->mode = acm_poll;
		args->receivedMsg->deviceLog = req->deviceLog;

		break;

	case sq_getStatus:

		args->receivedMsg->cmd = bc_systemStatus;
		args->receivedMsg->mode = acm_poll;
		break;

	case sq_ack:

		args->receivedMsg->cmd = bc_ack;
		args->receivedMsg->mode = acm_response;
		break;

	case sq_nak:

		args->receivedMsg->cmd = bc_nak;
		args->receivedMsg->mode = acm_response;
		break;

	case sq_endSession:

		GSM_TransferData(GSM_CMD_Ack);
		args->receivedMsg->cmd = bc_sessionFinished;
		args->receivedMsg->mode = acm_response;
		break;

	case sq_info:
	case sq_connectTCP:

		GSM_TransferData(GSM_CMD_Welcome_TCP);
		isBmpMsg = false;
		break;

	case sq_unknown:

		GSM_TransferData(GSM_CMD_Unknown);
		isBmpMsg = false;
		break;

	default:

		isBmpMsg = false;
		break;
	}

	return isBmpMsg;
}
