/******************************************************************************
 * @file:		gsm_inet.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	24. apr. 2016
 ******************************************************************************
 *	Handles all of the TCP/IP related communication with the GSM/GPRS module.
 ******************************************************************************/

#include <string.h>

#include "system_format.h"
#include "timer.h"
#include "err.h"

#include "gsm_inet.h"
#include "gsm_driver.h"

#include "stream.h"


/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static int ConfigureTransparentMode();

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

static bool inTransparentMode 		= false;
static bool wasInTransparentMode 	= false;
static bool clientConnected 		= false;
static bool isServer 				= false;
static bool isClient   				= false;

static char * apnInet, * tcpPort, localIp[16];

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief If the BMP protocol is not used for GSM, commands are identified
 * using the GSM commands stored in memory.
 * @return True if command OK, false otherwise.
 *******************************************************************************/
bool GSM_IdentifyTCPRequest(GSM_TCPRequest * tcpRequest){

	bool res = true;
	char date[6];
	char * fileName = NULL;

	/* Identify TCP Command */
	int fRes = STREAM_WaitAltResponse(
			stream_gsm, 200, 8,
			GSM_CMD_Info,
			GSM_CMD_GetStatus,
			GSM_CMD_StopTCP,
			GSM_CMD_ReadNewData,
			GSM_CMD_SpecifiedFile,
			GSM_CMD_Ack, GSM_CMD_Nak,
			"CLOSED");


	if (fRes == 8){ /* "CLOSED" */

		tcpRequest->cmd = sq_endSession;
		inTransparentMode = false;
		clientConnected = false;
		wasInTransparentMode = false;
	}

	else tcpRequest->cmd = GSM_TCP_MSG_TO_QUERY(fRes);


	if (tcpRequest->cmd == sq_readNewData ||
			tcpRequest->cmd == sq_readSpecifiedFile){

		/* Identify log to access */
		fRes = STREAM_GetValue(stream_gsm, 100);

		if (fRes >= 0 && fRes < N_LOG_DEVICES) tcpRequest->deviceLog = (Device_dev)fRes;

		else res = false;


		if (res && tcpRequest->cmd == sq_readSpecifiedFile){

			/* File to access */
			fRes = STREAM_GetString(
					stream_gsm, 100, '"', '"', date, 6);


			if (fRes > 0) fileName = SYS_MakeFileName(
					false, tcpRequest->deviceLog, date);

			/* Valid date and file name -> copy into tcpRequest */
			if (fileName) strcpy(tcpRequest->file, fileName);

			else res = false;
		}
	}

	STREAM_WaitAltResponse(stream_gsm, 50, 1, "\n");

	if (!res) ERR_RAISE(err_incompleteCmd, dev_gsm);

	return tcpRequest->cmd == sq_error ? false : res;
}


/*******************************************************************************
 * @brief Identifies incoming message from the GSM modem. Meant to identify
 * messages when not in transparent mode.
 *******************************************************************************/
GSM_Query_sq GSM_IdentifyData(){

	GSM_Query_sq query = sq_unknown;

	int fRes = STREAM_WaitAltResponse(
			stream_gsm, 200, 3, "CONNECT\r\n", "CLOSED\r\n", "OK\r\n");

	switch (fRes){

	case 1: /* Client connected */
		clientConnected = true;
		inTransparentMode = true;
		query = sq_connectTCP;
		break;

	case 2: /* Client/server closed connection */
		clientConnected = false;
		inTransparentMode = false;
		wasInTransparentMode = false;
		isClient = false;
		query = sq_endSession;
		break;

	default: /* Not recognized */
		break;
	}

	return query;
}


/*******************************************************************************
 * @brief Transfers the provided data buffer to stream.
 *******************************************************************************/
bool GSM_TransferData(const char * buf){

	bool res = true;
	if (!inTransparentMode) GSM_ReEnterTransparentMode();
	res = STREAM_Printf(stream_gsm, buf);

	return res;
}


/*******************************************************************************
 * @brief
 * @return true if module is currently acting as a TCP server. If false, it
 * can either be idle, or it is acting as a TCP client.
 *******************************************************************************/
GSM_TcpStat GSM_TcpStatus(){

	GSM_TcpStat stat;

	if (isClient && clientConnected) stat = tcp_client;

	else if (isServer) stat = tcp_server;

	else stat = tcp_notConnected;

	return stat;
}


/*******************************************************************************
 * @brief
 * @return true if module is in transparent mode.
 *******************************************************************************/
bool GSM_InTransparentMode(){

	return inTransparentMode;
}


/*******************************************************************************
 * @brief Exits transparent mode and goes to command mode.
 *******************************************************************************/
bool GSM_ExitTransparentMode(){

	if (inTransparentMode){
		wasInTransparentMode = true;
	}
	/* DTR must be configured to exit transparent mode
	 * on ON->OFF transition in data mode */
	GSM_LowPowerMode(true);
	Sleep_ms(200, true);
	GSM_LowPowerMode(false);
	Sleep_ms(100, true);

	inTransparentMode = false;
	return true;
}


/*******************************************************************************
 * @brief Re-enters transparent mode if a client is still connected to
 * the TCP server.
 *******************************************************************************/
bool GSM_ReEnterTransparentMode(){

	int fRes = 0;
	if (wasInTransparentMode && !inTransparentMode){

		fRes = STREAM_PrintfWaitResp(
				stream_gsm, 1000, "CONNECT\r\n", "ATO\r");

		if (!fRes) clientConnected = false;
		else if (fRes > 0) inTransparentMode = true;
	}
	return fRes >= 0 ? true : false;
}


/*******************************************************************************
 * @brief Set the APN to use for the GPRS connection. Service provider specific.
 *******************************************************************************/
void GSM_SetApn(char * apn){

	if(apn) apnInet = apn;
}


/*******************************************************************************
 * @brief Set the standard TCP port to use when connecting to remote TCP server.
 *******************************************************************************/
void GSM_SetStandardTcpPort(char * port){

	if (port) tcpPort = port;
}


/*******************************************************************************
 * @brief Connect to a TCP server with the specified IP address. If IP is not
 * provided, it is assumed that the IP address is available in the stream (provided in SMS).
 * If the IP address contains the port, it will be used. Otherwise,
 * the standard port is used.
 * @return True if everything OK.
 *******************************************************************************/
bool GSM_ConnectToServer(
		bool transparentMode,
		char * ipAddr){

	int fRes = -1;
	char ip[25];
	char * ipPtr = NULL, * portPtr;


	/* Get IP address from stream if not supplied in the argument */
	if (!ipAddr){
		fRes = STREAM_GetString(stream_gsm, 200, '"', '"', ip, 25);
		if (fRes > 0) ipPtr = &ip[0];
	}
	else ipPtr = ipAddr;


	/* No IP found */
	if (!ipPtr) goto exit;

	/* Search for a colon in the IP address */
	else if ((portPtr = strstr(ipPtr, ":"))){
		*portPtr = 0;
		portPtr += 1;
	}

	else portPtr = tcpPort;



	/* Check if modem is attached to GPRS */
	STREAM_Printf(stream_gsm, "AT+CGATT?\r");

	fRes = STREAM_WaitAltResponse(
			stream_gsm, 1000, 2, "+CGATT:1", "+CGATT:0");

	STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");

	if (fRes == 2){
		/* Try to attach to GPRS */
		fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK", "AT+CGATT=1\r");
	}

	/* Close any connections (reset state) */
	if (fRes > 0) STREAM_Printf(stream_gsm, "AT+CIPSHUT\r");
	if (fRes > 0) fRes = STREAM_WaitAltResponse(stream_gsm, 1000, 2, "OK", "SHUT OK");

	if (fRes > 0 && transparentMode) fRes = ConfigureTransparentMode();

	/* Set APN */
	if (fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK", "AT+CSTT=\"%s\"\r", apnInet);

	/* Bring up connection */
	if (fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 3000, "OK", "AT+CIICR\r");

	/* Get local ip address */
	if (fRes > 0) STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");
	if (fRes > 0) STREAM_Printf(stream_gsm, "AT+CIFSR\r");
	if (fRes > 0) STREAM_GetString(stream_gsm, 2000, '\n', '\r', localIp, 16);

	/* Connect to TCP server */
	if (fRes > 0) STREAM_Printf(stream_gsm, "AT+CIPSTART=\"TCP\",\"%s\",\"%s\"\r", ipPtr, portPtr);
	if (fRes > 0) fRes = STREAM_WaitAltResponse(stream_gsm, 2000, 3, "CONNECT\r\n", "CONNECT FAIL", "STATE");

	if (fRes > 1 || fRes < 1){
		STREAM_Printf(stream_gsm, "AT+CIPSHUT\r");
		STREAM_WaitAltResponse(stream_gsm, 1000, 2, "OK\r\n", "SHUT OK\r\n");
		ERR_RAISE(err_dataNotValid, dev_gsm);
		fRes = 0;
	}

	if (fRes > 0 && transparentMode){
		clientConnected = true;
		inTransparentMode = true;
		isClient = true;
		isServer = false;
	}

exit:
	return fRes > 0 ? ipPtr ? true : false : false;
}



/*******************************************************************************
 * @brief Disconnects from TCP server if connected.
 *******************************************************************************/
bool GSM_DisconnectFromServer(){

	if (inTransparentMode) GSM_ExitTransparentMode();

	/* Close any connections (reset state) */
	int fRes = STREAM_PrintfWaitResp(stream_gsm, 2000, "SHUT OK", "AT+CIPSHUT\r");

	if (!fRes) {ERR_RAISE(err_dataNotValid, dev_gsm);}

	else if (fRes >= 0){
		clientConnected = false;
		wasInTransparentMode = false;
		isClient = false;
	}

	return fRes >= 0 ? true : false;
}


/*******************************************************************************
 * @brief Starts a TCP server. Returns the IP address and standard port.
 * @return True if everything OK.
 *******************************************************************************/
bool GSM_EstablishTCPServer(
		bool transparentMode,
		char * ipAddr,
		uint8_t ipLen){


	int fRes;

	/* If server is already active,
	 * create the IP address string and return.
	 */
	if (isServer){
		strcat(ipAddr, localIp);
		strcat(ipAddr, ":");
		strcat(ipAddr, tcpPort);
		fRes = 1;
		goto exit;
	}

	/* Check if modem is attached to GPRS */
	STREAM_Printf(stream_gsm, "AT+CGATT?\r");

	fRes = STREAM_WaitAltResponse(
			stream_gsm, 1000, 2, "+CGATT:1", "+CGATT:0");

	STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");

	if (fRes == 2){
		/* Try to attach */
		fRes = STREAM_PrintfWaitResp(stream_gsm, 2000, "OK", "AT+CGATT=1\r");
	}


	if (fRes > 0 && transparentMode) fRes = ConfigureTransparentMode();

	/* Set APN */
	if (fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK", "AT+CSTT=\"%s\"\r", apnInet);

	/* Close any connections (reset state) */
	if (fRes > 0) STREAM_Printf(stream_gsm, "AT+CIPSHUT\r");
	if (fRes > 0) fRes = STREAM_WaitAltResponse(stream_gsm, 1000, 2, "OK", "SHUT OK");

	/* Start TCP server */
	if (fRes > 0) fRes = STREAM_PrintfWaitResp(stream_gsm, 3000, "SERVER OK", "AT+CIPSERVER=1,%s\r", tcpPort);

	/* Get local ip address */
	if (fRes > 0) STREAM_WaitAltResponse(stream_gsm, 100, 1, "OK");
	if (fRes > 0) STREAM_Printf(stream_gsm, "AT+CIFSR\r");
	if (fRes > 0) STREAM_GetString(stream_gsm, 3000, '\n', '\r', ipAddr, ipLen);

	if (fRes > 0){

		/* Copy received IP address into "localIP" */
		strcpy(localIp, ipAddr);

		/* Append port number to IP address string */
		strcat(ipAddr, ":");
		strncat(ipAddr, tcpPort, ipLen - strlen(ipAddr));

		/* Check server status */
		fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "SERVER LISTENING\r\n", "AT+CIPSTATUS\r");
	}

	if (fRes > 0) isServer = true;

	if (!fRes) ERR_RAISE(err_dataNotValid, dev_gsm);

exit:
	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Closes the currently open TCP server as well as any open connections
 * to clients.
 *******************************************************************************/
bool GSM_CloseTCPServer(){

	int fRes;

	if (inTransparentMode) GSM_ExitTransparentMode();

	fRes = STREAM_PrintfWaitResp(
					stream_gsm, 1000, "SERVER CLOSE", "AT+CIPSERVER=0\r");

	if (clientConnected && fRes > 0) fRes = STREAM_PrintfWaitResp(
					stream_gsm, 1000, "CLOSE OK", "AT+CIPCLOSE\r");


	if (fRes > 0) STREAM_Printf(stream_gsm, "AT+CIPSHUT\r");
	if (fRes > 0) fRes = STREAM_WaitAltResponse(stream_gsm, 1000, 2, "OK\r\n", "SHUT OK\r\n");

	if (fRes > 0){

		clientConnected = false;
		wasInTransparentMode = false;
		isServer = false;
	}


	if (!fRes) ERR_RAISE(err_dataNotValid, dev_gsm);

	return fRes > 0 ? true : false;
}


/*******************************************************************************
 * @brief Sends close call command to modem.
 *******************************************************************************/
bool GSM_CloseCall(){

	int fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK\r\n", "ATH\r");
	return fRes >= 0 ? true: false;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Configures the GSM modem to use transparent mode after connection
 * to remote host has been established.
 *******************************************************************************/
static int ConfigureTransparentMode(){

	int fRes;

	/* Close any connections (reset state) */
	STREAM_Printf(stream_gsm, "AT+CIPSHUT\r");
	fRes = STREAM_WaitAltResponse(stream_gsm, 500, 2, "OK", "SHUT OK");

	/* Set TCP/IP transparent mode */
	fRes = STREAM_PrintfWaitResp(stream_gsm, 1000, "OK", "AT+CIPMODE=1\r");

	/* Configure transfer settings in transparent mode */
	if (fRes > 0) fRes = STREAM_PrintfWaitResp(
			stream_gsm, 1500, "OK", "AT+CIPCCFG=5,%d,%d,1\r", GSM_SERIAL_PACKET_TIMEOUT, GSM_TCP_MAX_PACKET_LEN);

	return fRes;
}
