/******************************************************************************
 * @file:		log_manager.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	8. feb. 2016
 ******************************************************************************
 *	The module manages all contact with the log files and the file system
 *	modules. Also allows configuration through configuration files.
 ******************************************************************************/

#include <string.h>

#include "system.h"
#include "err.h"
#include "timer.h"

#include "log_manager.h"
#include "stream_functions.h"
#include "file.h"


static char * AllocateFileName();

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

/* Follows the "Device" enum order */
static LOG_Properties logs[2][6] = {
{
		{NULL, NULL, LOG_TBR_DIR, 0}, 		/* dev_tbr700 */
		{NULL, NULL, LOG_AQUA_DIR, 0},		/* dev_aquaTroll */
		{NULL, NULL, LOG_WS_DIR, 0},		/* dev_ws */
		{NULL, NULL, LOG_GPS_DIR, 0},		/* dev_gps */
		{NULL, NULL, LOG_IMU_DIR, 0},		/* dev_imu */
		{NULL, NULL, LOG_GAUGE_DIR, 0}},	/* dev_gauge */
{
		{NULL, NULL, LOG_TBR_CDIR, 0},		/* dev_tbr700 */
		{NULL, NULL, LOG_AQUA_CDIR, 0},		/* dev_aquaTroll */
		{NULL, NULL, LOG_WS_CDIR, 0},		/* dev_ws */
		{NULL, NULL, LOG_GPS_CDIR, 0},		/* dev_gps */
		{NULL, NULL, LOG_IMU_CDIR, 0},		/* dev_imu */
		{NULL, NULL, LOG_GAUGE_CDIR, 0}}	/* dev_gauge */
};

static char * parentDirs[2] = {LOG_PARENT_DIR, LOG_COMB_PAR_DIR};

static char logFileNames[N_LOG_DEVICES * 4][FILENAME_MAX_SIZE];

static FILE_InStreamFunc 	inStream[N_STREAM_DEVICES] = {
		STREAM_ReceiveDataFromRadio,
		STREAM_ReceiveDataFromTbr700,
		NULL,
		NULL,
		STREAM_ReceiveDataFromGsm
};

static FILE_OutStreamFunc 	outStream[N_STREAM_DEVICES] = {
		STREAM_TransferDataToRadio,
		NULL,
		NULL,
		NULL,
		STREAM_TransferDataToGSM
};

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Loads the last saved log operational parameters from file, and
 * allocates memory for log file names.
 *******************************************************************************/
bool LOG_Init(){

	bool res = true;
	int max = *SYS_GetID() == *SYS_GetGateway() ? 2 : 1;

	for (int deviceLog = 0; deviceLog < N_LOG_DEVICES; deviceLog++){

		/* Allocate memory for log variables */
		for (int log = 0; log < max; log++){

			if (!logs[log][deviceLog].lastReadFile)
				logs[log][deviceLog].lastReadFile = AllocateFileName();

			LOG_LoadDeviceOpParam(deviceLog, log);
		}
	}

	for (int opParam = 1; opParam < LOG_OP_PARAM_ITEMS; opParam++){
		if (res) res = LOG_LoadMiscOpParam(opParam);
	}

	LOG_CloseOpParamFile();

	return res;
}


/*******************************************************************************
 * @brief Provides the last read log file for the specified device log and
 * log type.
 *******************************************************************************/
void LOG_GetLastReadFile(Device_dev deviceLog, LOG_Type log, char * fName){

	if (fName && strcmp(fName, logs[log][deviceLog].lastReadFile)){
		strcpy(fName, logs[log][deviceLog].lastReadFile);
	}
}


/*******************************************************************************
 * @brief Creates a log directory for each log device. The log argument defines
 * whether to access the combined log, or the local log.
 *******************************************************************************/
bool LOG_CreateDirectories(LOG_Type log){

	bool resOk = true;

	/* Create parent directory */
	if (resOk) resOk = FILE_CreateDir(parentDirs[log]);

	for (int i = 0; i < N_LOG_DEVICES; i++){

		/* Create log directories */
		if (resOk) resOk = FILE_CreateDir(logs[log][i].dir);
	}

	return resOk;
}


/*******************************************************************************
 * @brief Creates a new file for the specified device log, and registers the file
 * as the current log file. The log argument defines whether to access the
 * combined log, or the local log.
 *******************************************************************************/
bool LOG_CreateNewFile(
		Device_dev deviceLog,
		char * fName,
		LOG_Type log){

	bool resOk = true;
	FIL * fileObjPtr;


	/* Allocate memory for new file if not already allocated */
	if (!logs[log][deviceLog].currentSavefile){

		logs[log][deviceLog].currentSavefile = AllocateFileName();
	}

	/* Copy new file name into memory location */
	strcpy(logs[log][deviceLog].currentSavefile, fName);

	/* Create new, or open existing file */

	if (resOk) resOk = FILE_ChangeDir(logs[log][deviceLog].dir);

	if (resOk) fileObjPtr = FILE_OpenNew(
			logs[log][deviceLog].currentSavefile);

	LOG_Close(deviceLog, log);


	/* If operational parameters file was empty ("lastReadFile" empty),
	 * update last read file with the current file name */
	if (logs[log][deviceLog].lastReadFile &&
			!strstr(logs[log][deviceLog].lastReadFile, ".txt")){

		strcpy(logs[log][deviceLog].lastReadFile, fName);
		logs[log][deviceLog].lastReadPos = 0;

		/* Save parameters to file */
		LOG_SaveDeviceOpParam(deviceLog, log);
	}

	return !fileObjPtr ? false : resOk;
}


/*******************************************************************************
 * @brief Closes the current log file.
 * Returns true if no files are open.
 *******************************************************************************/
bool LOG_Close(
		Device_dev deviceLog,
		LOG_Type log){


	FILE_Close(logs[log][deviceLog].currentSavefile);

	/* Return true if no files are open */
	return !FILE_GetNOpenFiles();
}


/*******************************************************************************
 * @brief Closes all open files.
 *******************************************************************************/
bool LOG_CloseAllOpen(){

	return FILE_CloseAll();
}


/*******************************************************************************
 * @brief Closes the operational parameters file.
 *******************************************************************************/
bool LOG_CloseOpParamFile(){

	FILE_Close(LOG_CONFIG);
	return !FILE_GetNOpenFiles();
}


/*******************************************************************************
 * @brief Loads the specified device log parameters from the operational param file.
 * This includes the last read file and last read file position for the
 * specified log type.
 * File is opened and accessed, but will not be closed afterwards.
 *******************************************************************************/
bool LOG_LoadDeviceOpParam(Device_dev device, LOG_Type log){

	bool resOk = true;
	char lastReadPos[11] = {0};

	resOk = LOG_LoadOpParam(
			device, param_device, log, FILENAME_MAX_SIZE,
			LOG_OP_PARAM_DEVICE_ITEMS, logs[log][device].lastReadFile, lastReadPos);


	if (resOk){ /* File successfully opened */

		logs[log][device].lastReadPos = (uint32_t)atoi(lastReadPos);
	}

	return resOk;
}


/*******************************************************************************
 * @brief Saves the specified device log parameters to operational param file.
 * This includes the last read file and last read file position for the
 * specified log type.
 * File is opened and accessed, but will not be closed afterwards.
 *******************************************************************************/
bool LOG_SaveDeviceOpParam(Device_dev device, LOG_Type log){

	bool resOk = true;

	char lastReadPos[11] = {0};

	sprintf(lastReadPos, ",%lu", logs[log][device].lastReadPos);

	resOk = LOG_SaveOpParam(
				device, param_device, log, LOG_OP_PARAM_DEVICE_ITEMS,
				logs[log][device].lastReadFile, lastReadPos);

	return resOk;
}


/*******************************************************************************
 * @brief Loads the specified parameter from the operational parameters file.
 * File is opened and accessed, but will not be closed afterwards.
 *******************************************************************************/
bool LOG_LoadMiscOpParam(LOG_Param param){

	bool resOk = true;
	char buf[LOG_MAX_LINE_LEN];
	uint32_t convertedValue;

	resOk = LOG_LoadOpParam(dev_noDevice, param,
			type_local, LOG_MAX_LINE_LEN, 1, buf);

	if (resOk){

		switch(param){

		case param_tbrTagDetections:

			convertedValue = (uint32_t)atoi(buf);
			SYS_TbrDetections(&convertedValue);
			break;

		case param_lastReadTbrFlashCode:

			convertedValue = (uint32_t)atoi(buf);
			SYS_TbrFlashCode(&convertedValue);
			break;

		default:
			break;
		}
	}
	return resOk;
}


/*******************************************************************************
 * @brief Saves the specified parameter to the operational parameters file.
 * File is opened and accessed, but will not be closed afterwards.
 *******************************************************************************/
bool LOG_SaveMiscOpParam(LOG_Param param){

	bool resOk;
	char buf[LOG_MAX_LINE_LEN];

	switch(param){

	case param_tbrTagDetections:

		sprintf(buf, "%010lu", *SYS_TbrDetections(NULL));
		break;

	case param_lastReadTbrFlashCode:

		sprintf(buf, "%010lu", *SYS_TbrFlashCode(NULL));
		break;

	default:
		break;
	}

	resOk = LOG_SaveOpParam(dev_noDevice, param, type_local, 1, buf);

	return resOk;
}



/*******************************************************************************
 * @brief Check if new data is available in the log, for the
 * specified device log and log type.
 *******************************************************************************/
bool LOG_NewDataAvailable(Device_dev deviceLog, LOG_Type log){

	bool dataAv = false;

	/* Check if log files exist */
	if (logs[log][deviceLog].lastReadFile &&
			logs[log][deviceLog].currentSavefile){

		/* Open the log directory */
		FILE_ChangeDir(logs[log][deviceLog].dir);

		/* Open the last read log file */
		FIL * fileObjPtr = FILE_OpenExisting(logs[log][deviceLog].lastReadFile);

		/* Check if new data is available since the last "readFromLast" poll */
		if (fileObjPtr && fileObjPtr->fsize >
			logs[log][deviceLog].lastReadPos) dataAv = true;

		else if (strcmp(logs[log][deviceLog].lastReadFile,
						logs[log][deviceLog].currentSavefile)){

			/* Last read file and current save file are not the same -> check if
			 * data is available in current save file */

			fileObjPtr = FILE_OpenExisting(logs[log][deviceLog].currentSavefile);

			if (fileObjPtr && fileObjPtr->fsize > 0) dataAv = true;
		}

		else {

			FILE_Close(logs[log][deviceLog].lastReadFile);
			FILE_Close(logs[log][deviceLog].currentSavefile);
		}
	}
	return dataAv;
}


/*******************************************************************************
 * @brief Reads incoming data from specified stream into the buffer,
 * until an EOT char is received.
 * Returns number of bytes received. Returns 0 if RX buffer timeout.
 *******************************************************************************/
uint16_t LOG_ReceiveToBuf(
		Stream stream,
		uint8_t * buf){

	bool transFinished = false;
	uint8_t dataAvailable = 0;
	uint16_t nBytesRecv = 0;

	while (!transFinished){

		StartTimer_ms(LOG_SEND_RECV_TIMEOUT_MS);

		/* Wait for data to be available */
		while (dataAvailable == 0){

			/* Timeout error if data has not become
			 * available within the specified time */
			if (TimeElapsed()){
				ERR_RAISE(err_rxTimeout, STREAM_TO_DEV(stream));
				nBytesRecv = 0;
				goto end;
			}

			/* Sense call */
			dataAvailable = inStream[stream](0, 1);
		}

		/* Do as long as data is available */
		while (dataAvailable != 0){

			dataAvailable = inStream[stream](buf++, 0);

			/* Signals that the "end of transmission"
			 * char has been received */
			if (dataAvailable == 0xFF){
				transFinished = true;
			}
			else if (dataAvailable == 1) nBytesRecv++;
		}
	}
end:
	return nBytesRecv;
}


/*******************************************************************************
 * @brief Transfer provided data buffer to the specified stream.
 * Bytes to transfer must be lower than MAX PACKET LEN - BMP HEADER LEN for the
 * data to be transferred in one packet.
 *******************************************************************************/
bool LOG_TransferFromBuf(
		Stream stream,
		uint8_t * buf,
		uint16_t bytesToTransfer){

	bool resOk = true;
	uint16_t bytesTransferred = 0;

	while (bytesToTransfer != bytesTransferred){

		/* Sense call to check if stream is ready to receive data */
		StartTimer_ms(LOG_SEND_RECV_TIMEOUT_MS);

		while (outStream[stream](buf, 0) != 1){

			if (TimeElapsed()) {
				ERR_RAISE(err_txTimeout, STREAM_TO_DEV(stream));
				resOk = false;
				goto end;
			}
		}

		/* Transfer call */
		bytesTransferred += outStream[stream](
							buf + bytesTransferred,
							bytesToTransfer - bytesTransferred);
	}
end:
	return resOk;
}


/*******************************************************************************
 * @brief Reads from the specified log (local or combined) and device log,
 * and writes to the output stream.
 *******************************************************************************/
bool LOG_StreamFromStorage(LOG_StreamArgs * args){

	bool resOk = true;
	FIL * fileObjPtr;

	/* last read file position for current -and specified file access */
	static uint32_t lastReadPosCurr_Spes = 0;


	STREAM_SetArgs(args->waitAfterTransferEnd_ms, args->nBytesToTransfer);


	/* Open the log directory */
	if (resOk) resOk = FILE_ChangeDir(
			logs[args->logToAccess][args->deviceLogToAccess].dir);


	switch (args->accessMode){

	case am_readCurrentFile:

		/* Open the current log file */
		if (resOk) fileObjPtr = FILE_OpenExisting(
				logs[args->logToAccess][args->deviceLogToAccess].currentSavefile);

		if (resOk && fileObjPtr) resOk = FILE_MoveToPosition(fileObjPtr, lastReadPosCurr_Spes);

		break;

	case am_readSpecifiedFile:

		/* Open the specified log file */
		if (resOk) fileObjPtr = FILE_OpenExisting(args->fileToAccess);

		if (resOk && fileObjPtr) resOk = FILE_MoveToPosition(fileObjPtr, lastReadPosCurr_Spes);

		break;

	case am_readFromLast:

		/* Open the last read log file */
		if (resOk) fileObjPtr = FILE_OpenExisting(
				logs[args->logToAccess][args->deviceLogToAccess].lastReadFile);

		if (resOk && fileObjPtr) resOk = FILE_MoveToPosition(fileObjPtr,
				logs[args->logToAccess][args->deviceLogToAccess].lastReadPos);

		break;

	default:
		break;
	}

	/* Transfer data from the specified log to the stream */
	if (resOk && fileObjPtr) resOk = FILE_TransferToStream(
			fileObjPtr,
			outStream[args->stream],
			args->nBytesToTransfer);


	if (resOk && fileObjPtr){

		if (args->accessMode == am_readFromLast){

			/* Update last read file position */
			logs[args->logToAccess][args->deviceLogToAccess].lastReadPos = fileObjPtr->fptr;

			if (fileObjPtr->fptr == fileObjPtr->fsize){ /* Reached end of file */

				FILE_Close(logs[args->logToAccess][args->deviceLogToAccess].lastReadFile);


				if (args->getNextFile && strcmp(logs[args->logToAccess][args->deviceLogToAccess].lastReadFile,
						logs[args->logToAccess][args->deviceLogToAccess].currentSavefile)){

					/* Last read file and current save file does not match */
					/* Call back for the next log file */
					args->getNextFile(
							logs[args->logToAccess][args->deviceLogToAccess].lastReadFile);

					/* Reset file position pointer */
					logs[args->logToAccess][args->deviceLogToAccess].lastReadPos = 0;

					/* If "next file" does not exist, set current save file as last read file */
					if (!FILE_OpenExisting(logs[args->logToAccess][args->deviceLogToAccess].lastReadFile)){

						strcpy(logs[args->logToAccess][args->deviceLogToAccess].lastReadFile,
								logs[args->logToAccess][args->deviceLogToAccess].currentSavefile);
					}
				}

				/* Call back on transfer complete */
				else if (args->transferComplete) args->transferComplete();

			}
		}
		else { /* mode != am_readFromLast */

			if (fileObjPtr->fptr == fileObjPtr->fsize){ /* Reached end of file */

				lastReadPosCurr_Spes = 0;

				if (args->transferComplete) args->transferComplete();
				FILE_Close(args->fileToAccess);
			}

			/* Update last read file position */
			else lastReadPosCurr_Spes = fileObjPtr->fptr;
		}
	}

	return fileObjPtr ? resOk: false;
}


/*******************************************************************************
 * @brief Write data to the specified log (local or combined) and device log,
 * from the input stream. The access mode defines how to access the file.
 *******************************************************************************/
bool LOG_StreamToStorage(LOG_StreamArgs * args){

	bool resOk = true;
	FIL * fileObjPtr;


	/* Open the log directory */
	if (resOk) resOk = FILE_ChangeDir(
			logs[args->logToAccess][args->deviceLogToAccess].dir);


	switch (args->accessMode){

	case am_appendToSpecifiedFile:

		/* Open log file with specified file name */
		if (resOk) fileObjPtr = FILE_OpenAppend(args->fileToAccess);

		break;

	case am_appendToCurrentFile:

		/* Open existing file to append data */
		if (resOk) fileObjPtr = FILE_OpenAppend(
				logs[args->logToAccess][args->deviceLogToAccess].currentSavefile);

		break;

	default:
		break;
	}

	/* Write data to file using the stream function pointer */
	if (fileObjPtr && resOk) resOk = FILE_WriteFromStream(
			fileObjPtr,
			inStream[args->stream]);


	return fileObjPtr ? resOk : false;
}


/*******************************************************************************
 * @brief Append provided data to the specified log (local or combined)
 * and device log.
 *******************************************************************************/
bool LOG_WriteToStorage(
		Device_dev deviceLog,
		LOG_Type log,
		char * data,
		uint16_t dataLen){

	bool resOk = true;
	FIL * fileObjPtr;


	/* Open the log directory */
	if (resOk) resOk = FILE_ChangeDir(
			logs[log][deviceLog].dir);

	/* Open existing file to append data */
	if (resOk) fileObjPtr = FILE_OpenAppend(
			logs[log][deviceLog].currentSavefile);

	/* Write data to file */
	if (resOk && fileObjPtr) resOk = FILE_Write(fileObjPtr, data, dataLen);


	return !fileObjPtr ? false : resOk;
}


/*******************************************************************************
 * @return pointer to an available log file name array.
 *******************************************************************************/
static char * AllocateFileName(){

	static int count = 0;

	if (count < N_LOG_DEVICES * 4){

		return &logFileNames[count++][0];
	}
	else return NULL;
}
