/******************************************************************************
 * @file:		log_param.c
 * @project:	BuoyController
 * @author:  	Ove
 * @version:	1.00
 * @created:	28. mai 2016
 ******************************************************************************
 *	Module allows simplified access to the configuration files.
 ******************************************************************************/

#include <string.h>
#include <stdarg.h>

#include "system.h"
#include "err.h"

#include "log_param.h"
#include "file.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static FIL * FindOpString(
		LOG_Type log,
		LOG_Param param,
		Device_dev device,
		char ** bufPos,
		uint32_t * filePointer);

static FIL * FindUserString(
		LOG_UserParam_user param,
		Device_dev device,
		char ** bufPos,
		char * buf,
		uint32_t * filePointer);

static bool CreateParam(
		FIL * fileObjPtr,
		LOG_Param param,
		LOG_Type log,
		Device_dev device);

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Load user configuration from file. Opens and closes the file.
 *******************************************************************************/
bool LOG_LoadUserConfig(LOG_UserConfig * config){

	bool resOk = true;
	char * bufPos;
	char buf[LOG_MAX_LINE_LEN];
	FIL * fileObjPtr;
	uint32_t filePointer;
	int readValueCounter = 0;

	for (int i = 0; i < N_DEVICES; i++){

		/* Initially set config as not found */
		config->configFound[i] = false;

		/* Clear buffer */
		memset(&buf[0], 0, LOG_MAX_LINE_LEN);

		/* Search for string in user file (user_apn is don't care!) */
		fileObjPtr = FindUserString(user_apn, i, &bufPos, buf, &filePointer);

		if (!fileObjPtr) goto exit;

		readValueCounter = 0;
		config->deviceOpConfig[readValueCounter][i] = 0; /* Initialize variable to zero */

		/* Read until new parameter, or end of buffer */
		while (bufPos && *bufPos != '\0'){

			if (*bufPos == ','){
				readValueCounter++;
				config->deviceOpConfig[readValueCounter][i] = 0; /* Initialize variable to zero */
			}

			/* Read values exceed values per line */
			if (readValueCounter >= LOG_USER_CONFIG_DEVICE_ITEMS) break;


			if (*bufPos < '0' || *bufPos > '9') bufPos++;

			else { /* Convert ASCII to integer value */

				/* Set configuration found */
				config->configFound[i] = true;

				config->deviceOpConfig[readValueCounter][i] *= 10;
				config->deviceOpConfig[readValueCounter][i] += (*bufPos - '0');
				bufPos++;
			}
		}
	}

	for (int i = 0; i < LOG_USER_CONFIG_MISC_ITEMS; i++){

		/* Initially set config as not found */
		config->configFound[i + N_DEVICES] = false;

		/* Clear buffer */
		memset(&buf[0], 0, LOG_MAX_LINE_LEN);

		/* Search for string in user file */
		fileObjPtr = FindUserString(i, dev_noDevice, &bufPos, buf, &filePointer);

		int pos = 0;
		int stringLen = LOG_USER_CONFIG_MISC_ITEMS_LEN;

		/* Read until new parameter, or end of buffer */
		while (bufPos && *++bufPos != '\0' && *bufPos != '\r'
				&& *bufPos != '\n' && *bufPos != ' '){

			/* Configuration found */
			config->configFound[i + N_DEVICES] = true;

			/* Stop before overflowing buffer */
			if (pos >= stringLen) break;

			/* Load read values into the provided char array */
			*(config->misc[i] + pos++) = *bufPos;
		}
	}

	FILE_Close(LOG_CONFIG_USER);

exit:
	return !fileObjPtr ? false : resOk;
}


/*******************************************************************************
 * @brief Clears the user parameter from user configuration file. Opens and
 * closes the file. Note: clear only one character!
 *******************************************************************************/
bool LOG_ClearUserParam(LOG_UserParam_user param){

	bool resOk = true;
	char * bufPos;
	char buf[LOG_MAX_LINE_LEN] = {0};

	FIL * fileObjPtr;
	uint32_t filePointer;

	/* Search for string in user file */
	fileObjPtr = FindUserString(param, dev_noDevice, &bufPos, buf, &filePointer);

	if (fileObjPtr && bufPos){

		resOk = FILE_MoveToPosition(fileObjPtr, filePointer + 1);
		if (resOk) resOk = FILE_Write(fileObjPtr, " ", 1);
	}

	if (fileObjPtr) FILE_Close(LOG_CONFIG_USER);

	return !fileObjPtr ? true : resOk;
}


/*******************************************************************************
 * @brief Loads the specified parameters from the operational param file.
 * nParams specifies how many parameters are provided, to be loaded
 * from one line in file.
 *******************************************************************************/
bool LOG_LoadOpParam(
		Device_dev device,
		LOG_Param param,
		LOG_Type log,
		uint8_t maxStringLen,
		uint8_t nParams, ...){

	bool resOk = true;
	char * stringPtr;
	char * bufPos;

	int memPos;
	uint32_t filePointer;
	va_list args;

	/* Search for string in param file */
	FIL * fileObjPtr = FindOpString(log, param, device, &bufPos, &filePointer);

	if (fileObjPtr){ /* File successfully opened */

		if (bufPos){ /* Operational parameter identifier found */

			va_start(args, nParams);

			for (int i = 0; i < nParams; i++){

				memPos = 0;
				stringPtr = va_arg(args, char *);

				/* Clear the provided memory */
				memset(&stringPtr[0], 0, maxStringLen);

				/* Read until new parameter, or end of buffer */
				while (*++bufPos != '\0' && *bufPos != ','){

					if (memPos >= maxStringLen) break;

					/* Load read values into the provided char arrays */

					*(stringPtr + memPos++) = *bufPos;
				}
			}
			va_end(args);
		}

		/* Parameter not found -> create it */
		else resOk = CreateParam(fileObjPtr, param, log, device);
	}
	return !fileObjPtr ? false : resOk;
}

/*******************************************************************************
 * @brief Saves the specified parameters to file, defined by the device type,
 * param and log type. nParams specifies how many parameters are provided to
 * be written on one line.
 *******************************************************************************/
bool LOG_SaveOpParam(
		Device_dev device,
		LOG_Param param,
		LOG_Type log,
		uint8_t nParams, ...){

	bool resOk = true;
	char * bufPos;
	char * stringPtr;

	uint32_t filePointer;
	uint8_t stringLen;
	va_list args;

	va_start(args, nParams);

	/* Search for string in param file */
	FIL * fileObjPtr = FindOpString(log, param, device, &bufPos, &filePointer);


	if (fileObjPtr && bufPos){ /* File successfully opened and string found */

		resOk = FILE_MoveToPosition(fileObjPtr, filePointer + 1);

		/* Make array of the string pointers and string lengths */

		for (int i = 0; i < nParams; i++){

			stringPtr = va_arg(args, char *);
			stringLen = strlen(stringPtr);

			if (resOk) resOk = FILE_Write(fileObjPtr, stringPtr, stringLen);
		}
	}

	va_end(args);
	return !fileObjPtr ? false : resOk;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Creates a new parameters line in file by writing the corresponding
 * string, and reserve a line length of LOG_MAX_LINE_LEN.
 *******************************************************************************/
static bool CreateParam(
		FIL * fileObjPtr,
		LOG_Param param,
		LOG_Type log,
		Device_dev device){

	bool resOk = true;
	char paramString[LOG_MAX_LINE_LEN] = {0};


	paramString[LOG_MAX_LINE_LEN - 3] = '\r';
	paramString[LOG_MAX_LINE_LEN - 2] = '\n';

	if (param == param_device){

		strcpy(paramString, DEVICE_TO_STRING(device));

		if (log == type_local) strcat(paramString, "_local:");

		else strcat(paramString, "_comb:");

		resOk = FILE_Write(fileObjPtr, paramString, LOG_MAX_LINE_LEN - 1);
	}

	else {

		strcpy(paramString, OP_PARAM_TO_STRING(param));
		resOk = FILE_Write(fileObjPtr, paramString, LOG_MAX_LINE_LEN - 1);
	}

	return resOk;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static FIL * FindUserString(
		LOG_UserParam_user param,
		Device_dev device,
		char ** bufPos,
		char * buf,
		uint32_t * filePointer){


	FIL * fileObjPtr = NULL;

	/* Open the parent directory */
	bool resOk = FILE_ChangeDir("/");

	if (resOk) fileObjPtr = FILE_OpenNew(LOG_CONFIG_USER);


	if (fileObjPtr){

		if (device != dev_noDevice) *bufPos = FILE_FindString(fileObjPtr,
				DEVICE_TO_STRING(device), buf, LOG_MAX_LINE_LEN, filePointer);


		else *bufPos = FILE_FindString(fileObjPtr,
				USER_PARAM_TO_STRING(param), buf, LOG_MAX_LINE_LEN, filePointer);
	}

	return fileObjPtr;
}


/*******************************************************************************
 * @brief Searches for the string corresponding to the provided arguments.
 * The provided bufPos will point to the next character after the string pattern.
 * The filePointer will point to the same character, but in the file, not in buffer.
 * @return Pointer to the FIL object structure.
 * Includes a static buffer to load the read string into
 *******************************************************************************/
static FIL * FindOpString(
		LOG_Type log,
		LOG_Param param,
		Device_dev device,
		char ** bufPos,
		uint32_t * filePointer){

	static char buffer[LOG_MAX_LINE_LEN];

	FIL * fileObjPtr = NULL;
	char matchString[25] = {0};

	/* Clear buffer */
	memset(&buffer[0], 0, strlen(buffer));

	/* Open the parent directory */
	bool resOk = FILE_ChangeDir("/");

	if (resOk) fileObjPtr = FILE_OpenNew(LOG_CONFIG);

	if (fileObjPtr){

		if (param == param_device){

			/* Assemble string to search for */
			strcpy(matchString, DEVICE_TO_STRING(device));

			if (log == type_local) strcat(matchString, "_local");

			else strcat(matchString, "_comb");

			/* Search for the provided "matchString" */

			*bufPos = FILE_FindString(fileObjPtr,
					matchString, buffer, LOG_MAX_LINE_LEN, filePointer);
		}

		else *bufPos = FILE_FindString(fileObjPtr,
				OP_PARAM_TO_STRING(param), buffer, LOG_MAX_LINE_LEN, filePointer);
	}
	return fileObjPtr;
}
