/******************************************************************************
 * @file:		stream.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	24. apr. 2016
 ******************************************************************************
 *	Provides Printf functionality to every FIFO stream available. Can also
 *	parse incoming data.
 ******************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "timer.h"
#include "err.h"

#include "stream.h"
#include "fifo_config.h"
#include "fifo_rx.h"
#include "fifo_tx.h"

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

static bool (*FlushPtr[N_STREAM_DEVICES])();

static Stream currentStream;

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Initializes the stdout buffer and the FIFO buffers
 *******************************************************************************/
void STREAM_Init(){

	/* Set temporary buffer for stdout to null, meaning no buffering will
	 * be done by the printf/sprintf functions */
	setbuf(stdout, NULL);

	FIFO_RX_Init();
	FIFO_TX_Init();
}


/*******************************************************************************
 * @brief If data available in stream, the Stream enumerator is returned.
 *******************************************************************************/
Stream STREAM_DataAvailable(){

	for (int i = 0; i < N_STREAM_DEVICES; i++){

		if (!FIFO_RX_Empty(i)){
			return i;
		}
	}
	return stream_none;
}


/*******************************************************************************
 * @brief Print the formatted data to stream. No float support!
 *******************************************************************************/
bool STREAM_Printf(Stream stream, const char * fmt, ...){

	va_list args;
	currentStream = stream;

	va_start(args, fmt);
	int ret = vprintf(fmt, args);
	va_end(args);

	return ret > 0 ? true : false;
}


/*******************************************************************************
 * @brief Print the formatted data to stream, and wait for a response match.
 * No float support!
 * @return > 0: The provided string matched with the received data.
 * return = 0: The provided string did not match.
 * return = -1: RX timeout, no data received. Raises new common exception.
 *******************************************************************************/
int STREAM_PrintfWaitResp(
		Stream stream,
		uint16_t timeout_ms,
		const char * exResp,
		const char * fmt, ...){

	int res;
	va_list args;
	currentStream = stream;

	va_start(args,fmt);

	vprintf(fmt, args);
	va_end(args);

	res = STREAM_WaitAltResponse(
			stream, timeout_ms, 1, exResp);

	return res;
}


/*******************************************************************************
 * @brief Read from specified stream until one of the provided strings are
 * matched/received, or timeout occurs.
 * All upper -and lower case characters will match.
 * @return > 0: Provided string found. String number found is returned.
 * return = 0: None of the provided strings matched.
 * return = -1: RX timeout, no data received. Raises new err_rxTimeout exception.
 *******************************************************************************/
int	STREAM_WaitAltResponse(
		Stream stream,
		uint16_t timeout_ms,
		uint8_t nAltResp, ...){

	if (nAltResp < 1) return -1;

	va_list args;
	int 	stringFound = -1;
	bool 	entireStringFound = false;
	bool 	charFound = false;

	uint8_t recvByte;


	char * 	stringPtr[nAltResp];
	uint8_t nCharsRecv[nAltResp];
	uint8_t stringLen[nAltResp];
	uint8_t nStrings = nAltResp;


	memset(&nCharsRecv[0], 0, sizeof(nCharsRecv));

	va_start(args, nAltResp);

	/* Make array of the string pointers and string lengths */
	while (nStrings){

		stringPtr[nAltResp - nStrings] = va_arg(args, char *);
		stringLen[nAltResp - nStrings] = strlen(stringPtr[nAltResp - nStrings]);
		nStrings--;
	}

	StartTimer_ms(timeout_ms);

	while (!entireStringFound && !TimeElapsed()){

		while(FIFO_RX_Empty(stream) && !TimeElapsed());

		if (!TimeElapsed() && stringFound < 0) stringFound = 0;

		while (!entireStringFound && FIFO_RX_Get(stream, &recvByte)){

			for (int i = 0; i < nAltResp; i++){

				if (*stringPtr[i] != '\0'){


					charFound = true;

					/* Lower case character to be found: match with upper case as well */
					if (*stringPtr[i] >= 'a' && *stringPtr[i] <= 'z'){

						if ((char)recvByte == *stringPtr[i] ||
								(char)recvByte == (*stringPtr[i] - 0x20)){

							nCharsRecv[i]++;
							stringPtr[i]++;
						}
						else charFound = false;
					}

					/* Upper case character to be found: match with lower case as well */
					else if (*stringPtr[i] >= 'A' && *stringPtr[i] <= 'Z'){

						if ((char)recvByte == *stringPtr[i] ||
								(char)recvByte == (*stringPtr[i] + 0x20)){

							nCharsRecv[i]++;
							stringPtr[i]++;
						}
						else charFound = false;
					}

					/* Match but no character */
					else if ((char)recvByte == *stringPtr[i]){

						nCharsRecv[i]++;
						stringPtr[i]++;
					}
					else charFound = false;


					/* The current match string was broken off */
					if (!charFound && i == stringFound - 1){
						stringFound = 0;
					}


					/* Character match for one of the provided strings */
					if (!stringFound && charFound && nCharsRecv[i]){
						stringFound = i + 1;
					}

					/* Entire match string received */
					if (nCharsRecv[i] == stringLen[i] && i == stringFound - 1){
						entireStringFound = true;
						break;
					}
				}
			}
		}
	}
	va_end(args);

	if (stringFound < 0) ERR_RAISE(err_rxTimeout, STREAM_TO_DEV(stream));

	if (!entireStringFound && stringFound > 0) stringFound = 0;

	return stringFound;
}


/*******************************************************************************
 * @brief Reads from the specified stream until numbers are read. A character
 * different from a digit ends the read, or if more than 10 digits are read.
 * @return >= 0: The value received on the stream.
 * return = -1: No digits received. Raises new err_rxTimeout exception.
 *******************************************************************************/
int32_t STREAM_GetValue(Stream stream, uint16_t timeout_ms){

	bool startOfValueFound = false;
	bool valueFound = false;

	uint8_t bytesRead = 0;
	const uint8_t bufLen = 11; /* Max number of digits in an int32 */

	char buf[bufLen];

	StartTimer_ms(timeout_ms);

	while (!valueFound && !TimeElapsed()){

		while (FIFO_RX_Empty(stream) && !TimeElapsed());

		while (!valueFound && FIFO_RX_Get(stream, (uint8_t*)(buf + bytesRead))){

			/* A valid digit is found ->
			 * consider start of value found */
			if (!startOfValueFound &&
					*(buf + bytesRead) >= '0' &&
					*(buf + bytesRead) <= '9'){

				startOfValueFound = true;
				bytesRead++;
			}

			/* An invalid digit found: consider value found */
			else if ((startOfValueFound &&
					(*(buf + bytesRead) < '0' ||
					*(buf + bytesRead) > '9')) ||
					bytesRead >= bufLen){

				valueFound = true;
				*(buf + bytesRead) = '\0';
			}

			/* Reading value */
			else if (startOfValueFound) bytesRead++;
		}
	}
	if (startOfValueFound) return atoi(buf);

	else {
		ERR_RAISE(err_rxTimeout, STREAM_TO_DEV(stream));
		return -1;
	}
}


/*******************************************************************************
 * @brief Read from specified stream until the end char is matched,
 *  the number of bytes is equal to bufLen, or timeout occurs. Starts loading
 *  the buffer when the start char is found.
 * @return = 1: String found. Buffer contains the null terminated string.
 * return = 0: The provided start char was not matched.
 * return = -1: RX timeout, no data received. Raises new err_rxTimeout exception.
 *******************************************************************************/
int STREAM_GetString(
		Stream stream,
		uint16_t timeout_ms,
		char start,
		char end,
		char * buf,
		uint8_t bufLen){

	int res = -1;
	bool stringFound = false;
	bool startFound = false;
	uint8_t bytesRead = 0;

	StartTimer_ms(timeout_ms);

	while (!stringFound && !TimeElapsed()){

		while (FIFO_RX_Empty(stream) && !TimeElapsed());

		res = 0;

		while (!stringFound && FIFO_RX_Get(stream, (uint8_t*)(buf + bytesRead))){

			/* Look for start character */
			if (!startFound && *(buf + bytesRead) == start) startFound = true;

			/* Look for end character */
			else if ((startFound && *(buf + bytesRead) == end) ||
					bytesRead >= bufLen){

				stringFound = true; /* Null terminate found string */
				*(buf + bytesRead) = '\0';
			}

			/* Receive character */
			else if (startFound) bytesRead++;
		}
	}
	if (res < 0) ERR_RAISE(err_rxTimeout, STREAM_TO_DEV(stream));

	return stringFound ? 1 : res;
}


/*******************************************************************************
 * @brief Flush data from specified stream, if pointer registered using
 * STREAM_SetFlushCallback
 *******************************************************************************/
bool STREAM_Flush(Stream stream){

	return FlushPtr[stream] ? FlushPtr[stream]() : false;
}


/*******************************************************************************
 * @brief Check for available data in the specified incoming stream.
 *******************************************************************************/
bool STREAM_DataAv(Stream stream){

	return !FIFO_RX_Empty(stream);
}


/*******************************************************************************
 * @brief Check if the specified outgoing stream is empty/ready to receive data.
 *******************************************************************************/
bool STREAM_Ready(Stream stream){

	return FIFO_TX_Empty(stream);
}


/*******************************************************************************
 * @brief Register function for flushing the stream.
 *******************************************************************************/
void STREAM_SetFlushCallback(Stream stream, bool (*callback)()){

	FlushPtr[stream] = callback;
}


/*******************************************************************************
 * @brief Function used by printf functions to write data to stream
 *******************************************************************************/
int _write(int fileHandle, char * buf, int len){

	int bytesWritten = 0;

	/* FIFO stream */
	if (currentStream < N_STREAM_DEVICES){

		for (bytesWritten = 0; bytesWritten < len; bytesWritten++){

			StartTimer_ms(FIFO_STREAM_TX_TIMEOUT);

			while (!FIFO_TX_Free(currentStream) && !TimeElapsed());

			if (TimeElapsed()) break;

			FIFO_TX_Put(currentStream, *buf++);
		}

		STREAM_Flush(currentStream);
	}

	/* Debug print */
	else {

		for (int x = 0; x < len; x++){

			ITM_SendChar(*buf++);
			Delay_ms(2);
		}
		bytesWritten = len;
	}

	if (bytesWritten < len) ERR_RAISE(err_txTimeout, STREAM_TO_DEV(currentStream));

	return bytesWritten;
}
