/******************************************************************************
 * @file:		stream_functions.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	12. mai 2016
 ******************************************************************************
 *	Stream functions provided to allow access to the incoming -and outgoing
 *	FIFO buffers.
 ******************************************************************************/


#include "system.h"

#include "stream_functions.h"

#include "fifo_tx.h"
#include "fifo_rx.h"
#include "stream.h"

#include "bmp_parser.h"
#include "tbr_driver.h"


static char * FormatSystemID();

static bool radioEndTransfer = false;
static uint8_t _waitAfterEndTransfer;
static uint16_t _bytesToTransfer;

/*******************************************************************************
 * @brief
 *******************************************************************************/
void STREAM_ResetRadioTransfer(){

	radioEndTransfer = false;
}


/*******************************************************************************
 * @brief Set the provided arguments for the coming data transmission.
 *******************************************************************************/
void STREAM_SetArgs(
		uint8_t waitAfterEndTransfer,
		uint16_t bytesToTransfer){

	_bytesToTransfer = bytesToTransfer;
	_waitAfterEndTransfer = waitAfterEndTransfer;

	radioEndTransfer = false;
}


/*******************************************************************************
 * @brief Transfers the provided data with length "bytesToTransfer" to the
 * radio tx fifo buffer. Enables transmission from fifo to the radio.
 *******************************************************************************/
uint STREAM_TransferDataToRadio(
		const uint8_t * data,
		uint bytesToTransfer){

	uint8_t nLinesTransferred = 0;
	uint16_t lineLen = 0;
	uint16_t bytesTransferred = 0;


	/* Enable transmit from FIFO to radio */
	STREAM_Flush(stream_radio);

	/* Sense call from caller function to determine if
	 * stream is ready to receive new data */
	if (!bytesToTransfer){

		if (FIFO_TX_Free(stream_radio) && !radioEndTransfer){
			bytesTransferred = 1;
		}
	}
	/* Transfer call */
	else {
		do {
			if (*data == '\n' && bytesTransferred != 0){

				/* Count lines of data transferred */
				if (!lineLen) lineLen = bytesTransferred + 1;
				nLinesTransferred++;

				/* End transfer if sending another line of data
				 * will make the packet larger than max size */
				if (lineLen * (nLinesTransferred + 1) > _bytesToTransfer){
					radioEndTransfer = true;

				}
			}

			FIFO_TX_Put(stream_radio, *data++);
			bytesTransferred++;

			if (radioEndTransfer || bytesTransferred == bytesToTransfer){

				BMP_EndTransfer(stream_radio, _waitAfterEndTransfer);
			}
		}
		while (
				bytesTransferred < bytesToTransfer &&
				FIFO_TX_Free(stream_radio) &&
				!radioEndTransfer);
	}

	return bytesTransferred;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
uint STREAM_TransferDataToGSM(
		const uint8_t * data,
		uint bytesToTransfer){


	uint16_t bytesTransferred = 0;

	/* Enable transmit from FIFO to radio */
	STREAM_Flush(stream_gsm);

	/* Sense call from caller function to determine if
	 * stream is ready to receive new data */
	if (bytesToTransfer == 0){

		if (FIFO_TX_Free(stream_gsm)){
			bytesTransferred = 1;
		}
	}
	/* Transfer call */
	else {
		do {

			FIFO_TX_Put(stream_gsm, *data++);
			bytesTransferred++;

			if (bytesToTransfer == bytesTransferred){
				BMP_EndTransfer(stream_gsm, _waitAfterEndTransfer);
			}

		}
		while (
				bytesTransferred < bytesToTransfer &&
				FIFO_TX_Free(stream_gsm));
	}

	return bytesTransferred;
}


/*******************************************************************************
 * @brief Receives one byte for each call, if data is accessible in the fifo
 * buffer.
 *******************************************************************************/
uint STREAM_ReceiveDataFromRadio(uint8_t * recvByte, uint sense){

	uint8_t retVal = 1;

	/* Sense call from caller function to determine if
	 * data is available */
	if (sense){

		if (FIFO_RX_Empty(stream_radio)) retVal = 0;
	}

	/* Receive call */
	else if (FIFO_RX_Get(stream_radio, recvByte)){

		if (*recvByte == BMP_EOT) retVal = 0xFF; 	/* End of transmission byte received */

		else retVal = 1; /* Data received */
	}

	/* Stream empty */
	else retVal = 0;

	return retVal;
}

/*******************************************************************************
 * @brief Receives one byte for each call, if data is accessible in the fifo
 * buffer.
 *******************************************************************************/
uint STREAM_ReceiveDataFromGsm(uint8_t * recvByte, uint sense){

	uint8_t retVal = 1;

	/* Sense call from caller function to determine if
	 * data is available */
	if (sense){ if (FIFO_RX_Empty(stream_gsm)) retVal = 0;}


	/* Receive call */
	else if (FIFO_RX_Get(stream_gsm, recvByte)){

		if (*recvByte == BMP_EOT) retVal = 0xFF; 	/* End of transmission byte received */

		else retVal = 1; /* Data received */
	}

	/* Stream empty */
	else retVal = 0;

	return retVal;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
uint STREAM_ReceiveDataFromTbr700(uint8_t * recvByte, uint sense){

	static int newLineCharAdded = 0;
	static char * uidBuf;
	static bool isFormattingUid = false;
	uint8_t retVal = 1;

	/* Sense call from caller function to determine if
	 * data is available */
	if (sense){ if (FIFO_RX_Empty(stream_tbr700)) retVal = 0; }


	/* End of line -> add newline char */
	else if (newLineCharAdded < 2){

		if (!newLineCharAdded){
			*recvByte = '\n';
			newLineCharAdded++;
		}

		else if (newLineCharAdded){
			newLineCharAdded++;
			/* Finished reading */
			retVal = 0xFF;
		}
	}

	/* Replace TBR ID with the system UID */
	else if (isFormattingUid && *uidBuf){
		*recvByte = *uidBuf++;
	}

	/* Receive call */
	else if (FIFO_RX_Get(stream_tbr700, recvByte)){

		TBR_IdentifyTemperatureData(recvByte);
		TBR_IdentifyFlashCode(recvByte);

		if (TBR_IdentifyNewTagData(recvByte)){

			/* New tag -or temperature data identified */

			/* Get system UID to replace TBR ID */
			isFormattingUid = true;
			uidBuf = FormatSystemID();
			*recvByte = *uidBuf++;
		}

		else if (isFormattingUid){

			int emptyCount = 0;

			while (*recvByte != ',' && emptyCount < 100000){

				if (!FIFO_RX_Get(stream_tbr700, recvByte)) emptyCount++;
				else emptyCount = 0;
				TBR_IdentifyFlashCode(recvByte);
			}

			/* Found comma -> end format */
			isFormattingUid = false;
		}

		else if (*recvByte == '\r') newLineCharAdded = 0;

		/* Finished reading (When reading from TBR memory)*/
		else if (*recvByte == 'F') retVal = 0xFF;
	}

	/* Stream empty */
	else retVal = 0;

	return retVal;
}

/*******************************************************************************
 * @brief Format the system UID.
 *******************************************************************************/
static char * FormatSystemID(){

	static char uidBuf[15];

	sprintf(uidBuf, "UID:%lu", *SYS_GetID());

	return &uidBuf[0];
}
