/******************************************************************************
 * @file:		system.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	24. feb. 2016
 ******************************************************************************
 * System module providing access to the System Info structures.
 ******************************************************************************/

#include <sys/time.h>
#include "stdarg.h"
#include <string.h>

#include "system.h"
#include "system_format.h"
#include "ee_printf.h"

#include "timer.h"
#include "err.h"

#include "fifo_tx.h"
#include "fifo_rx.h"
#include "stream.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

static uint8_t 		nRegisteredSystems = 1; /* This system is number 1 */
static uint8_t 		flags = 0;

static SystemInfo 	regSystems[N_MAX_SYSTEMS] = {{0}};

static uint32_t		totalTbrDetections = 0;
static uint16_t		newTbrDetections = 0;
static uint32_t		lastReadTbrFlashCode = 0;

static int 			dataAcqCounter = 0;
/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Sets the "enabled devices" variable in the system info structure.
 * If contact has been established with the GSM/GPRS modem, the system is assumed
 * to be a gateway node.
 *******************************************************************************/
void SYS_SetEnabledDevice(Device_dev dev){

	regSystems[0].enabledDevices |= (1 << dev);

	/* Contact with gprs -> assumed to be a gateway system */
	if (dev == dev_gsm) regSystems[0].gateway = true;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool SYS_GetEnabledDevice(Device_dev dev){

	return regSystems[0].enabledDevices & (1 << dev);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
uint8_t	SYS_GetnRegSystems(){

	return nRegisteredSystems;
}


/*******************************************************************************
 * @brief Sets the "faulty devices" variable in the system info structure.
 * If no contact has been established with the GPRS, the system is assumed
 * to be a router system.
 *******************************************************************************/
void SYS_SetFaultyDevice(Device_dev dev){

	regSystems[0].faultyDevices |= (1 << dev);

	/* No contact with gprs -> assumed to be a router system */
	if (dev == dev_gsm) regSystems[0].gateway = true;//false;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
bool SYS_GetFaultyDevice(Device_dev dev){

	return regSystems[0].faultyDevices & (1 << dev);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void SYS_SetFlags(uint8_t mask){

	regSystems[0].flags |= mask;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void SYS_ClearFlags(uint8_t mask){

	regSystems[0].flags &= ~mask;
}


/*******************************************************************************
 * @brief Initiates data acquisition from network slave nodes.
 *******************************************************************************/
void SYS_StartDataAcquisition(){


 	if (dataAcqCounter < SYS_DATA_ACQ_RETRIES){

		flags |= SYS_START_DATA_ACQ_MASK;
		dataAcqCounter++;
	}
	else {

		dataAcqCounter = 0;
		/* No response from any router node -> end session */
		flags |= SYS_END_ACQ_SESSION_MASK;
	}
}


/*******************************************************************************
 * @brief Call to report data acquisition from slave nodes finished.
 *******************************************************************************/
void SYS_DataAcqFinished(){

	flags |= SYS_DATA_ACQ_FINISHED_MASK;

	dataAcqCounter = 0;
}


/*******************************************************************************
 * @brief Get next available system task.
 * @return SYS_Task enumerator.
 *******************************************************************************/
SYS_Task SYS_GetTask(){

	SYS_Task task = task_ok;
	int mask = 1;
	bool maskMatch = false;
	static int count = 0;

	if (flags){

		while (!maskMatch && mask <= 0x80){

			maskMatch = true;

			switch(flags & mask){

			case SYS_START_DATA_ACQ_MASK:

				flags &= ~SYS_START_DATA_ACQ_MASK;
				task = task_startDataAcq;

				/* Callback if no data is received within the time limit */
				StartTimer_s(SYS_DATA_ACQ_TIMEOUT_S, SYS_StartDataAcquisition);
				break;

			case SYS_DATA_ACQ_FINISHED_MASK:

				flags &= ~SYS_DATA_ACQ_FINISHED_MASK;

				switch(count){

				case 0:
					task = task_pollTbrData;
					count++;
					break;

				case 1:
					task = task_pollGpsData;
					count++;
					break;

				case 2:
					task = task_pollGaugeData;
					count++;
					break;

				case 3:
					task = task_endDataAcqSession;
					count = 0;
					break;
				}

				break;

			case SYS_END_ACQ_SESSION_MASK:

				flags &= ~SYS_END_ACQ_SESSION_MASK;
				task = task_endDataAcqSession;
				break;

			default:
				maskMatch = false;
				break;
			}
			mask <<= 1;
		}
	}

	else if (!maskMatch){

		for (int i = 0; i < nRegisteredSystems; i++){

			if (regSystems[i].flags > SYS_CHARGING_MASK){

				task = task_sysParamAboveTH;
				maskMatch = true;
				break; /* For loop */
			}
		}
	}
	return task;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void SYS_UpdateOnBoardTemp(int8_t temp){

	regSystems[0].onBoardTemp_degC = temp;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void SYS_UpdateSeaTemp(int8_t temp){

	regSystems[0].seaTemp_degC = temp;
}


/*******************************************************************************
 * @brief Update last contact to the specified system. Set the current time stamp.
 *******************************************************************************/
void SYS_UpdateLastContact(uint8_t systemIndex){

	struct timeval time;
	gettimeofday(&time, NULL);

	regSystems[systemIndex].lastContact = time.tv_sec;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void SYS_UpdateGeoPos(GeoPosition * pos){

	struct timeval t;

	regSystems[0].lastPosition = *pos;

	gettimeofday(&t, NULL);
	regSystems[0].lastPosition.atTime = t.tv_sec;
}


/*******************************************************************************
 * @brief Adds the provided TBR detections to the system info variable.
 *******************************************************************************/
void SYS_AddTbrDetections(uint32_t rec){

	regSystems[0].tbrTagDetections += rec;
}


/*******************************************************************************
 * @brief Get or set current TBR detections.
 *******************************************************************************/
uint32_t * SYS_TbrDetections(uint32_t * detections){

	if (detections) regSystems[0].tbrTagDetections = *detections;

	return &regSystems[0].tbrTagDetections;
}


/*******************************************************************************
 * @brief Get or set current TBR flash code number.
 *******************************************************************************/
uint32_t * SYS_TbrFlashCode(uint32_t * flashCode){

	if(flashCode) lastReadTbrFlashCode = *flashCode;

	return &lastReadTbrFlashCode;
}


/*******************************************************************************
 * @brief Update the power info of the system.
 *******************************************************************************/
void SYS_UpdatePowerInfo(
		uint16_t battVoltage_mV,
		int16_t avgCurrent_mA,
		uint8_t stateOfCharge){

	regSystems[0].battVoltage_mV = battVoltage_mV;
	regSystems[0].avgCurrent_mA = avgCurrent_mA;
	regSystems[0].stateOfCharge = stateOfCharge;
}


/*******************************************************************************
 * @brief Get the unique ID of this system.
 *******************************************************************************/
uint32_t * SYS_GetID(){

	return &regSystems[0].id;
}


/*******************************************************************************
 * @brief Set the unique ID of this system.
 *******************************************************************************/
void SYS_SetID(uint32_t * uid){

	regSystems[0].id = *uid;
}


/*******************************************************************************
 * @brief Gets the unique ID of the system which functions as the gateway.
 *******************************************************************************/
uint32_t * SYS_GetGateway(){

	for (int i = 0; i < N_MAX_SYSTEMS; i++){
		if (regSystems[i].gateway) return &regSystems[i].id;
	}
	return NULL;
}


/*******************************************************************************
 * @brief Identify the provided system UID.
 * @return the system index, which is the array index of the System Info structures.
 *******************************************************************************/
uint8_t SYS_IdentifySystem(uint32_t addr){

	uint8_t index = 0;

	/* Use index 0 (this system) to index the GSM module */

	if (addr != STANDARD_UID_GSM){

		for (int i = 0; i < nRegisteredSystems; i++){

			if (regSystems[i].id == addr){
				index = i;
				break;
			}
		}

		if (!index){

			nRegisteredSystems++;

			if (nRegisteredSystems > N_MAX_SYSTEMS){} // TO MANY CLIENTS!

			index = nRegisteredSystems - 1;
			regSystems[index].id = addr;
		}
	}

	return index;
}


/*******************************************************************************
 * @brief Identify the provided system index.
 * @return the registered unique ID corresponding to the provided system index.
 *******************************************************************************/
uint32_t * SYS_IdentifyIndex(uint8_t index){

	return &regSystems[index].id;
}


/*******************************************************************************
 * @brief Print a short system info message to send by for example SMS.
 * The message is loaded into the specified stream.
 *******************************************************************************/
bool SYS_PrintShortSysInfo(Stream stream){

	bool res;
	int mask = 1;
	char buf[20];
	bool maskMatch = false;
	uint16_t meanBatVolt = 0;

	uint8_t flagCounts[6] = {0};


	/* Analyze system flags for each registered node and sum them up */

	for (int i = 0; i < nRegisteredSystems; i++){

		if (regSystems[i].flags > 1){

			while(!maskMatch && mask <= 0x80){

				maskMatch = true;

				switch(regSystems[i].flags & mask){

				case SYS_NOCONTACT_MASK:
					flagCounts[0]++;
					break;

				case SYS_CHRGERR_MASK:
					flagCounts[1]++;
					break;

				case SYS_FAULTY_DEV_MASK:
					flagCounts[2]++;
					break;

				case SYS_HIGHTEMP_MASK:
					flagCounts[3]++;

					break;

				case SYS_LOWSOC_MASK:
					flagCounts[4]++;
					break;

				case SYS_LOWVOLTAGE_MASK:
					flagCounts[5]++;
					break;

				default:
					maskMatch = false;
					break;
				}
				mask <<= 1;
			}

			if (maskMatch){

				break; /* For loop */
			}
		}
	}

	/* One SMS = max 160 characters */

	/* Send TBR tag detection info */
	if (newTbrDetections){
		STREAM_Printf(stream, "New tag detections since last check :%u. ", newTbrDetections);
		newTbrDetections = 0;
	}
	else STREAM_Printf(stream, "No new tag detections since last check. ");


	/* Get mean battery voltage */
	for (int i = 0; i < nRegisteredSystems; i++){

		if (regSystems[i].battVoltage_mV){
			meanBatVolt += regSystems[i].battVoltage_mV;
		}
	}
	meanBatVolt /= (nRegisteredSystems - 1);

	/* Send mean battery voltage for all nodes */
	STREAM_Printf(stream, "Mean voltage:");
	ee_sprintf(buf, "%.2fV. ", (double)meanBatVolt / 1000);
	STREAM_Printf(stream, buf);


	/* Send information about exceeded threshold values */

	if (flagCounts[0]) res = STREAM_Printf(stream, "Thres Last contact:%u. ", flagCounts[0]);
//	if (flagCounts[1]) res = STREAM_Printf(stream, "Thres ChargeErr:%u", flagCounts[1]);
//	if (flagCounts[2]) res = STREAM_Printf(stream, "Thres faulty dev:%u", flagCounts[2]);
//	if (flagCounts[3]) res = STREAM_Printf(stream, "Thres High temp:%u", flagCounts[3]);
//	if (flagCounts[4]) res = STREAM_Printf(stream, "Thres Low SOC:%u", flagCounts[4]);
	if (flagCounts[5]) res = STREAM_Printf(stream, "Thres Low Voltage:%u.", flagCounts[5]);

	return res;
}


/*******************************************************************************
 * @brief Serialize system info to specified stream.
 * Either ASCII representation or standard data format. Formatted includes a time stamp.
 *******************************************************************************/
void SYS_PrintSystemInfo(
		Stream stream,
		uint8_t systemIndex,
		bool format){


	if (format){ /* Serialize to ASCII representation */

		char buf[30];
		uint16_t nBytes = 0;

		STREAM_Printf(stream,"\r\nUID:%lu,Gateway:%d,",
				regSystems[systemIndex].id, regSystems[systemIndex].gateway);

		SYS_FormatDateTime(buf, &regSystems[systemIndex].lastContact);

		STREAM_Printf(stream, "Last Contact:%s,System flags:%d%d%d%d%d%d%d%d,", buf,
				BYTE_TO_BINARY(regSystems[systemIndex].flags));

		STREAM_Printf(stream, "Last position:");
		nBytes = ee_sprintf(buf + nBytes, "%.5f,", (double)regSystems[systemIndex].lastPosition.latitude / 10000000);
		nBytes += ee_sprintf(buf + nBytes,"%.5f,", (double)regSystems[systemIndex].lastPosition.longitude / 10000000);
		nBytes += ee_sprintf(buf + nBytes,"%.1fz", (double)regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm / 1000);

		STREAM_Printf(stream, "%s", buf);


		SYS_FormatDateTime(buf, &regSystems[systemIndex].lastPosition.atTime);
		STREAM_Printf(stream, ",at time:%s,", buf);

		ee_sprintf(buf, "Battery voltage:%.2fV,", (double)regSystems[systemIndex].battVoltage_mV / 1000);
		STREAM_Printf(stream, "%s", buf);

		STREAM_Printf(stream, "Avg. current:%dmA,SOC:%u%%,",
				regSystems[systemIndex].avgCurrent_mA,
				regSystems[systemIndex].stateOfCharge);

		STREAM_Printf(stream, "OnBoardTemp:%d\260\103,SeaTemp:%d\260\103,",
				regSystems[systemIndex].onBoardTemp_degC,
				regSystems[systemIndex].seaTemp_degC);

		STREAM_Printf(stream, "EnDev:%d%d%d%d%d%d%d%d%d,FauDev:%d%d%d%d%d%d%d%d%d,",
				regSystems[systemIndex].enabledDevices >> 8,
				BYTE_TO_BINARY(regSystems[systemIndex].enabledDevices),
				regSystems[systemIndex].faultyDevices >> 8,
				BYTE_TO_BINARY(regSystems[systemIndex].faultyDevices));

		STREAM_Printf(stream,"Total tagDet:%lu\r\n", regSystems[systemIndex].tbrTagDetections);

	}
	else { /* Serialize data as is */

		FIFO_TX_Put(stream, regSystems[systemIndex].id >> 24);
		FIFO_TX_Put(stream, regSystems[systemIndex].id >> 16);
		FIFO_TX_Put(stream, regSystems[systemIndex].id >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].id);

		FIFO_TX_Put(stream, regSystems[systemIndex].gateway);
		FIFO_TX_Put(stream, regSystems[systemIndex].flags);


		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.latitude >> 24);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.latitude >> 16);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.latitude >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.latitude);

		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.longitude >> 24);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.longitude >> 16);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.longitude >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.longitude);

		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm >> 24);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm >> 16);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm);

		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.atTime >> 24);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.atTime >> 16);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.atTime >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].lastPosition.atTime);

		FIFO_TX_Put(stream, regSystems[systemIndex].battVoltage_mV >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].battVoltage_mV);

		FIFO_TX_Put(stream, regSystems[systemIndex].avgCurrent_mA >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].avgCurrent_mA);

		FIFO_TX_Put(stream, regSystems[systemIndex].stateOfCharge);

		FIFO_TX_Put(stream, regSystems[systemIndex].onBoardTemp_degC);
		FIFO_TX_Put(stream, regSystems[systemIndex].seaTemp_degC);

		FIFO_TX_Put(stream, regSystems[systemIndex].enabledDevices >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].enabledDevices);

		FIFO_TX_Put(stream, regSystems[systemIndex].faultyDevices >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].faultyDevices);

		FIFO_TX_Put(stream, regSystems[systemIndex].tbrTagDetections >> 24);
		FIFO_TX_Put(stream, regSystems[systemIndex].tbrTagDetections >> 16);
		FIFO_TX_Put(stream, regSystems[systemIndex].tbrTagDetections >> 8);
		FIFO_TX_Put(stream, regSystems[systemIndex].tbrTagDetections);
	}
}


/*******************************************************************************
 * @brief De-serializes received data into the corresponding system info
 * structure. Reads directly from stream.
 *******************************************************************************/
bool SYS_UpdateSystemInfo(
		Stream stream,
		uint8_t systemIndex){

	uint8_t recvByte;
	int i = 0;
	bool res = true;

	StartTimer_ms(200);

	while (i < 37){

		while (!FIFO_RX_Get(stream, &recvByte) && !TimeElapsed());

		if (TimeElapsed()){
			res = false;
			goto exit;
		}

		/* Update data fields */
		switch(i++){
		case 0: regSystems[systemIndex].id 									 = recvByte << 24;			break;
		case 1: regSystems[systemIndex].id 									|= recvByte << 16;			break;
		case 2: regSystems[systemIndex].id 									|= recvByte << 8; 			break;
		case 3: regSystems[systemIndex].id 									|= recvByte; 				break;

		case 4: regSystems[systemIndex].gateway 							 = recvByte; 				break;
		case 5: regSystems[systemIndex].flags 								 = recvByte; 				break;

		case 6: regSystems[systemIndex].lastPosition.latitude  				 = recvByte << 24; 			break;
		case 7: regSystems[systemIndex].lastPosition.latitude 			   	|= recvByte << 16; 			break;
		case 8: regSystems[systemIndex].lastPosition.latitude 				|= recvByte << 8; 			break;
		case 9: regSystems[systemIndex].lastPosition.latitude 				|= recvByte;				break;

		case 10: regSystems[systemIndex].lastPosition.longitude  			 = recvByte << 24; 			break;
		case 11: regSystems[systemIndex].lastPosition.longitude 			|= recvByte << 16; 			break;
		case 12: regSystems[systemIndex].lastPosition.longitude 			|= recvByte << 8; 			break;
		case 13: regSystems[systemIndex].lastPosition.longitude 			|= recvByte; 				break;

		case 14: regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm  = recvByte << 24; 			break;
		case 15: regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm |= recvByte << 16;			break;
		case 16: regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm |= recvByte << 8;			break;
		case 17: regSystems[systemIndex].lastPosition.heightAboveMeanSea_mm |= recvByte;				break;

		case 18: regSystems[systemIndex].lastPosition.atTime 				 = recvByte << 24;			break;
		case 19: regSystems[systemIndex].lastPosition.atTime 				|= recvByte << 16;			break;
		case 20: regSystems[systemIndex].lastPosition.atTime 				|= recvByte << 8;			break;
		case 21: regSystems[systemIndex].lastPosition.atTime 				|= recvByte;				break;

		case 22: regSystems[systemIndex].battVoltage_mV  					 = recvByte << 8; 			break;
		case 23: regSystems[systemIndex].battVoltage_mV 					|= recvByte;				break;

		case 24: regSystems[systemIndex].avgCurrent_mA  					 = recvByte << 8;			break;
		case 25: regSystems[systemIndex].avgCurrent_mA 						|= recvByte;				break;

		case 26: regSystems[systemIndex].stateOfCharge  					 = recvByte;				break;

		case 27: regSystems[systemIndex].onBoardTemp_degC 					 = recvByte;				break;
		case 28: regSystems[systemIndex].seaTemp_degC 						 = recvByte;				break;

		case 29: regSystems[systemIndex].enabledDevices  					 = recvByte << 8;			break;
		case 30: regSystems[systemIndex].enabledDevices 					|= recvByte;				break;

		case 31: regSystems[systemIndex].faultyDevices  					 = recvByte << 8;			break;
		case 32: regSystems[systemIndex].faultyDevices 						|= recvByte;				break;

		case 33: regSystems[systemIndex].tbrTagDetections  					 = recvByte << 24;			break;
		case 34: regSystems[systemIndex].tbrTagDetections  					|= recvByte << 16;			break;
		case 35: regSystems[systemIndex].tbrTagDetections  					|= recvByte << 8;			break;
		case 36: regSystems[systemIndex].tbrTagDetections 					|= recvByte;				break;
		}
	}
exit:
	if (!res) ERR_RAISE(err_rxTimeout, STREAM_TO_DEV(stream));
	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
void SYS_InitSystemClocks(){

	/** Enable oscillators **/

	/* Start LF crystal osc. and wait until it is stable */
	CMU_OscillatorEnable(cmuOsc_LFXO, true, true);

	/** HF Clock **/

	/* Set HF RC osc. band */
	CMU_HFRCOBandSet(HFRCO_BAND);

	/* Set HFPER and HFCORE clock Div */
	CMU_ClockDivSet(cmuClock_HFPER, HFPER_DIV); /* Peripheral clock source */
	CMU_ClockDivSet(cmuClock_HF, HFCORE_DIV);	/* Core clock source */

	/* Enable clock to the interface of the low energy modules */
	CMU_ClockEnable(cmuClock_CORELE, true);


	/** LFA Clock **/
	/** Clock source for the RTC and the LETIMER **/

	/* Route the LFXO clock to the LFA clock branch */
	CMU_ClockSelectSet(cmuClock_LFA, cmuSelect_LFXO);

	/** LFB Clock **/
	/** Clock source for the LEUART **/

	/* Route the CORELE/2 clock to the LFB clock branch */
	CMU_ClockSelectSet(cmuClock_LFB, cmuSelect_CORELEDIV2);
}


/**************************************************************************//**
 * @brief Configure SWO - serial wire output
 *****************************************************************************/
#if DEBUG_OPERATION
void SYS_SetupSWOForPrint() {
	/* Enable GPIO clock. */
	CMU ->HFPERCLKEN0 |= CMU_HFPERCLKEN0_GPIO;

	/* Enable Serial wire output pin */
	GPIO ->ROUTE |= GPIO_ROUTE_SWOPEN;

	/* Set location 0 */
	GPIO ->ROUTE = (GPIO ->ROUTE
			& ~(_GPIO_ROUTE_SWLOCATION_MASK)) | GPIO_ROUTE_SWLOCATION_LOC0;

	/* Enable output on pin - GPIO Port F, Pin 2 */
	GPIO ->P[5].MODEL &=
			~(_GPIO_P_MODEL_MODE2_MASK);
	GPIO ->P[5].MODEL |= GPIO_P_MODEL_MODE2_PUSHPULL;

	/* Enable debug clock AUXHFRCO */
	CMU ->OSCENCMD = CMU_OSCENCMD_AUXHFRCOEN;

	/* Wait until clock is ready */
	while (!(CMU ->STATUS & CMU_STATUS_AUXHFRCORDY));

	/* Enable trace in core debug */
	CoreDebug ->DEMCR |=
			CoreDebug_DEMCR_TRCENA_Msk;

	ITM ->LAR = 0xC5ACCE55;
	ITM ->TER = 0x0;
	ITM ->TCR = 0x0;
	TPI ->SPPR = 2;
	TPI ->ACPR = 0xf;
	ITM ->TPR = 0x0;
	DWT ->CTRL = 0x400003FE;
	ITM ->TCR = 0x0001000D;
	TPI ->FFCR = 0x00000100;
	ITM ->TER = 0x1;
}
#endif


/*******************************************************************************
 * @brief Analyzes and compares the system info of all registered systems to
 * the defined threshold values.
 *******************************************************************************/
void SYS_AnalyzeSystemInfo(){

	struct timeval t;

	gettimeofday(&t, NULL);

	for (int i = 0; i < nRegisteredSystems; i++){

//		if (regSystems[i].enabledDevices == 0) continue; /* Skip the remote system (remote user) */

		/* On-board temperature */
		if (regSystems[i].onBoardTemp_degC > SYS_BOARD_TEMP_THRES){
			regSystems[i].flags |= SYS_HIGHTEMP_MASK;
		}
		else regSystems[i].flags &= ~SYS_HIGHTEMP_MASK;

		/* Last contact with system */
		if (i > 0 && regSystems[i].lastContact + SYS_LASTCONTACT_THRES < t.tv_sec){

			regSystems[i].flags |= SYS_NOCONTACT_MASK;
		}
		else regSystems[i].flags &= ~SYS_NOCONTACT_MASK;

		/* Battery voltage */
		if (regSystems[i].battVoltage_mV < BAT_VOLTAGE_LOW_THRES){

			regSystems[i].flags |= SYS_LOWVOLTAGE_MASK;
		}
		else regSystems[i].flags &= ~SYS_LOWVOLTAGE_MASK;

		/* Battery state of charge (%) */
		if (regSystems[i].stateOfCharge < BAT_SOC_LOW_THRES){

			regSystems[i].flags |= SYS_LOWSOC_MASK;
		}
		else regSystems[i].flags &= ~SYS_LOWSOC_MASK;

		/* TBR tag detections */
		totalTbrDetections += regSystems[i].tbrTagDetections;
	}

	newTbrDetections += totalTbrDetections;
	totalTbrDetections = 0;

	SYS_DataAcqFinished();
}
