/******************************************************************************
 * @file:		system_format.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. mai 2016
 ******************************************************************************
 *	Provides functions for file name -and time stamp formatting.
 ******************************************************************************/

#include <sys/time.h>
#include <string.h>

#include "system_format.h"


static struct tm *	calendar;


/*******************************************************************************
 * @brief Format of output string: DD/MM/YY HH:MM:SS. If usetime set to null,
 * the current system time will be used.
 * @return Number of bytes.
 *******************************************************************************/
uint16_t SYS_FormatDateTime(char * buf, time_t * usetime){


	struct timeval timeToConvert;

	if (usetime) timeToConvert.tv_sec = *usetime; /* Use provided time stamp */

	else gettimeofday(&timeToConvert, NULL); /* Get current time */

	calendar = localtime(&timeToConvert.tv_sec); /* Get calendar format of time */

	uint16_t nBytes = 0;

	nBytes  = sprintf(buf, "%02d/", calendar->tm_mday);
	nBytes += sprintf(buf + nBytes, "%02d/", 1 + calendar->tm_mon);
	nBytes += sprintf(buf + nBytes, "%d ", 1900 + calendar->tm_year - 2000);
	nBytes += sprintf(buf + nBytes, "%02d:", calendar->tm_hour);
	nBytes += sprintf(buf + nBytes, "%02d:", calendar->tm_min);
	nBytes += sprintf(buf + nBytes, "%02d", calendar->tm_sec);

	*(buf + nBytes) = '\0'; //?

	return nBytes;
}


/*******************************************************************************
 * @brief Gets the current UNIX time stamp in string format.
 *******************************************************************************/
uint16_t SYS_FormatTimeStamp(char * buf){

	struct timeval currentTime;
	gettimeofday(&currentTime, NULL); /* Get current time */

	return sprintf(buf, "%ld", currentTime.tv_sec);
}


/*******************************************************************************
 * @brief Date format: DDMMYY.
 * @return Number of bytes.
 *******************************************************************************/
uint16_t SYS_FormatDate(char * buf){

	struct timeval currentTime;

	gettimeofday(&currentTime, NULL); /* Get current time */
	calendar = localtime(&currentTime.tv_sec); /* Get calendar format of current time */

	uint16_t nBytes = 0;

	nBytes  = sprintf(buf, "%02d", calendar->tm_mday);
	nBytes += sprintf(buf + nBytes, "%02d", 1 + calendar->tm_mon);
	nBytes += sprintf(buf + nBytes, "%d", 1900 + calendar->tm_year - 2000);

	return nBytes;
}


/*******************************************************************************
 * @brief Creates a new file name for the specified log, either local(true) or
 * combined (false), and the device log. If date (DDMMYY) is provided,
 * the function use that instead of the current date.
 *******************************************************************************/
char * SYS_MakeFileName(bool local, Device_dev deviceLog, char * date){


	static char path[FILENAME_MAX_SIZE]; /* Small file name */


	if (date && strlen(date) <= 6) strcpy(&path[2], date);

	else if (!date) SYS_FormatDate(path + 2);

	else return NULL;


	strcpy(path + 8, ".txt");

	if (local) path[0] = 'L';
	else path[0] = 'C';

	path[1] = ('0' + deviceLog);

	return &path[0];
}


/*******************************************************************************
 * @brief Convert file name to accommodate the current naming convention.
 *******************************************************************************/
void SYS_FormatReceivedFileName(char * fileName){

	fileName[0] = 'C';
}


/*******************************************************************************
 * @brief Input date format: DDMMYY. Adds one day to the date.
 *******************************************************************************/
void SYS_GetNextFile(char * fileName){

	time_t inputTime;
	int i, j = 0;
	char * firstElemBuf = fileName;

	struct tm calendar = (struct tm){0};

	while (*fileName != '\0' && *fileName != '.'){

		while (j < 2){
			if (i == 0) calendar.tm_mday = 10 * calendar.tm_mday + *fileName - '0';
			if (i == 1) calendar.tm_mon = 10 * calendar.tm_mon + *fileName - '0';
			if (i == 2) calendar.tm_year = 10 * calendar.tm_year + *fileName - '0';
			fileName++;
			j++;
		}
		fileName++;
		i++;
		j = 0;
	}

	calendar.tm_year += 2000;
	calendar.tm_mon -= 1;

	inputTime = mktime(&calendar); /* Get input time in UNIX format */
	inputTime += 24 * 3600; /* Add one day */

	calendar = *localtime(&inputTime); /* Get calendar format of next day */

	uint16_t nBytes = 0;

	nBytes  = sprintf(firstElemBuf, "%02d", calendar.tm_mday);
	nBytes += sprintf(firstElemBuf + nBytes, "%02d", 1 + calendar.tm_mon);
	nBytes += sprintf(firstElemBuf + nBytes, "%d", 1900 + calendar.tm_year - 2000);
}

