/******************************************************************************
 * @file:		time_manager.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	6. feb. 2016
 ******************************************************************************
 *	The time manager manages the real time counter (RTC) and the system clock,
 *	and incorporates a task scheduler that calculates compare values for the
 *	next system wake up event. The major events are the wake up and shutdown
 *	of the different devices. In addition, other events are also available, and
 *	new custom events can easily be added.
 ******************************************************************************/

#include <stdio.h>
#include <sys/time.h>

#include "time_manager.h"
#include "device_settings.h"

/*******************************************************************************
 * 							Private function declarations
 *******************************************************************************/

static void 	InitRTC();
static bool 	ScheduleNextWakeUp();
static bool	 	ScheduleNextShutdown();
static void 	SetMidnight(const struct timeval * time);
static void 	SyncScheduler();
static uint64_t GetNextWakeUpCompareVal();
static uint32_t *GetNextShutdownCompareVal();

/*******************************************************************************
 * 							Variable declarations
 *******************************************************************************/

/*******************************************************************************
 * @brief Time manager status flags.
 * Bit explanation:
 * 	bit22	   bit21		 bit20		 bit19		 bit18		 bit17 - bit0
 *			  Watchdog   ind timeout	Midnight   TempComp	   Activate device
 *
 * Activate device: Time to activate(wake up) or de-activate(shut down) device.
 * Bit nr. 2*n signals wake up, while bit 2*n+1 signals shut down,
 * where n = device.
 * Temp comp: Time for RTC temperature compensation.
 * Midnight:
 * ind timeout: If in normal mode, LEDs will be de-activated.
 * Watchdog: Check connection to devices etc.
 *******************************************************************************/
static uint32_t 	flags = 0;

static uint16_t 	activeDevices = 0; 	/* To keep control of which devices are currently awake */
static uint32_t 	wakeUpFlagsToSet = 0, shutdwnFlagsToSet = 0;

static time_t 		midnightTime;

/* Compensation for the scheduler to keep in sync with UTC time */
static uint32_t		schedulerComp_neg = 0, schedulerComp_pos = 0;
static bool			doSchedulerSync = false;

static uint32_t 	rtcrflowCounter = 0;
static uint32_t 	intervalCount[N_DEVICES + TIME_N_MISC_INTERVALS] = {0};
static uint32_t		shdnStartCounts[N_DEVICES][2] = {{0},{0}};


/* Other tasks than wake up and shutdown of devices */
static uint32_t 	miscIntervals[TIME_N_MISC_INTERVALS] = {
		TIME_IND_TIMEOUT,
		TIME_MIDNIGHT_INTERVAL,
		TIME_TEMP_COMP_INTERVAL,
		TIME_WATCHDOG_INTERVAL
};

/* Function pointers */
static bool 	 (*GetEnabled)(Device_dev dev);
static uint16_t *(*GetOnPeriod)(Device_dev dev);
static uint32_t *(*GetInterval)(Device_dev dev);

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Update the system time since the UNIX epoch. Timezone not supported.
 *******************************************************************************/
int settimeofday(const struct timeval * time, const struct timezone * tz){

	static uint16_t count = 0;
	static bool onInit = true;

	clockSetStartTime(time);
	SetMidnight(time);

	/* Do scheduler sync with the sync scheduler rate */
	if (!count || !(count % SYNC_SCHEDULER_RATE)) doSchedulerSync = true;

	if (onInit){
		RTC_CompareSet(0, RTC->CNT + RTC_TICKS_PR_SEC);
		onInit = false;
	}
	count++;
	return 0;
}


/*******************************************************************************
 * @brief Initiates all of the clocks and timers.
 *******************************************************************************/
void TIME_Init(struct timeval * inittime){

	/* Initialize the temperature compensated calendar */
	Clock_Init_TypeDef init;

	if (inittime != NULL) init.startTime = *inittime;

	else {
#if (DEBUG_OPERATION > 1)
		struct tm defaultCalendar  = CLOCK_DEFAULT_START_DATE;
		init.startTime.tv_sec = mktime(&defaultCalendar);
		init.startTime.tv_usec = 0;
#else
		init.startTime.tv_sec = 0;
#endif
	}
	init.rtcCountsPerSec = RTC_TICKS_PR_SEC;
	init.tempCompInterval = TIME_TEMP_COMP_INTERVAL;

	clockInit(&init);

	InitRTC();
#if (DEBUG_OPERATION > 1)
	SetMidnight();
#endif

	/* Set initial radio wake up */
	flags |= 1 << 2 * dev_radio;
	activeDevices |= dev_radio;


	/* Set function pointers to get information about device wake up and shut down */
    GetEnabled = DEV_GetEnabledDevice;
    GetInterval = DEV_GetDeviceInterval;
    GetOnPeriod = DEV_GetDeviceOnPeriod;
}


/*******************************************************************************
 * @brief Enters EM1 or EM2 sleep, where EM2 is deep sleep with only LFXO enabled.
 *******************************************************************************/
void TIME_Sleep(bool light){

	if (!light){
		/* Enter EM2 with no clock restore */
		EMU_EnterEM2(false);
	}
	else EMU_EnterEM1();
}


/*******************************************************************************
 * @brief Executes a crystal temperature compensation based on the built in MCU
 * temperature sensor. Returns the temperature reading.
 *******************************************************************************/
int8_t TIME_DoTempComp(){

	return clockDoTemperatureCompensation();
}


/*******************************************************************************
 * @brief Get next event generated by the scheduler.
 *******************************************************************************/
TIME_Event TIME_GetEvent(Device_dev * device){

	TIME_Event nextEvent = event_none;

	if (flags & TIME_MIDNIGHT_MASK){

		flags &= ~TIME_MIDNIGHT_MASK;
		nextEvent = event_midnight;
	}
	else if (flags & TIME_TEMP_COMP_MASK){

		flags &= ~TIME_TEMP_COMP_MASK;
		nextEvent = event_rtcTempComp;
	}
	else if (flags & TIME_IND_TIMEOUT_MASK){

		flags &= ~TIME_IND_TIMEOUT_MASK;
		nextEvent = event_indicatorTimeout;
	}
	else if (flags & TIME_WATCHDOG_MASK){

		flags &= ~TIME_WATCHDOG_MASK;
		nextEvent = event_watchdog;
	}
	else {

		for (int i = 0; i < N_DEVICES; i++){

			/* Wake up */
			if (flags & (1 << 2 * i)){

				*device = i;
				flags &= ~(1 << (2 * i));
				nextEvent = event_wakeupDevice;
				break;
			}

			/* Shut down */
			else if (flags & (1 << (2 * i + 1))){

				*device = i;
				flags &= ~(1 << (2 * i + 1));
				nextEvent = event_shutdownDevice;
				break;
			}
		}
	}
	return nextEvent;
}


/*******************************************************************************
 * 							Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void InitRTC(){

	CMU_ClockEnable(cmuClock_RTC, true);

	/* Prescaler of 32768 = 1 sek ticks */
	CMU_ClockDivSet(cmuClock_RTC, RTC_CLK_DIV);

	RTC_Init_TypeDef rtcInit;
	rtcInit.enable = true;
	rtcInit.comp0Top = false;
	rtcInit.debugRun = false;


	/* COMP0 interrupt to trigger device wake up */
	/* COMP1 interrupt to trigger device shutdown */
	/* OF interrupt to keep track of counter overflows */
	RTC_IntEnable(RTC_IEN_OF | RTC_IEN_COMP0);
 	NVIC_EnableIRQ(RTC_IRQn);

 	#if (DEBUG_OPERATION > 1)
 	RTC_CompareSet(0, RTC_TICKS_PR_SEC);
#else
 	RTC_CompareSet(0, RTC_LEN);
#endif
	/* Initialize the RTC */
	RTC_Init(&rtcInit);
}


/*******************************************************************************
 * @brief Updates the midnight time.
 *******************************************************************************/
static void SetMidnight(const struct timeval * time){

	struct timeval currentTime;
	const struct timeval * currentTimePtr;

	if (time) currentTimePtr = time;

	else {
		gettimeofday(&currentTime, NULL); 		/* Get current time */
		currentTimePtr = &currentTime;
	}

	struct tm * calendar = localtime(&currentTimePtr->tv_sec); 	/* Get calendar format of current time */

	struct tm midnightTimeCal = *calendar;

	midnightTimeCal.tm_hour = 0;
	midnightTimeCal.tm_min = 0;
	midnightTimeCal.tm_sec = 0;

	/* Add one day */
	midnightTime = mktime(&midnightTimeCal) + 3600 * 24;

	/* Set next midnightTime wake up */
	miscIntervals[TIME_MIDNIGHT_INTERVAL_POS] = midnightTime - currentTimePtr->tv_sec;
}


/*******************************************************************************
 * @brief Sync the scheduler to the newly updated system time. Uses the radio
 * device as a reference to calculate the scheduler compensation.
 *******************************************************************************/
static void SyncScheduler(){

	/* Only sync when the radio wake up flag is set */
	if (flags & 1 << (2 * dev_radio)){

		struct timeval currentTime;
		gettimeofday(&currentTime, NULL); 	/* Get current time */

		time_t secToMidnight = midnightTime - currentTime.tv_sec;
		int i = 0;

		/* Find how many radio intervals there is until midnight */
		while (*GetInterval(dev_radio) * i++ < secToMidnight);


		if (*GetInterval(dev_radio) * --i - secToMidnight >= *GetInterval(dev_radio) / 2){

			/* Positive compensation if the compensation value is greater than half the interval */
			schedulerComp_pos += (secToMidnight - *GetInterval(dev_radio) * --i);
		}

		else {
			schedulerComp_neg += (*GetInterval(dev_radio) * i - secToMidnight);
		}

		doSchedulerSync = false;
	}
}


/*******************************************************************************
 * @brief Calculates the next wake up compare value to be set.
 *******************************************************************************/
static uint64_t GetNextWakeUpCompareVal(){

	uint32_t smallestCompareVal = 0xFFFFFFFF;
	uint32_t cnt = RTC->CNT;

	/* Adjusts interval counts if intervals have been changed,
	 * or a device has been enabled */
	for (int i = 0; i < N_DEVICES; i++){

		/* Ignore device if not enabled, or interval set to 0 */
		if (GetEnabled(i) && *GetInterval(i)){

			while ((schedulerComp_pos + *GetInterval(i) * intervalCount[i]) * RTC_TICKS_PR_SEC >
					cnt + rtcrflowCounter * RTC_LEN + schedulerComp_neg * RTC_TICKS_PR_SEC){

				if (intervalCount[i] == 0) break;
				else intervalCount[i]--;
			}

			while ((schedulerComp_pos + *GetInterval(i) * intervalCount[i]) * RTC_TICKS_PR_SEC <=
					cnt + rtcrflowCounter * RTC_LEN + schedulerComp_neg * RTC_TICKS_PR_SEC){

				intervalCount[i]++;
			}
		}
	}
	/* Adjust interval counts for misc. intervals if changed, or schedulerComps have changed */
	for (int i = 0; i < TIME_N_MISC_INTERVALS; i++){

		/* Ignore if interval is set to 0 */
		if (miscIntervals[i]){

			while ((schedulerComp_pos + miscIntervals[i] * intervalCount[N_DEVICES + i]) * RTC_TICKS_PR_SEC >
					cnt + rtcrflowCounter * RTC_LEN + schedulerComp_neg * RTC_TICKS_PR_SEC){

				if (intervalCount[N_DEVICES + i] == 0) break;
				else intervalCount[N_DEVICES + i]--;
			}

			while ((schedulerComp_pos + miscIntervals[i] * intervalCount[N_DEVICES + i]) * RTC_TICKS_PR_SEC <=
					cnt + rtcrflowCounter * RTC_LEN + schedulerComp_neg * RTC_TICKS_PR_SEC){

				intervalCount[N_DEVICES + i]++;
			}
		}
	}

	/* Identify the device which is closest to wake up */
	for (int i = 0; i < N_DEVICES; i++){

		/* Ignore device if not enabled, or interval set to 0 */
		if (GetEnabled(i) && *GetInterval(i)){

			if (*GetInterval(i) * intervalCount[i] +
					schedulerComp_pos - schedulerComp_neg < smallestCompareVal){

				smallestCompareVal =
						*GetInterval(i) * intervalCount[i] + schedulerComp_pos - schedulerComp_neg;
			}
		}
	}

	/* Check if any of the misc. intervals are closer */
	for (int i = 0; i < TIME_N_MISC_INTERVALS; i++){

		/* Ignore if interval is set to 0 */
		if (miscIntervals[i]){

			if (miscIntervals[i] * intervalCount[N_DEVICES + i] +
					schedulerComp_pos - schedulerComp_neg <= smallestCompareVal){

				smallestCompareVal =
						miscIntervals[i] * intervalCount[N_DEVICES + i] + schedulerComp_pos - schedulerComp_neg;

				/* Add a flag to set on next wake up */
				wakeUpFlagsToSet |= 1 << (2 * N_DEVICES + i);
			}
			/* Clear any flags that may be set */
			else wakeUpFlagsToSet &= ~1 << (2 * N_DEVICES + i);
		}
	}


	/* Identify other devices that may have the same wake up time */
	for (int i = 0; i < N_DEVICES; i++){

		/* Ignore device if not enabled, or interval set to 0 */
		if (GetEnabled(i) && *GetInterval(i) != 0){

			if (*GetInterval(i) * intervalCount[i] +
					schedulerComp_pos - schedulerComp_neg <= smallestCompareVal){

				/* Add a flag to set on next wake up */
				wakeUpFlagsToSet |= (1 << 2 * i);
			}
			/* Clear any flags that may be set */
			else wakeUpFlagsToSet &= ~(1 << 2 * i);
		}
	}

	return
			smallestCompareVal * RTC_TICKS_PR_SEC -
			rtcrflowCounter * RTC_LEN;
}


/*******************************************************************************
 * @brief Calculates the next shutdown compare value to be set.
 *******************************************************************************/
static uint32_t * GetNextShutdownCompareVal(){

	static uint32_t smallestCompareVal;
	uint16_t * shortestOnPeriod;

	smallestCompareVal = 0xFFFFFFFF;

	/* Identify the device which is closest to shut down */
	for (int i = 0; i < N_DEVICES; i++){

		/* Only consider currently active devices */
		if (activeDevices & (1 << i)){

			/* Set start count to counter value if device was just woken up */
			if (shdnStartCounts[i][0] == 0){

				shdnStartCounts[i][0] = RTC->CNT; 			/* Set start counter value */
				shdnStartCounts[i][1] = rtcrflowCounter; /* Set start overflow counter value */
			}

			if (	/* Identify the device that is closest to shut down */
					shdnStartCounts[i][0] +
					*GetOnPeriod(i) * RTC_TICKS_PR_SEC -
					(rtcrflowCounter - shdnStartCounts[i][1]) * RTC_LEN
					< smallestCompareVal){

				smallestCompareVal =
						shdnStartCounts[i][0] +
						*GetOnPeriod(i) * RTC_TICKS_PR_SEC -
						(rtcrflowCounter - shdnStartCounts[i][1]) * RTC_LEN;

				shortestOnPeriod = GetOnPeriod(i);
			}
		}
	}

	/* Identify other devices that may have the same shut down time */
	for (int i = 0; i < N_DEVICES; i++){

		/* Only consider currently active devices */
		if (activeDevices & (1 << i)){

			if (
					shdnStartCounts[i][0] +
					*GetOnPeriod(i) * RTC_TICKS_PR_SEC -
					(rtcrflowCounter - shdnStartCounts[i][1]) * RTC_LEN
					<= smallestCompareVal){

				/* Add a flag to set on next wake up */
				shutdwnFlagsToSet |= (1 << (2 * i + 1));

			} /* Clear any flags that may already be set */
			else shutdwnFlagsToSet &= ~(1 << (2 * i + 1));
		}
	}

	if (*shortestOnPeriod == 0) smallestCompareVal = 0;
	return &smallestCompareVal;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool ScheduleNextWakeUp(){

	uint64_t nextCompareVal;
	bool res = true;

	/* Sync the scheduler to midnight time if system time has been updated */
	if (doSchedulerSync) SyncScheduler();

	/* Calculate next wake up */
	nextCompareVal = GetNextWakeUpCompareVal();

	/* If next compare value will result in overflow */
	if (nextCompareVal > RTC_LEN) res = false;

	/* Set wake up immediately if compare value is smaller than current count */
	else if (nextCompareVal <= RTC->CNT) RTC_IntSet(RTC_IEN_COMP0);

	/* Set next wake up compare value */
	else RTC_CompareSet(0, nextCompareVal);

	return res;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool ScheduleNextShutdown(){

	uint32_t * nextCompareVal;
	bool res = true;

	if (activeDevices){

		/* Calculate next shut down */
		nextCompareVal = GetNextShutdownCompareVal();

		/* If device(s) to be shut down immediately after task finished (cycle time is set to 0) */
		if (*nextCompareVal == 0) RTC_IntSet(RTC_IEN_COMP1);

		/* If next compare value will result in overflow */
		else if (*nextCompareVal > RTC_LEN) res = false;

		/* To be sure wake up will happen */
		else if (*nextCompareVal <= RTC->CNT) RTC_IntSet(RTC_IEN_COMP1);

		/* Set next shut down compare value */
		else RTC_CompareSet(1, *nextCompareVal);
	}
	else res = false;

	return res;
}


/*******************************************************************************
 * 							Interrupt handlers
 *******************************************************************************/


void RTC_IRQHandler(){

	INT_Disable();

	/*****************************************************
	 * Wake up interrupt for devices and misc. intervals.
	 *****************************************************/
	if (RTC_IntGet() & RTC_IF_COMP0){

		RTC_IntClear(RTC_IFC_COMP0);

		/* Set the wake-up flags calculated on the previous wake-up */
		flags |= wakeUpFlagsToSet;

		for (int i = 0; i < N_DEVICES; i++){

			/* Check if wake up flag set */
			if (wakeUpFlagsToSet & (1 << 2 * i)){

				/* Increase interval count */
				intervalCount[i]++;

				/* Clear wake up flag */
				wakeUpFlagsToSet &= ~(1 << 2 * i);

				/* Flag active device */
				activeDevices |= (1 << i);
			}
		}

		/* Update interval count for misc. intervals */
		for (int i = 0; i < TIME_N_MISC_INTERVALS; i++){

			/* Check if wake up flag set */
			if (wakeUpFlagsToSet & 1 << (2 * N_DEVICES + i)){

				/* Increase interval count */
				intervalCount[N_DEVICES + i]++;

				/* Clear wake up flag */
				wakeUpFlagsToSet &= ~1 << (2 * N_DEVICES + i);

				/* Reset midnight interval (because of midnight sync) */
				if (i == TIME_MIDNIGHT_INTERVAL_POS) miscIntervals[i] = TIME_MIDNIGHT_INTERVAL;

				/* Disable "LED timeout" after first wake up */
				if (i == TIME_IND_TIMEOUT_POS) miscIntervals[i] = 0;
			}
		}

		/* Calculate next shut down compare value */
		if (ScheduleNextShutdown()){
			RTC_IntEnable(RTC_IEN_COMP1);
		}

		else RTC_IntDisable(RTC_IEN_COMP1);

		/* Calculate next wake up compare value */
		if (!ScheduleNextWakeUp()){
			RTC_IntDisable(RTC_IEN_COMP0);
		}
	}

	/*****************************************************
	 * Shut down interrupt for devices.
	 *****************************************************/
	if (RTC_IntGet() & RTC_IF_COMP1){

		RTC_IntClear(RTC_IFC_COMP1);

		/* Flag that the device(s) can be shut down */
		flags |= shutdwnFlagsToSet;

		for (int i = 0; i < N_DEVICES; i++){

			/* Check if shutdown flag is set */
			if (shutdwnFlagsToSet & (1 << (2 * i + 1))){

				/* Flag device as inactive */
				activeDevices &= ~(1 << i);

				/* Clear shutdown flag */
				shutdwnFlagsToSet &= ~(1 << (2 * i + 1));

				/* Reset start count */
				shdnStartCounts[i][0] = 0;
			}
		}

		/* Disable interrupt if no more active devices */
		if (!ScheduleNextShutdown()){
			RTC_IntDisable(RTC_IEN_COMP1);
		}
	}

	/*****************************************************
	 * Counter overflow
	 *****************************************************/
	if (RTC_IntGet() & RTC_IF_OF){

		RTC_IntClear(RTC_IFC_OF);

		clockOverflow();
		rtcrflowCounter++;


		if (~RTC_IntGet() & RTC_IEN_COMP0){

			/* Schedule wake up if COMP0 interrupt has been disabled */
			if (ScheduleNextWakeUp()){

				/* Enable interrupt */
				RTC_IntEnable(RTC_IEN_COMP0);
			}
		}

		if (~RTC_IntGet() & RTC_IEN_COMP1){

			/* Schedule next shut down if COMP1 interrupt has been disabled */
			if (ScheduleNextShutdown()){

				/* Enable interrupt */
				RTC_IntEnable(RTC_IEN_COMP1);
			}
		}
	}

	INT_Enable();
}
