 /******************************************************************************
 * @file:		timer.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	28. jan. 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include "em_int.h"
#include "shared.h"

#include "timer.h"

#if TIMER_USE_LETIMER
static void 	InitLETimer();
#endif

static void 	(*OnElapsedLEFunc)(), (*OnElapsedFunc)();

static bool 	timer0timeout, timer2timeout;
static bool 	timer0started = false, timer1started = false;
static uint32_t timer0val, timer1val, timer2val;

/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Initialize TIMER0 and TIMER1.
 *******************************************************************************/
void Timer_Init(){

#if TIMER_USE_LETIMER
	InitLETimer();
#endif

	/* Enable clock for timers */
	CMU_ClockEnable(cmuClock_TIMER0, true);
	CMU_ClockEnable(cmuClock_TIMER1, true);
	CMU_ClockEnable(cmuClock_TIMER2, true);

	TIMER_Init_TypeDef timerInit = {
	       .enable     = false,
	       .debugRun   = false,
	       .prescale   = timerPrescale1024,
	       .clkSel     = timerClkSelHFPerClk,
	       .fallAction = timerInputActionNone,
	       .riseAction = timerInputActionNone,
	       .mode       = timerModeUp,
	       .dmaClrAct  = false,
	       .quadModeX4 = false,
	       .oneShot    = false,
	       .sync       = false,
	};

	/* Enable overflow interrupt */
	TIMER0->IEN |= TIMER_IEN_OF;
	TIMER1->IEN |= TIMER_IEN_OF;
//	TIMER2->IEN |= TIMER_IEN_OF;

	/* Enable interrupt handlers */
	NVIC_EnableIRQ(SysTick_IRQn);
	NVIC_EnableIRQ(TIMER0_IRQn);
	NVIC_EnableIRQ(TIMER1_IRQn);
//	NVIC_EnableIRQ(TIMER2_IRQn);

	/* Configure timers */
	TIMER_Init(TIMER0, &timerInit);
	TIMER_Init(TIMER1, &timerInit);
	TIMER_Init(TIMER2, &timerInit);

	CMU_ClockEnable(cmuClock_TIMER0, false);
	CMU_ClockEnable(cmuClock_TIMER1, false);
	CMU_ClockEnable(cmuClock_TIMER2, false);
}


/*******************************************************************************
 * @brief Starts TIMER0. Call "TimerElapsed" to check if the time has elapsed.
 * Use with care as lower level implementations exist.
 *******************************************************************************/
void StartTimer_ms(uint16_t millis){

	if (!timer0started){
		/* Enable clock */
		CMU_ClockEnable(cmuClock_TIMER0, true);
		timer0started = true;
	}

	TIMER0->CNT = 0;

	/* Set top value to provided timeout value if smaller than the timer length */
	if (TIMER_MULTI * millis < TIMER_LEN){
		TIMER0->TOP = TIMER_MULTI * millis;
		timer0val = 0;
	}

	else {

		TIMER0->TOP = TIMER_LEN;
		timer0val = TIMER_MULTI * millis - TIMER_LEN;
	}

	TIMER0->CMD = TIMER_CMD_START;

	timer0timeout = false;
}


/*******************************************************************************
 * @brief Checks if time set in StartTimer_ms for TIMER0 has elapsed.
 *******************************************************************************/
bool TimeElapsed(){

	return timer0timeout;
}


/*******************************************************************************
 * @brief Starts TIMER1 and sets the specified timeout in seconds. Provided
 * callback function will be called on timeout.
 * Mostly used for timeout when gateway is polling nodes for data.
 *******************************************************************************/
void StartTimer_s(uint16_t sec, void(*OnElapsed)){


	if (!timer1started){
		/* Enable clock */
		CMU_ClockEnable(cmuClock_TIMER1, true);
		timer1started = true;
	}

	OnElapsedFunc = OnElapsed;

	TIMER1->CNT = 0;

	/* Set top value to provided timeout value if smaller than the timer length */
	if (TIMER_MULTI * sec * 1000 < TIMER_LEN){
		TIMER1->TOP = TIMER_MULTI * sec * 1000;
		timer1val = 0;
	}

	else {
		TIMER1->TOP = TIMER_LEN;
		timer1val = TIMER_MULTI * sec * 1000 - TIMER_LEN;
	}

	TIMER1->CMD = TIMER_CMD_START;
}


/*******************************************************************************
 * @brief Resets TIMER1 set by "StartTmer_s".
 *******************************************************************************/
void ResetTimer(){

	OnElapsedFunc = NULL;
	TIMER1->CMD = TIMER_CMD_STOP;

	if (timer1started){

		/* Disable clock */
		CMU_ClockEnable(cmuClock_TIMER1, false);
		timer1started = false;
	}
}


///*******************************************************************************
// * @brief Starts TIMER2 and blocks until the specified time has elapsed.
// *******************************************************************************/
//void Delay_ms(uint16_t millis){
//
//	/* Enable clock */
//	CMU_ClockEnable(cmuClock_TIMER2, true);
//
//	/* Clear TIMER0 counter value */
//	TIMER2->CNT = 0;
//	timer2timeout = false;
//
//	if (TIMER_MULTI * millis < TIMER_LEN){
//		TIMER2->TOP = TIMER_MULTI * millis;
//		timer2val = 0;
//	}
//	else {
//		TIMER2->TOP = TIMER_LEN;
//		timer2val = TIMER_MULTI * millis - TIMER_LEN;
//	}
//
//	TIMER2->CMD = TIMER_CMD_START;
//
//	/* Wait for time to elapse */
//	while (!timer2timeout);
//}

/*******************************************************************************
 * @brief Starts TIMER2 and blocks until the specified time has elapsed.
 * Max 2000ms!
 *******************************************************************************/
void Delay_ms(uint16_t millis){

	/* Enable clock */
	CMU_ClockEnable(cmuClock_TIMER2, true);

	/* Clear TIMER0 counter value */
	TIMER2->CNT = 0;
	TIMER2->CMD = TIMER_CMD_START;
	while (millis * TIMER_MULTI > TIMER2->CNT);
	TIMER2->CMD = TIMER_CMD_STOP;
	CMU_ClockEnable(cmuClock_TIMER2, false);
}


/*******************************************************************************
 * @brief Starts SysTick timer and goes to sleep (EM1) until the specified time
 * has elapsed, or an interrupt causes a wake up. untilElapsed defines whether
 * to sleep until the defined time has elapsed. The RTC must be enabled for
 * the timer to work!
 *******************************************************************************/
void Sleep_ms(uint16_t millis, bool untilElapsed){

	SysTick->LOAD = (RTC_TICKS_PR_SEC / 1000) * millis - 1;
	SysTick->VAL = 0;

	SysTick->CTRL &= ~SysTick_CTRL_CLKSOURCE_Msk; /* Set clock source to bit 0 of RTC counter */
	SysTick->CTRL |= SysTick_CTRL_ENABLE_Msk | SysTick_CTRL_TICKINT_Msk;


	do {
		/* Enter sleep */
		EMU_EnterEM1();
	}
	/* Continue until time elapsed if untilElapsed set */
	while (untilElapsed &&
			!(SysTick->CTRL & SysTick_CTRL_COUNTFLAG_Msk));

	/* Disable timer */
	SysTick->CTRL &= ~SysTick_CTRL_ENABLE_Msk;
}


#if TIMER_USE_LETIMER
/*******************************************************************************
 * @brief Start low energy timer with the selected seconds timeout. Callback to
 * provided function when time has elapsed.
 *******************************************************************************/
void StartLETimer_s(uint16_t sec, void(*OnElapsed)()){

	OnElapsedLEFunc = OnElapsed;

	if (LETIMER0->CNT < sec * LETIMER_TICKS_PR_SEC){

		LETIMER_CompareSet(LETIMER0, 0, LETIMER_LEN - sec * LETIMER_TICKS_PR_SEC + LETIMER0->CNT);
	}

	else LETIMER_CompareSet(LETIMER0, 0, LETIMER0->CNT - sec * LETIMER_TICKS_PR_SEC);
}


void ResetLETimer(){

	OnElapsedLEFunc = NULL;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void InitLETimer(){

	CMU_ClockEnable(cmuClock_LETIMER0, true);

	/* Prescaler of 32768 = 1 sek ticks */
	CMU_ClockDivSet(cmuClock_LETIMER0, LETIMER_CLK_DIV);

	LETIMER_Init_TypeDef letimerInit = {
	  .enable         = true,
	  .debugRun       = false,
	  .rtcComp0Enable = false,
	  .rtcComp1Enable = false,
	  .comp0Top       = false,
	  .bufTop         = false,
	  .out0Pol        = 0,
	  .out1Pol        = 0,
	  .ufoa0          = letimerUFOANone,
	  .ufoa1          = letimerUFOANone,
	  .repMode        = letimerRepeatFree
	};

	/* Initialize LETIMER */
	LETIMER_Init(LETIMER0, &letimerInit);

	LETIMER_IntEnable(LETIMER0, LETIMER_IEN_COMP0);
	NVIC_EnableIRQ(LETIMER0_IRQn);
}
#endif


/*******************************************************************************
 * 							Interrupt handlers
 *******************************************************************************/


void SysTick_Handler(){}


void LETIMER0_IRQHandler(){

	if (LETIMER0->IF & LETIMER_IF_COMP0){

		LETIMER0->IFC = LETIMER_IF_COMP0;

		/* Call registered callback function when time elapse */
		if (OnElapsedLEFunc) OnElapsedLEFunc();
	}
}


void TIMER0_IRQHandler(){

	uint32_t flags = TIMER0->IF;
	TIMER0->IFC = flags;

	if (flags & TIMER_IF_OF){

		if (timer0val){

			if (timer0val < TIMER_LEN){
				TIMER0->TOP = timer0val;
				timer0val = 0;
			}

			else timer0val -= TIMER_LEN;
		}

		else {

			TIMER0->CMD = TIMER_CMD_STOP;
			timer0started = false;

			/* Disable clock */
			CMU_ClockEnable(cmuClock_TIMER0, false);

			timer0timeout = true;
		}
	}
}


void TIMER1_IRQHandler(){

	uint32_t flags = TIMER1->IF;
	TIMER1->IFC = flags;

	if (flags & TIMER_IF_OF){

		if (timer1val){

			if (timer1val < TIMER_LEN){
				TIMER1->TOP = timer1val;
				timer1val = 0;
			}

			else timer1val -= TIMER_LEN;
		}

		else {

			TIMER1->CMD = TIMER_CMD_STOP;
			timer1started = false;

			/* Disable clock */
			CMU_ClockEnable(cmuClock_TIMER1, false);

			/* Invoke callback function when time has elapsed */
			if (OnElapsedFunc) OnElapsedFunc();
		}
	}
}


void TIMER2_IRQHandler(){

	uint32_t flags = TIMER2->IF;
	TIMER2->IFC = flags;

	if (flags & TIMER_IF_OF){

		if (timer2val){

			if (timer2val < TIMER_LEN){
				TIMER2->TOP = timer2val;
				timer2val = 0;
			}

			else timer2val -= TIMER_LEN;
		}

		else {

			TIMER2->CMD = TIMER_CMD_STOP;

			/* Disable clock */
			CMU_ClockEnable(cmuClock_TIMER2, false);

			timer2timeout = true;
		}
	}
}
