/******************************************************************************
 * @file:		ubx_parser.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	26. jan. 2016
 ******************************************************************************
 *	Parser for the proprietary UBX protocol for the U-blox GPS module.
 ******************************************************************************/

#include "ubx_parser.h"
#include "err.h"

/*******************************************************************************
 * 				Private function -and variable declarations
 *******************************************************************************/


static bool inSPIMode;

static void WrapUbxMessage(UBX_Message * msg, uint8_t * ubxPacket);
static bool UnwrapRecvUbxPacket(uint8_t * recvUbxPacket,
		uint8_t * recvPayload, UBX_Message * sentMsg);
static bool ReceiveAck(UBX_Message * sentMsg);
static void SendPacket(uint8_t * packet, uint16_t packetLen);
static bool ReceivePacket(uint8_t * recvPacket, uint16_t recvPacketLen);

/* Function pointers */
static void 	(*TransmitBytePtr)(uint8_t);
static uint8_t 	(*ReceiveBytePtr)();


/*******************************************************************************
 * 								Public functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
void UBX_Init(
		void (*TransmitByte)(uint8_t),
		uint8_t (*ReceiveByte)(),
		bool spiMode){

	inSPIMode = spiMode;
	TransmitBytePtr = TransmitByte;
	ReceiveBytePtr = ReceiveByte;
}


/*******************************************************************************
 * @brief Transmits provided UBX message, and puts received data in recvPayload.
 *******************************************************************************/
bool UBX_TransmitMessage(
		UBX_Message * msg,
		uint8_t * recvPayload){

	int retryCount = 0;
	bool packetOK = false;

	uint16_t ubxPacketLen = msg->payloadLen + 8;
	uint8_t ubxPacket[ubxPacketLen];

	uint16_t recvUbxPacketLen = msg->recvPayloadLen + 8;
	uint8_t recvUbxPacket[recvUbxPacketLen];

	WrapUbxMessage(msg, ubxPacket);

	if (inSPIMode) SPI_Select(spi_gps);

	/* ALlows one retry if bad data */
	while (retryCount < 2 && !packetOK){

		SendPacket(ubxPacket, ubxPacketLen);

		if (!msg->poll) packetOK = ReceiveAck(msg);

		else {
			packetOK = ReceivePacket(recvUbxPacket, recvUbxPacketLen);

			if (packetOK) packetOK = UnwrapRecvUbxPacket(recvUbxPacket, recvPayload, msg);
		}
		retryCount++;
	}

	if (inSPIMode) SPI_DeSelect(spi_gps);
	Sleep_ms(15, true); /* To let the GPS module process the message */

	return packetOK;
}


/*******************************************************************************
 * @brief Receive defined UBX message into recvPayload. Does not poll for data.
 *******************************************************************************/
bool UBX_ReceiveMessage(
		UBX_Message * msg,
		uint8_t * recvPayload){

	bool packetOK = false;
	uint16_t recvUbxPacketLen = msg->recvPayloadLen + 8;
	uint8_t	recvUbxPacket[recvUbxPacketLen];


	if (inSPIMode) SPI_Select(spi_gps);

	packetOK = ReceivePacket(recvUbxPacket, recvUbxPacketLen);

	if (packetOK) packetOK = UnwrapRecvUbxPacket(recvUbxPacket, recvPayload, msg);

	if (inSPIMode) SPI_DeSelect(spi_gps);

	return packetOK;
}


/*******************************************************************************
 * 								Private functions
 *******************************************************************************/


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool ReceiveAck(UBX_Message * sentMsg){

	bool packetOK = true;

	UBX_Message ackMsg = UBX_MESSAGE_ACK;

	uint16_t ackNakPacketLen = ackMsg.recvPayloadLen + 8;
	uint8_t ackNakPacket[ackNakPacketLen];

	uint8_t recvPayload[ackMsg.recvPayloadLen];

	packetOK = ReceivePacket(ackNakPacket, ackNakPacketLen);

	if (packetOK) packetOK = UnwrapRecvUbxPacket(
			ackNakPacket, recvPayload, &ackMsg);

	if (!packetOK &&
			error.erRes == err_dataNotValid) ERR_RAISE(err_nakRecv, dev_gps);

	/* Check if the Ack belongs to the sent packet */
	if (packetOK && (recvPayload[0] != sentMsg->msgClass ||
			recvPayload[1] != sentMsg->msgID)){

		packetOK = false;
		ERR_RAISE(err_dataNotValid, dev_gps);
	}

	return packetOK;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void SendPacket(uint8_t * packet, uint16_t packetLen){

	for (uint16_t i = 0; i < packetLen; i++){

		TransmitBytePtr(packet[i]);
	}
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool ReceivePacket(uint8_t * recvPacket, uint16_t recvPacketLen){

	uint8_t recvByte = 0xFF;

	/* Wait for the first valid msg char.
	 * Timeout if it is not received within
	 * the specified time.
	 */
	StartTimer_ms(UBX_RECEIVE_TIMEOUT_MS);
	while (recvByte != UBX_SYNC_CHAR_1){

		if (TimeElapsed()){

			ERR_RAISE(err_rxTimeout, dev_gps);
			return false;
		}

		recvByte = ReceiveBytePtr();
	}

	recvPacket[0] = recvByte;

	for (uint16_t j = 1; j < recvPacketLen; j++){

		 recvPacket[j] = ReceiveBytePtr();
	}
	return true;
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void WrapUbxMessage(UBX_Message * msg, uint8_t * ubxPacket){

	 uint8_t ck_a = 0, ck_b = 0;
	 uint8_t lb1, lb2;

	 /* Split payload_len into two bytes */
	 lb1 = (uint8_t)((msg->payloadLen & 0xFF00) >> 8);
	 lb2 = (uint8_t)(msg->payloadLen & 0x00FF);

	 /* Build message */
	 ubxPacket[0] = UBX_SYNC_CHAR_1;
	 ubxPacket[1] = UBX_SYNC_CHAR_2;
	 ubxPacket[2] = msg->msgClass;
	 ubxPacket[3] = msg->msgID;;
	 ubxPacket[4] = lb2; /* Little endian format */
	 ubxPacket[5] = lb1;

	 for (uint16_t i = 0; i < msg->payloadLen; i++){
		 ubxPacket[i + 6] = msg->payload[i];
	 }

	 /* Calculate checksums */
	 for (uint16_t a = 2; a < msg->payloadLen + 6; a++){

		 ck_a = ck_a + ubxPacket[a];
		 ck_b = ck_b + ck_a;
	 }

	 ubxPacket[5 + msg->payloadLen + 1] = ck_a;
	 ubxPacket[5 + msg->payloadLen + 2] = ck_b;

}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static bool UnwrapRecvUbxPacket(uint8_t * recvUbxPacket,
		uint8_t * recvPayload, UBX_Message * sentMsg){

	 uint8_t ck_a = 0, ck_b = 0;
	 bool res = true;

	 /* Calculate checksums */

	 for (uint16_t a = 2; a < sentMsg->recvPayloadLen + 6; a++){

		 ck_a = ck_a + recvUbxPacket[a];
		 ck_b = ck_b + ck_a;
	 }


	 /* Compare calculated and received checksum */

	 if ((recvUbxPacket[5 + sentMsg->recvPayloadLen + 1] != ck_a) ||
			 (recvUbxPacket[5 + sentMsg->recvPayloadLen + 2] != ck_b)){

		 res = false;
		 ERR_RAISE(err_badCheckSum, dev_gps);
		 goto end;
	 }


	 /* Compare sent class -and message IDs to received message. */

	 if (sentMsg->msgClass != recvUbxPacket[2] ||
			 sentMsg->msgID != recvUbxPacket[3]){

		 res = false;
		 ERR_RAISE(err_dataNotValid, dev_gps);
		 goto end;
	 }


	 for (uint16_t i = 0; i < sentMsg->recvPayloadLen; i++){

		 recvPayload[i] = recvUbxPacket[i + 6];
	 }
end:
	 return res;
}
