/******************************************************************************
 * @file:		checkStates.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	17. mai 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include "shared.h"

#include "checkStates.h"
#include "deviceStates.h"
#include "comStates.h"
#include "updateStates.h"
#include "exceptionState.h"
#include "initState.h"

/*******************************************************************************
 * 							Transition functions
 *******************************************************************************/


void TransitionToCheckStream(ControllerStatePtr state){

	state->state = CheckStream;
	state->transition = TransitionToCheckTasks;
}


void TransitionToCheckTasks(ControllerStatePtr state){

	state->state = CheckTasks;
	state->transition = TransitionToCheckComState;
}


void TransitionToCheckComState(ControllerStatePtr state){

	state->state = CheckComState;
	state->transition = TransitionToCheckForSystemTasks;
}


void TransitionToCheckForSystemTasks(ControllerStatePtr state){

	state->state = CheckForSystemTasks;
	state->transition = TransitionToCheckGsmAv;
}


void TransitionToCheckGsmAv(ControllerStatePtr state){

	state->state = CheckGsmAvailable;
	state->transition = TransitionToCheckGpsAv;
}


void TransitionToCheckGpsAv(ControllerStatePtr state){

	state->state = CheckGpsAvailable;
	state->transition = TransitionToSystemSleep;
}


void TransitionToSystemSleep(ControllerStatePtr state){

	DefaultImplementation(state);
	state->state = SystemSleep;
}


/*******************************************************************************
 * 							 State functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Check for data available in stream.
 *******************************************************************************/
void CheckStream(ControllerStatePtr state){

	Stream currentStream = STREAM_DataAvailable();

	if (currentStream != stream_none){

		state->currentDevice = STREAM_TO_DEV(currentStream);

		state->transition = TransitionToReceiveData;
	}
}


/*******************************************************************************
 * @brief Check communication state.
 *******************************************************************************/
void CheckComState(ControllerStatePtr state){

	Device_dev deviceLog;
	COM_State comState = COM_GetState(&deviceLog);

	switch(comState){

	case state_recvFinished_local:
	case state_transFinished_local:
	case state_transInProgress_local:

		state->currentDevice = dev_radio;
		state->transition = TransitionToSendData;
		break;

	case state_recvFinished_remote:
	case state_transFinished_remote:
	case state_transInProgress_remote:

		state->currentDevice = dev_gsm;
		state->transition = TransitionToSendData;
		break;

	case state_ackReceived_local:
	case state_ackReceived_remote:

		if (deviceLog != dev_noDevice){

			/* Save last read file and file position */
			LOG_SaveDeviceOpParam(deviceLog, type_local);

			if (comState == state_ackReceived_remote){
				LOG_SaveDeviceOpParam(deviceLog, type_combined);
			}

			if (LOG_CloseOpParamFile()){
				DEV_ShutdownDevice(dev_sdCard);
			}
		}

		break;

	case state_nakReceived_local:
	case state_recoverLogParam_local:

		if (deviceLog != dev_noDevice){

			DEV_WakeUpDevice(dev_sdCard);

			/* Reset last read file and file position */
			LOG_LoadDeviceOpParam(deviceLog, type_local);

			if (LOG_CloseOpParamFile()){
				DEV_ShutdownDevice(dev_sdCard);
			}
		}

		if (comState != state_recoverLogParam_local){

			/* Retransmit */
			state->currentDevice = dev_radio;
			state->transition = TransitionToSendData;
		}

		break;

	case state_nakReceived_remote:
	case state_recoverLogParam_remote:

		if (deviceLog != dev_noDevice){

			DEV_WakeUpDevice(dev_sdCard);

			/* Reset last read file and file position */
			LOG_LoadDeviceOpParam(deviceLog, type_local);
			LOG_LoadDeviceOpParam(deviceLog, type_combined);

			if (LOG_CloseOpParamFile()){
				DEV_ShutdownDevice(dev_sdCard);
			}
		}

		if (comState != state_recoverLogParam_remote){
			/* Retransmit */
			state->currentDevice = dev_gsm;
			state->transition = TransitionToSendData;
		}

		break;

	case state_localSessionComplete:

		state->currentDevice = dev_radio;
		state->transition = TransitionToShutdownDevice;

		/* Close all log files and shut down SD card */
		LOG_CloseAllOpen();
		DEV_ShutdownDevice(dev_sdCard);
		break;

	case state_remoteSessionComplete:

		state->currentDevice = dev_gsm;
		state->transition = TransitionToShutdownDevice;

		/* Close all log files and shut down SD card */
		LOG_CloseAllOpen();
		DEV_ShutdownDevice(dev_sdCard);
		break;

	default:
		break;
	}
}


/*******************************************************************************
 * @brief Check system task flags.
 *******************************************************************************/
void CheckForSystemTasks(ControllerStatePtr state){

	static BMP_MessageType_msg msgToSend;

	state->argPtr = &msgToSend;

	SYS_Task stat = SYS_GetTask();


	switch(stat){

	case task_startDataAcq:

		state->currentDevice = dev_radio;
		state->transition = TransitionToSendSpecifiedMessage;
		msgToSend = msg_pollClientStatus;
		break;

	case task_pollTbrData:

		state->currentDevice = dev_radio;
		state->transition = TransitionToSendSpecifiedMessage;
		msgToSend = msg_pollTbrDetections;
		break;

	case task_sysParamAboveTH:

//		state->currentDevice = dev_gsm;
//		state->transition = TransitionToWakeAndSendSms;
		break;

	case task_pollGpsData:

		state->currentDevice = dev_radio;
		state->transition = TransitionToSendSpecifiedMessage;
		msgToSend = msg_pollGpsData;
		break;

	case task_pollGaugeData:

		state->currentDevice = dev_radio;
		state->transition = TransitionToSendSpecifiedMessage;
		msgToSend = msg_pollGaugeData;
		break;

	case task_endDataAcqSession:

		state->currentDevice = dev_radio;
		state->transition = TransitionToEndSession;
		msgToSend = msg_sessionFinished;

		/* Close all log files and shut down SD card */
		LOG_CloseAllOpen();
		DEV_ShutdownDevice(dev_sdCard);
		break;

	default:
		break;
	}
}


/*******************************************************************************
 * @brief Check for data available in GPS buffer (if TX ready feature enabled).
 *******************************************************************************/
void CheckGpsAvailable(ControllerStatePtr state){

	if (DEV_DataAv(dev_gps)){

		state->currentDevice = dev_gps;
		state->transition = TransitionToCollectData;
	}
}


/*******************************************************************************
 * @brief Check for data available from GSM modem (SMS or call).
 *******************************************************************************/
void CheckGsmAvailable(ControllerStatePtr state){

	if (DEV_DataAv(dev_gsm)){

		state->transition = TransitionToWakeAndReceiveSms;
	}
}


/*******************************************************************************
 * @brief Check time manager generated tasks.
 *******************************************************************************/
void CheckTasks(ControllerStatePtr state){


	TIME_Event transitionTask = TIME_GetEvent(&state->currentDevice);

	switch(transitionTask){

	case event_indicatorTimeout:

	#if !DEBUG_OPERATION
		indicatorTimeout = true;
		DEV_SetLED(false, led_green);
		DEV_SetLED(false, led_yellow);
		DEV_SetLED(false, led_red);
	#endif

		break;

	case event_midnight:
		state->transition = TransitionToCreateLogFiles;
		break;

	case event_rtcTempComp:
		state->transition = TransitionToTempComp;
		break;

	case event_watchdog:
		/* For example:
		 * Check connections to devices
		 * Check free space on memory card -> delete files if over threshold
		 * Check system threshold flags -> report/ log errors etc.
		 */
		break;

	case event_shutdownDevice:
		state->transition = TransitionToShutdownDevice;
		break;

	case event_wakeupDevice:
		state->transition = TransitionToWakeUpDevice;
		break;

	default:
		break;
	}
}


/*******************************************************************************
 * @brief Takes the system into sleep, either EM1 or EM2 depending on if there
 *  are any devices active that requires to stay in EM1 (HF OSC available).
 *******************************************************************************/
void SystemSleep(ControllerStatePtr state){

	if (!indicatorTimeout){
		DEV_SetLED(false, led_green);
		DEV_SetLED(true, led_yellow);
	}

	TIME_Sleep(DEV_LightSleepDeviceActive());

	if (!indicatorTimeout){
		DEV_SetLED(true, led_green);
		DEV_SetLED(false, led_yellow);
	}
}
