/******************************************************************************
 * @file:		exceptionState.c
 * @project:	BuoyController
 * @author:  	Ove Nesse
 * @version:	1.00
 * @created:	17. mai 2016
 ******************************************************************************
 *	File info here
 ******************************************************************************/

#include "checkStates.h"
#include "deviceStates.h"
#include "comStates.h"
#include "updateStates.h"
#include "exceptionState.h"
#include "initState.h"

static BMP_MessageType_msg msgToSend;
static void * lastTransition, * lastState;

static void TransitionToPrevious(ControllerStatePtr state);
static void OnSendNak(ControllerStatePtr state);
/*******************************************************************************
 * 							Transition functions
 *******************************************************************************/


void OnException(ControllerStatePtr state){

	lastTransition = state->transition;
	lastState = state->state;

	state->transition = TransitionToHandleException;
}


void TransitionToHandleException(ControllerStatePtr state){

	state->transition = TransitionToPrevious;
	state->state = HandleException;
}


static void TransitionToPrevious(ControllerStatePtr state){

	state->transition = lastTransition;
	state->state = lastState;
}


/*******************************************************************************
 * 							 State functions
 *******************************************************************************/


/*******************************************************************************
 * @brief Highest level exception handler.
 *******************************************************************************/
void HandleException(ControllerStatePtr state){


	/* Continue when currently in init state */
	if (state->init) state->transition = TransitionToPrevious;

	/* Only handle error if not resolved in
	 * the lower level error handler */
	else if (error.retVal == ret_notResolved){

		switch (error.erRes){

		case err_majorFailure:
		case err_simError: /* SIM card could not be recognized */

			/* Cannot continue before error
			 * is resolved and system restart */
			while(true){

				DEV_SetLED(true, led_red);
				Sleep_ms(100, true);
				DEV_SetLED(false, led_red);
				Sleep_ms(100, true);
			}
			break;

		case err_badFileName:
		case err_noFile:
		case err_headerErr:
		case err_rxTimeout:

			/* Only send NAK if error origin via one of the
			 * communication devices (provided file name not
			 * correct etc).
			 */
			if (state->currentDevice == dev_radio ||
					state->currentDevice == dev_gsm){

				if (error.errCount < 2) OnSendNak(state);

				else {
					ERR_RESET();
					DefaultImplementation(state);
				}
			}

			break;

		case err_sohTimeout:

			/* Disregard error, nothing to be done, return to default state */
			DefaultImplementation(state);

			break;

		case err_storageFull:
			// Delete log file(s) ?
			break;

		case err_sdCardNotDetected:

			/* Error is handled in device manager since it is a
			 * device-related error. */
			DEV_Error_Handler();

			break;

		case err_txTimeout:
			DefaultImplementation(state);
			break;

		default:
			break;
		}
	}

	/* Return to the previous state. */
	else if (error.retVal == ret_resolvedRetry){
		state->transition = TransitionToPrevious;
	}

	/* Go to default state if error resolved */
	else DefaultImplementation(state);
}


/*******************************************************************************
 * @brief
 *******************************************************************************/
static void OnSendNak(ControllerStatePtr state){

	DefaultImplementation(state);

	if (state->currentDevice == dev_gsm ||
		state->currentDevice == dev_radio){

		state->transition = TransitionToSendSpecifiedMessage;
		msgToSend = msg_nak;
		state->argPtr = &msgToSend;
	}
}
